/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::NJointController
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "NJointController.h"

#include <ArmarXCore/core/util/algorithm.h>

#include "../RobotUnit.h"

using namespace armarx;

const NJointControllerPtr NJointController::NullPtr
{
    nullptr
};

NJointControllerDescription NJointController::getControllerDescription(const Ice::Current&) const
{
    NJointControllerDescription d;
    d.className = getClassName();
    ARMARX_CHECK_EXPRESSION(getProxy(-1));
    d.controller = NJointControllerInterfacePrx::uncheckedCast(getProxy(-1));
    d.controlModeAssignment = controlDeviceControlModeMap;
    d.instanceName = getInstanceName();
    d.deletable = deletable;
    d.internal = internal;
    return d;
}

boost::optional<std::vector<char> > NJointController::isNotInConflictWith(const std::vector<char>& used) const
{
    ARMARX_CHECK_EXPRESSION(used.size() == controlDeviceUsedBitmap.size());
    auto result = used;
    for (std::size_t i = 0; i < controlDeviceUsedBitmap.size(); ++i)
    {
        if (controlDeviceUsedBitmap.at(i))
        {
            if (used.at(i))
            {
                return boost::optional<std::vector<char>>();
            }
            result.at(i) = true;
        }

    }
    return {true, std::move(result)};
}

NJointControllerStatus NJointController::getControllerStatus(const Ice::Current&) const
{
    NJointControllerStatus s;
    s.active = isActive;
    s.instanceName = getInstanceName();
    s.requested = isRequested;
    s.error = deactivatedBecauseOfError;
    s.timestampUSec = std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now().time_since_epoch()).count();
    return s;
}

NJointControllerDescriptionWithStatus NJointController::getControllerDescriptionWithStatus(const Ice::Current&) const
{
    NJointControllerDescriptionWithStatus ds;
    ds.description = getControllerDescription();
    ds.status = getControllerStatus();
    return ds;
}

RobotUnitInterfacePrx NJointController::getRobotUnit(const Ice::Current&) const
{
    ARMARX_CHECK_EXPRESSION(robotUnit);
    return RobotUnitInterfacePrx::uncheckedCast(robotUnit->getProxy(-1));
}

void NJointController::activateController(const Ice::Current&) const
{
    ARMARX_CHECK_EXPRESSION(robotUnit);
    robotUnit->activateNJointController(getInstanceName());
}

void NJointController::deactivateController(const Ice::Current&) const
{
    ARMARX_CHECK_EXPRESSION(robotUnit);
    robotUnit->deactivateNJointController(getInstanceName());
}

void NJointController::deleteController(const Ice::Current&) const
{
    ARMARX_CHECK_EXPRESSION(robotUnit);
    robotUnit->deleteNJointController(getInstanceName());
}

void NJointController::publish(const SensorAndControl& sac, const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer)
{
    const bool active = isActive;
    if (publishActive.exchange(active) != active)
    {
        if (active)
        {
            ARMARX_DEBUG << "activating publishing for " << getInstanceName();
            try
            {
                onPublishActivation(draw, observer);
            }
            catch (std::exception& e)
            {
                ARMARX_WARNING << "onPublishActivation of '" << getInstanceName() << "' threw an exception!\nWhat:\n" << e.what();
            }
            catch (...)
            {
                ARMARX_WARNING << "onPublishActivation of '" << getInstanceName() << "' threw an exception!";
            }
        }
        else
        {
            ARMARX_DEBUG << "deactivating publishing for " << getInstanceName();
            try
            {
                onPublishDeactivation(draw, observer);
            }
            catch (std::exception& e)
            {
                ARMARX_WARNING << "onPublishDeactivation of '" << getInstanceName() << "' threw an exception!\nWhat:\n" << e.what();
            }
            catch (...)
            {
                ARMARX_WARNING << "onPublishDeactivation of '" << getInstanceName() << "' threw an exception!";
            }
        }
    }
    if (active)
    {
        try
        {
            onPublish(sac, draw, observer);
        }
        catch (std::exception& e)
        {
            ARMARX_WARNING << "onPublish of '" << getInstanceName() << "' threw an exception!\nWhat:\n" << e.what();
        }
        catch (...)
        {
            ARMARX_WARNING << "onPublish of '" << getInstanceName() << "' threw an exception!";
        }
    }
}

void NJointController::deactivatePublish(const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer)
{
    if (publishActive.exchange(false))
    {
        ARMARX_DEBUG << "forced deactivation of publishing for " << getInstanceName();
        onPublishDeactivation(draw, observer);
    }
}

void NJointController::rtActivateController()
{
    if (!isActive)
    {
        deactivatedBecauseOfError = false;
        rtPreActivateController();
        isActive = true;
    }
}

void NJointController::rtDeactivateController()
{
    if (isActive)
    {
        isActive = false;
        rtPostDeactivateController();
    }
}

void NJointController::rtDeactivateControllerBecauseOfError()
{
    deactivatedBecauseOfError = true;
    rtDeactivateController();
}

void NJointController::robotUnitInit(
    RobotUnit* ru,
    std::size_t ctrlId,
    std::map<std::string, const JointController*> ctrlDeviceUsedJointController,
    std::vector<char> ctrlDeviceUsedBitmap,
    std::vector<std::size_t> ctrlDeviceUsedIndices,
    bool ctrlDeletable,
    bool ctrlInternal)
{
    ARMARX_CHECK_EXPRESSION(ru);
    ARMARX_CHECK_EXPRESSION_W_HINT(ctrlDeviceUsedJointController.size(), "Controller requested zero joints");
    ARMARX_CHECK_EXPRESSION(ctrlId != std::numeric_limits<std::size_t>::max());
    ARMARX_CHECK_EXPRESSION_W_HINT(ru->getNumberOfControlDevices() == ctrlDeviceUsedBitmap.size(),
                                   "ru->getNumberOfControlDevices() == ctrlDeviceUsedBitmap.size() ->" +
                                   to_string(ru->getNumberOfControlDevices()) + " > " + to_string(ctrlDeviceUsedBitmap.size()));
    ARMARX_CHECK_EXPRESSION(ctrlDeviceUsedJointController.size() == ctrlDeviceUsedIndices.size());

    robotUnit = ru;
    id = ctrlId;
    controlDeviceUsedJointController = std::move(ctrlDeviceUsedJointController);
    for (const auto& pair : controlDeviceUsedJointController)
    {
        ARMARX_CHECK_EXPRESSION(pair.second);
        controlDeviceControlModeMap[pair.first] = pair.second->getControlMode();
    }
    controlDeviceUsedBitmap     = std::move(ctrlDeviceUsedBitmap);
    controlDeviceUsedIndices    = std::move(ctrlDeviceUsedIndices);
    deletable = ctrlDeletable;
    internal = ctrlInternal;
}

