/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::NJointController
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <map>
#include <atomic>
#include <mutex>
#include <functional>
#include <unordered_set>

#include <boost/optional.hpp>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/util/Registrar.h>
#include <ArmarXCore/core/util/TripleBuffer.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <ArmarXGui/interface/WidgetDescription.h>

#include <RobotAPI/interface/units/RobotUnit/NJointController.h>
#include <RobotAPI/interface/visualization/DebugDrawerInterface.h>
#include <RobotAPI/interface/units/RobotUnit/RobotUnitInterface.h>

#include "../ControlTargets/ControlTargetBase.h"

//units
#include "../Devices/ControlDevice.h"
#include "../Devices/SensorDevice.h"

#include "../util.h"
#include "../util/JointAndNJointControllers.h"
#include "../util/ControlThreadOutputBuffer.h"

namespace VirtualRobot
{
    class Robot;
}

namespace armarx
{
    TYPEDEF_PTRS_HANDLE(NJointController);
    TYPEDEF_PTRS_HANDLE(NJointControllerDescriptionProviderInterface);

    class RobotUnit;

    class NJointControllerDescriptionProviderInterface: virtual public Ice::Object
    {
    public:
        ~NJointControllerDescriptionProviderInterface() override = default;

        virtual ConstSensorDevicePtr   getSensorDevice(const std::string& sensorDeviceName) const = 0;
        virtual const SensorValueBase* getSensorValue(const std::string& sensorDeviceName) const;
        virtual ConstControlDevicePtr getControlDevice(const std::string& deviceName) const = 0;
        virtual ControlTargetBase* getControlTarget(const std::string& deviceName, const std::string& controlMode) = 0;

        virtual std::string getName() const = 0;

        template<class T>
        const T* getSensorValue(const std::string& sensorDeviceName) const;
        template<class T>
        T* getControlTarget(const std::string& deviceName, const std::string& controlMode);
    };

    /**
    * @defgroup Library-RobotUnit-NJointControllers NJointControllers
    * @ingroup Library-RobotUnit
    *
    *
    * @defgroup Library-RobotUnit RobotUnit
    * @ingroup RobotAPI
    * A description of the library RobotUnit.
    */

    /**
    * @ingroup Library-RobotUnit-NJointControllers
    * @brief A high level controller writing its results into targets.
    *
    * This is the abstract base class for all NJointControllers.
    * It implements basic management routines required by the RobotUnit and some basic ice calls.
    * NJointControllers are intantiated and managed by RobotUnits.
    *
    * Each NJointController has to do the following in their constructor:
    * \li Request joints in exactly one control mode from the robot unit (the control mode can't be changed later)
    *     You will receive pointer targets where it can store its results.
    * \li Get pointer to the state of required joints / sensors. (you can't request other data at a later point)
    *     You will receive pointer to the values.
    *
    * Each controller has two parts:
    * \li The RT controll loop
    * \li The NonRT ice communication
    *
    * \section rtcontrollloop The RT controll loop (rtRun)
    * Here you have access to the robots current state and write your results into targets.
    * \warning This part has to satisfy realtime conditions!
    * All realtime able functions of this class have the 'rt' prefix.
    *
    * You must not:
    * \li call any blocking operation
    * \li allocate ram on the heap (since this is a blocking operation)
    * \li resize any datastructure (e.g. vector::resize) (this sometimes allocates ram on the heap)
    * \li insert new datafields into datastructures (e.g. vector::push_back) (this sometimes allocates ram on the heap)
    * \li write to any stream (e.g. ARMARX_VERBOSE, std::cout, print, filestreams) (this sometimes blocks/allocates ram on the heap)
    * \li make network calls (e.g. through ice) (this blocks/allocates ram on the heap) (using begin_... end_... version of ice calls IS NO SOLUTION)
    * \li do any expensive calculations (e.g. calculate IK, run some solver, invert big matrices)
    *
    * \section nonrtpart The NonRT ice communication
    * This part consits of any ice communication.
    * Here you could get new controll parameters or targets from other components.
    *
    * \section rtnrtcomm Communication between RT and NonRT
    * All communication between RT and NonRT has to be lockfree.
    * So you have to use constructs like atomics or TripleBuffers (See armarx::NJointControllerWithTripleBuffer).
    *
    * \image html NJointControllerGeneralDataFlow.svg "General Dataflow in a NJointController"
    *
    *
    *
    *
    * \image html NJointControllerAtomicDataFlow.svg "Dataflow in a NJointController using atomics for communication between RT and NonRT"
    *
    *
    *
    *
    * \image html NJointControllerTripleBufferDataFlow.svg "Dataflow in a NJointController using a triple buffer for communication between RT and NonRT"
    *
    * \section expensivework How to do expensive calculations
    * You have to execute expensive calculations in a different worker thread.
    * This thread could calculate target values at its own speed (e.g. 100 Hz).
    *
    * While rtRun runs at a higher frequency (e.g. 1000 Hz) and:
    * \li reads target values
    * \li optionally passes the target values to a PID controller
    * \li writes them to the targets
    * \li sends the sensor values to the worker thread.
    *
    * If you do some additional calculation in rtRun, you maybe need to pass config parameters from NonRT to RT using a nonblocking method.
    *
    * \image html NJointControllerWorkerThreadDataFlow.svg "Dataflow in a NJointController using a worker thread"
    *
    */
    class NJointController:
        virtual public NJointControllerInterface,
        virtual public ManagedIceObject
    {
    protected:
        std::string getDefaultName() const override;
        void onInitComponent() override;
        void onConnectComponent() override;
    public:
        using ConfigPtrT = NJointControllerConfigPtr;
        //add these static functions if you want to provide a gui for construction
        // static WidgetDescription::WidgetPtr GenerateConfigDescription
        // (
        //      const boost::shared_ptr<VirtualRobot::Robot>&,
        //      const std::map<std::string, ConstControlDevicePtr>& controlDevices,
        //      const std::map<std::string, ConstSensorDevicePtr>& sensorDevices
        // )
        // {
        //      return a nullptr / Widget if you want remote configuration but do not need any parameters
        // }
        // static ConfigPtrT GenerateConfigFromVariants(const StringVariantBaseMap&)
        // {
        //      return the config that is passed to your controller
        // }
        using GenerateConfigDescriptionFunctionSignature = WidgetDescription::WidgetPtr(*)
                (
                    const boost::shared_ptr<VirtualRobot::Robot>&,
                    const std::map<std::string, ConstControlDevicePtr>& controlDevices,
                    const std::map<std::string, ConstSensorDevicePtr>& sensorDevices
                );
        template<class ConfigPrtType>
        using GenerateConfigFromVariantsFunctionSignature = ConfigPrtType(*)(const StringVariantBaseMap&);

        static const NJointControllerPtr NullPtr;

        ~NJointController() override = default;
        //ice interface (must not be called in the rt thread)
        bool isControllerActive(const Ice::Current& = GlobalIceCurrent) const final override;
        bool isControllerRequested(const Ice::Current& = GlobalIceCurrent) const final override;
        bool isDeletable(const Ice::Current& = GlobalIceCurrent) const final override;
        bool hasControllerError(const Ice::Current& = GlobalIceCurrent) const final override;

        std::string getClassName(const Ice::Current& = GlobalIceCurrent) const override  = 0;
        std::string getInstanceName(const Ice::Current& = GlobalIceCurrent) const final override;

        NJointControllerDescription getControllerDescription(const Ice::Current& = GlobalIceCurrent) const final override;
        NJointControllerStatus getControllerStatus(const Ice::Current& = GlobalIceCurrent) const final override;
        NJointControllerDescriptionWithStatus getControllerDescriptionWithStatus(const Ice::Current& = GlobalIceCurrent) const final override;
        RobotUnitInterfacePrx getRobotUnit(const Ice::Current& = GlobalIceCurrent) const final override;

        void activateController(const Ice::Current& = GlobalIceCurrent) const final override;
        void deactivateController(const Ice::Current& = GlobalIceCurrent) const final override;
        void deleteController(const Ice::Current& = GlobalIceCurrent) const final override;

        WidgetDescription::StringWidgetDictionary getFunctionDescriptions(const Ice::Current& = GlobalIceCurrent) const override;
        void callDescribedFunction(const std::string&, const StringVariantBaseMap&, const Ice::Current& = GlobalIceCurrent) override;

        std::size_t getId() const;

        //c++ interface (local calls) (must be called in the rt thread)
        virtual void rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) = 0;
        virtual void rtSwapBufferAndRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
    protected:
        /**
         * @brief This function is called before the controller is activated.
         * You can use it to activate a thread again (DO NOT SPAWN NEW THREADS!) e.g. via a std::atomic_bool.
         */
        virtual void rtPreActivateController() {}
        /**
         * @brief This function is called after the controller is deactivated.
         * You can use it to deactivate a thread (DO NOT JOIN THREADS!) e.g. via a std::atomic_bool.
         */
        virtual void rtPostDeactivateController() {}
    public:
        bool rtUsesControlDevice(std::size_t deviceIndex) const;

        std::size_t rtGetId() const;

        //used control devices
        StringStringDictionary getControlDeviceUsedControlModeMap(const Ice::Current& = GlobalIceCurrent) const final override;
        const std::vector<char>& getControlDeviceUsedBitmap() const;
        const std::vector<std::size_t>& rtGetControlDeviceUsedIndices() const;
        const std::vector<std::size_t>& getControlDeviceUsedIndices() const;

        const std::map<std::string, const JointController*>& getControlDevicesUsedJointController();

        //check for conflict
        boost::optional<std::vector<char>> isNotInConflictWith(const NJointControllerPtr& other) const;
        boost::optional<std::vector<char>> isNotInConflictWith(const std::vector<char>& used) const;

        template<class ItT>
        static boost::optional<std::vector<char>> AreNotInConflict(ItT first, ItT last);
    protected:
        //publish thread hooks
        virtual void onPublishActivation(const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx&) {}
        virtual void onPublishDeactivation(const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx&) {}
        virtual void onPublish(const SensorAndControl&, const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx&) {}
    private:
        void publish(const SensorAndControl& sac, const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer);
        void deactivatePublish(const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer);

        void rtActivateController();
        void rtDeactivateController();
        void rtDeactivateControllerBecauseOfError();

        //firends
        friend class RobotUnitNJointControllerAccess;

        void robotUnitInit(RobotUnit* ru,
                           std::size_t ctrlId,
                           std::map<std::string, const JointController*> ctrlDeviceControlModeMap,
                           std::vector<char> ctrlDeviceUsedBitmap,
                           std::vector<std::size_t> ctrlDeviceUsedIndices,
                           bool ctrlDeletable,
                           bool ctrlInternal);

        RobotUnit* robotUnit;
        std::size_t id = std::numeric_limits<std::size_t>::max();
        StringStringDictionary controlDeviceControlModeMap;
        std::map<std::string, const JointController*> controlDeviceUsedJointController;
        std::vector<char> controlDeviceUsedBitmap;
        std::vector<std::size_t> controlDeviceUsedIndices;

        //this data is filled by the robot unit to provide convenience functions
        std::atomic_bool isActive {false};
        std::atomic_bool isRequested {false};
        std::atomic_bool deactivatedBecauseOfError {false};
        bool deletable {false};
        bool internal {false};

        std::atomic_bool publishActive {false};

        std::atomic_bool statusReportedActive {false};
        std::atomic_bool statusReportedRequested {false};

    public:
        struct IdComp
        {
            template<class PtrT1, class PtrT2>
            bool operator()(PtrT1 l, PtrT2 r) const;
        };
    };

    class NJointControllerRegistryEntry
    {
    private:
        friend class RobotUnit;
        virtual NJointControllerPtr create(NJointControllerDescriptionProviderInterfacePtr, const NJointControllerConfigPtr&, const boost::shared_ptr<VirtualRobot::Robot>&) const = 0;
        virtual WidgetDescription::WidgetPtr GenerateConfigDescription(const boost::shared_ptr<VirtualRobot::Robot>&, const std::map<std::string, ConstControlDevicePtr>&, const std::map<std::string, ConstSensorDevicePtr>&) const = 0;
        virtual NJointControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap&) const = 0;
        virtual bool hasRemoteConfiguration() const = 0;
    };
    using NJointControllerRegistry = Registrar<std::unique_ptr<NJointControllerRegistryEntry>>;

    template<class ControllerType>
    struct NJointControllerRegistration;

    template <typename ControlDataStruct>
    class NJointControllerWithTripleBuffer: virtual public NJointController
    {
    public:
        using MutexType = std::recursive_mutex;
        using LockGuardType = std::lock_guard<std::recursive_mutex>;

        NJointControllerWithTripleBuffer(const ControlDataStruct& initialCommands = ControlDataStruct()):
            controlDataTripleBuffer {initialCommands}
        {
        }

        void rtSwapBufferAndRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) override;

    protected:
        const ControlDataStruct& rtGetControlStruct() const;
        bool                     rtUpdateControlStruct();

        void writeControlStruct();
        ControlDataStruct& getWriterControlStruct();

        void setControlStruct(const ControlDataStruct& newStruct);

        void reinitTripleBuffer(const ControlDataStruct& initial);

        mutable MutexType controlDataMutex;
    private:
        WriteBufferedTripleBuffer<ControlDataStruct> controlDataTripleBuffer;
    };

    template <typename ControlDataStruct>
    using NJointControllerWithTripleBufferPtr = IceInternal::Handle<NJointControllerWithTripleBuffer<ControlDataStruct>>;
}

//inline functions
namespace armarx
{
    //NJointControllerDescriptionProviderInterface
    inline const SensorValueBase* NJointControllerDescriptionProviderInterface::getSensorValue(const std::string& sensorDeviceName) const
    {
        auto sd = getSensorDevice(sensorDeviceName);
        return sd ? sd->getSensorValue() : nullptr;
    }

    template<class T>
    inline const T* NJointControllerDescriptionProviderInterface::getSensorValue(const std::string& sensorDeviceName) const
    {
        return dynamic_cast<const T*>(getSensorValue(sensorDeviceName));
    }

    template<class T>
    inline T* NJointControllerDescriptionProviderInterface::getControlTarget(const std::string& deviceName, const std::string& controlMode)
    {
        return dynamic_cast<T*>(getControlTarget(deviceName, controlMode));
    }

    //NJointController
    inline std::size_t NJointController::getId() const
    {
        return id;
    }

    inline void NJointController::rtSwapBufferAndRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtRun(sensorValuesTimestamp, timeSinceLastIteration);
    }

    inline bool NJointController::rtUsesControlDevice(std::size_t deviceIndex) const
    {
        return controlDeviceUsedBitmap.at(deviceIndex);
    }

    inline std::size_t NJointController::rtGetId() const
    {
        return id;
    }

    inline StringStringDictionary NJointController::getControlDeviceUsedControlModeMap(const Ice::Current&) const
    {
        return controlDeviceControlModeMap;
    }

    inline const std::vector<char>& NJointController::getControlDeviceUsedBitmap() const
    {
        return controlDeviceUsedBitmap;
    }

    inline const std::vector<std::size_t>& NJointController::rtGetControlDeviceUsedIndices() const
    {
        return controlDeviceUsedIndices;
    }

    inline const std::vector<std::size_t>& NJointController::getControlDeviceUsedIndices() const
    {
        return controlDeviceUsedIndices;
    }

    inline const std::map<std::string, const JointController*>& NJointController::getControlDevicesUsedJointController()
    {
        return controlDeviceUsedJointController;
    }

    inline boost::optional<std::vector<char> > NJointController::isNotInConflictWith(const NJointControllerPtr& other) const
    {
        return isNotInConflictWith(other->getControlDeviceUsedBitmap());
    }

    inline std::string NJointController::getDefaultName() const
    {
        return getClassName();
    }

    inline void NJointController::onInitComponent()
    {
    }

    inline void NJointController::onConnectComponent()
    {
    }

    inline bool NJointController::isControllerActive(const Ice::Current&) const
    {
        return isActive;
    }

    inline bool NJointController::isControllerRequested(const Ice::Current&) const
    {
        return isRequested;
    }

    inline bool NJointController::isDeletable(const Ice::Current&) const
    {
        return deletable;
    }

    inline bool NJointController::hasControllerError(const Ice::Current&) const
    {
        return deactivatedBecauseOfError;
    }

    inline std::string NJointController::getInstanceName(const Ice::Current&) const
    {
        return getName();
    }

    inline WidgetDescription::StringWidgetDictionary NJointController::getFunctionDescriptions(const Ice::Current&) const
    {
        return {};
    }

    inline void NJointController::callDescribedFunction(const std::string&, const StringVariantBaseMap&, const Ice::Current&)
    {
    }

    template<class PtrT1, class PtrT2>
    inline bool NJointController::IdComp::operator()(PtrT1 l, PtrT2 r) const
    {
        //null is bigger than any other
        if (!l)
        {
            return false;
        }
        if (!r)
        {
            return true;
        }
        return l->id < r->id;
    }

    //NJointControllerWithTripleBuffer<ControlDataStruct>
    template <typename ControlDataStruct>
    void NJointControllerWithTripleBuffer<ControlDataStruct>::rtSwapBufferAndRun(
        const IceUtil::Time& sensorValuesTimestamp,
        const IceUtil::Time& timeSinceLastIteration)
    {
        rtUpdateControlStruct();
        rtRun(sensorValuesTimestamp, timeSinceLastIteration);
    }

    template <typename ControlDataStruct>
    const ControlDataStruct& NJointControllerWithTripleBuffer<ControlDataStruct>::rtGetControlStruct() const
    {
        return controlDataTripleBuffer.getReadBuffer();
    }

    template <typename ControlDataStruct>
    bool NJointControllerWithTripleBuffer<ControlDataStruct>::rtUpdateControlStruct()
    {
        return controlDataTripleBuffer.updateReadBuffer();
    }

    template <typename ControlDataStruct>
    void NJointControllerWithTripleBuffer<ControlDataStruct>::writeControlStruct()
    {
        //just lock to be save and reduce the impact of an error
        //also this allows code to unlock the mutex before calling this function
        //(can happen if some lockguard in a scope is used)
        LockGuardType lock(controlDataMutex);
        controlDataTripleBuffer.commitWrite();
    }

    template <typename ControlDataStruct>
    ControlDataStruct& NJointControllerWithTripleBuffer<ControlDataStruct>::getWriterControlStruct()
    {
        return controlDataTripleBuffer.getWriteBuffer();
    }

    template <typename ControlDataStruct>
    void NJointControllerWithTripleBuffer<ControlDataStruct>::setControlStruct(const ControlDataStruct& newStruct)
    {
        LockGuardType lock {controlDataMutex};
        getWriterControlStruct() = newStruct;
        writeControlStruct();
    }

    template <typename ControlDataStruct>
    void NJointControllerWithTripleBuffer<ControlDataStruct>::reinitTripleBuffer(const ControlDataStruct& initial)
    {
        controlDataTripleBuffer.reinitAllBuffers(initial);
    }
}
#include "NJointController.hpp"
