/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::NJointController
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_NJointController_HPP
#define _ARMARX_LIB_RobotAPI_NJointController_HPP

#include <ArmarXCore/core/util/TemplateMetaProgramming.h>
#include "NJointController.h"

namespace armarx
{
    template<class ItT>
    boost::optional<std::vector<char>> NJointController::AreNotInConflict(ItT first, ItT last)
    {
        if (first == last)
        {
            return {true, std::vector<char>{}};
        }
        std::size_t n = (*first)->getControlDeviceUsedBitmap().size();
        std::vector<char> inuse(n, false);
        while (first != last)
        {
            auto r = (*first)->isNotInConflictWith(inuse);
            if (!r)
            {
                return r;
            }
            inuse = std::move(r.get());
            ++first;
        }
        return {true, std::move(inuse)};
    }

    namespace detail
    {
        template<class NJointControllerT, class = void, class = void>
        class NJointControllerRegistryEntryHelper : public NJointControllerRegistryEntry
        {
            NJointControllerPtr create(NJointControllerDescriptionProviderInterfacePtr prov,
                                       const NJointControllerConfigPtr& config,
                                       const boost::shared_ptr<VirtualRobot::Robot>& rob) const final override
            {
                ARMARX_CHECK_EXPRESSION_W_HINT(prov, "DataProvider is NULL!");
                using ConfigPtrT = typename NJointControllerT::ConfigPtrT;
                ConfigPtrT cfg = ConfigPtrT::dynamicCast(config);
                ARMARX_CHECK_EXPRESSION_W_HINT(cfg, "The configuration is of the wrong type! it has to be an instance of: "
                                               << GetTypeString<ConfigPtrT>());
                return new NJointControllerT(prov, cfg, rob);
            }
            WidgetDescription::WidgetPtr GenerateConfigDescription(const boost::shared_ptr<VirtualRobot::Robot>&,
                    const std::map<std::string, ConstControlDevicePtr>&,
                    const std::map<std::string, ConstSensorDevicePtr>&) const final override
            {
                ARMARX_CHECK_EXPRESSION(!"This function should never be called");
            }
            NJointControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap&) const override
            {
                ARMARX_CHECK_EXPRESSION(!"This function should never be called");
            }
            bool hasRemoteConfiguration() const override
            {
                return false;
            }
        };

        ARMARX_META_MAKE_HAS_MEMBER_FNC_CHECK(hasGenerateConfigDescription,  GenerateConfigDescription,
                                              NJointController::GenerateConfigDescriptionFunctionSignature);
        ARMARX_META_MAKE_HAS_MEMBER_FNC_CHECK(hasGenerateConfigFromVariants,  GenerateConfigFromVariants,
                                              NJointController::GenerateConfigFromVariantsFunctionSignature<typename T::ConfigPtrT>);
#define ARMARX_ASSERT_NJOINTCONTROLLER_HAS_CONSTRUCTION_GUI(T)                                                \
    static_assert(::armarx::detail::hasGenerateConfigDescription<T>::value ,                                \
                  #T " does not offer a construction gui (missing: static GenerateConfigDescription)" );    \
    static_assert(::armarx::detail::hasGenerateConfigFromVariants<T>::value,                                \
                  #T " does not offer a construction gui (missing: static GenerateConfigFromVariants)")

        template<class NJointControllerT>
        class NJointControllerRegistryEntryHelper <
            NJointControllerT,
        //            meta::void_t<decltype(NJointControllerT::GenerateConfigDescription)>,
            typename std::enable_if < hasGenerateConfigDescription<NJointControllerT>::value >::type,
        //            meta::void_t<decltype(NJointControllerT::GenerateConfigFromVariants)
            typename std::enable_if < hasGenerateConfigFromVariants<NJointControllerT>::value >::type
            > : public NJointControllerRegistryEntry
        {
            NJointControllerPtr create(NJointControllerDescriptionProviderInterfacePtr prov,
                                       const NJointControllerConfigPtr& config,
                                       const boost::shared_ptr<VirtualRobot::Robot>& rob) const final override
            {
                ARMARX_CHECK_EXPRESSION_W_HINT(prov, "DataProvider is NULL!");
                using ConfigPtrT = typename NJointControllerT::ConfigPtrT;
                ConfigPtrT cfg = ConfigPtrT::dynamicCast(config);
                ARMARX_CHECK_EXPRESSION_W_HINT(cfg, "The configuration is of the wrong type! it has to be an instance of: "
                                               << GetTypeString<ConfigPtrT>());
                return new NJointControllerT(prov, cfg, rob);
            }
            WidgetDescription::WidgetPtr GenerateConfigDescription(
                const boost::shared_ptr<VirtualRobot::Robot>& robot,
                const std::map<std::string, ConstControlDevicePtr>& controlDevices,
                const std::map<std::string, ConstSensorDevicePtr>& sensorDevices
            ) const final override
            {
                try
                {
                    return NJointControllerT::GenerateConfigDescription(robot, controlDevices, sensorDevices);
                }
                catch (Ice::UserException& e)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigDescription'"
                                 << "\n---- file = " << e.ice_file()
                                 << "\n---- line = " << e.ice_line()
                                 << "\n---- name = " << e.ice_name()
                                 << "\n---- what:\n" << e.what()
                                 << "\n---- stacktrace:\n" << e.ice_stackTrace();
                    throw;
                }
                catch (std::exception& e)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigDescription'"
                                 << "\n---- what:\n" << e.what();
                    throw;
                }
                catch (...)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigDescription'";
                    throw;
                }
            }
            NJointControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap& variants) const override
            {
                try
                {
                    return NJointControllerT::GenerateConfigFromVariants(variants);
                }
                catch (Ice::UserException& e)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigFromVariants'"
                                 << "\n---- file = " << e.ice_file()
                                 << "\n---- line = " << e.ice_line()
                                 << "\n---- name = " << e.ice_name()
                                 << "\n---- what:\n" << e.what()
                                 << "\n---- stacktrace:\n" << e.ice_stackTrace();
                    throw;
                }
                catch (std::exception& e)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigFromVariants'"
                                 << "\n---- what:\n" << e.what();
                    throw;
                }
                catch (...)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigFromVariants'";
                    throw;
                }
            }
            bool hasRemoteConfiguration() const override
            {
                return true;
            }
        };
    }

    using NJointControllerRegistry = Registrar<std::unique_ptr<NJointControllerRegistryEntry>>;
    template<class ControllerType>
    struct NJointControllerRegistration
    {
        NJointControllerRegistration(const std::string& name)
        {
            NJointControllerRegistry::registerElement(
                name,
                std::unique_ptr<NJointControllerRegistryEntry>(new detail::NJointControllerRegistryEntryHelper<ControllerType>));
        }
    };
}
#endif
