/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::RobotUnit::RobotSynchronization
 * @author     Stefan Reither ( stef dot reither at web dot de )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "RobotSynchronization.h"
#include <RobotAPI/interface/units/RobotUnit/RobotUnitInterface.h>
#include "SensorValues/SensorValue1DoFActuator.h"

#include "RobotUnit.h"

using namespace armarx;

RobotSynchronization::RobotSynchronization(RobotUnit* robotUnit, VirtualRobot::RobotPtr& robot, const std::vector<std::string>& synchronizingJoints)
{
    this->robotUnit = robotUnit;
    this->robot = robot;
    Ice::StringSeq names = robotUnit->getSensorDeviceNames(GlobalIceCurrent);

    for (std::string name : names)
    {
        // skip, if the current sensor device does not correspond to a joint.
        if (!this->robot->hasRobotNode(name))
        {
            continue;
        }

        ConstSensorDevicePtr device = robotUnit->getSensorDevice(name);

        // skip, if the sensor value of the current sensor device can't provide a position.
        if (!device->getSensorValue()->isA<SensorValue1DoFActuatorPosition>())
        {
            continue;
        }

        std::vector<std::string>::const_iterator it = std::find(synchronizingJoints.begin(), synchronizingJoints.end(), name);
        if (it == synchronizingJoints.end())
        {
            continue;
        }
        else
        {
            jointSensorDeviceMap[*it] = device;
        }
    }
}

void RobotSynchronization::sync()
{
    if (robotUnit->getRobotUnitState() == RobotUnitState::InitializingUnits || robotUnit->getRobotUnitState() == RobotUnitState::Running)
    {
        for (auto const& entry : jointSensorDeviceMap)
        {
            const SensorValueBase* sv = entry.second->getSensorValue();

            ARMARX_CHECK_EXPRESSION(sv->asA<SensorValue1DoFActuatorPosition>());
            robot->getRobotNode(entry.first)->setJointValueNoUpdate(sv->asA<SensorValue1DoFActuatorPosition>()->position);
        }
        robot->applyJointValues();
    }
}
