/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotUnit
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "RobotUnit.h"

#include <algorithm>

#include <boost/algorithm/string/predicate.hpp>

#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/Nodes/ForceTorqueSensor.h>
#include <VirtualRobot/Robot.h>
#include <VirtualRobot/RobotNodeSet.h>
#include <VirtualRobot/Visualization/VisualizationNode.h>
#include <VirtualRobot/Obstacle.h>
#include <VirtualRobot/SceneObjectSet.h>
#include <VirtualRobot/Nodes/RobotNode.h>
#include <ArmarXCore/core/system/cmake/CMakePackageFinder.h>
#include <ArmarXCore/observers/variant/TimestampVariant.h>
#include <ArmarXCore/core/IceManager.h>
#include <ArmarXCore/core/IceGridAdmin.h>
#include <ArmarXCore/core/util/FileSystemPathBuilder.h>
#include <ArmarXCore/core/system/ArmarXDataPath.h>

#include "Units/ForceTorqueSubUnit.h"
#include "Units/InertialMeasurementSubUnit.h"
#include "Units/KinematicSubUnit.h"
#include "Units/PlatformSubUnit.h"
#include "SensorValues/SensorValue1DoFActuator.h"
#include "Units/TCPControllerSubUnit.h"
#include "Units/TrajectoryControllerSubUnit.h"

namespace std
{
    std::string to_string(armarx::RobotUnitState s)
    {
        switch (s)
        {
            case armarx::RobotUnitState::PreComponentInitialization:
                return "PreComponentInitialization";
            case armarx::RobotUnitState::InitializingDevices:
                return "InitializingDevices";
            case armarx::RobotUnitState::InitializingUnits:
                return "InitializingUnits";
            case armarx::RobotUnitState::WaitingForRTThreadInitialization:
                return "WaitingForRTThreadInitialization";
            case armarx::RobotUnitState::Running:
                return "Running";
            case armarx::RobotUnitState::Exiting:
                return "Exiting";
        }
        throw std::invalid_argument {"Unknown state " + std::to_string(static_cast<std::size_t>(s))};
    }
}

namespace armarx
{
    class RobotUnitEmergencyStopMaster :
        virtual public ManagedIceObject,
        virtual public EmergencyStopMasterInterface
    {
    public:
        RobotUnitEmergencyStopMaster(RobotUnit* robotUnit, std::string emergencyStopTopicName) :
            robotUnit {robotUnit},
                  emergencyStopTopicName {emergencyStopTopicName}
        {}

        void setEmergencyStopState(EmergencyStopState state, const Ice::Current& = GlobalIceCurrent) final override
        {
            robotUnit->setEmergencyStopState(state);
            emergencyStopTopic->reportEmergencyStopState(state);
        }
        EmergencyStopState getEmergencyStopState(const Ice::Current& = GlobalIceCurrent) const final override
        {
            return robotUnit->getEmergencyStopState();
        }

    protected:
        void onInitComponent() final override
        {
            armarx::ManagedIceObject::offeringTopic(emergencyStopTopicName);
        }
        void onConnectComponent() final override
        {
            emergencyStopTopic = armarx::ManagedIceObject::getTopic<EmergencyStopListenerPrx>(emergencyStopTopicName);
            setEmergencyStopState(EmergencyStopState::eEmergencyStopInactive);
        }

        std::string getDefaultName() const final override
        {
            return "EmergencyStopMaster";
        }

    protected:
        RobotUnit* const robotUnit;
        const std::string emergencyStopTopicName;
        EmergencyStopListenerPrx emergencyStopTopic;
    };

    namespace detail
    {
        std::function<std::string(const std::string&)> makeDebugObserverNameFixer(const std::string& prefix)
        {
            return
                [prefix](const std::string & name)
            {
                std::string s = prefix + name;
                std::replace(s.begin(), s.end(), '.', '_');
                std::replace(s.begin(), s.end(), ':', '_');
                return s;
            };
        }
    }

    const std::string armarx::RobotUnit::rtThreadTimingsSensorDeviceName = "RTThreadTimings";

    const std::set<RobotUnitState> armarx::RobotUnit::devicesReadyStates
    {
        RobotUnitState::InitializingUnits,
        RobotUnitState::WaitingForRTThreadInitialization,
        RobotUnitState::Running
    };
    armarx::RobotUnit::GuardType armarx::RobotUnit::getGuard() const
    {
        if (std::this_thread::get_id() == rtThreadId)
        {
            throw std::logic_error {"Attempted to lock mutex in RT thread"};
        }
        return GuardType {dataMutex};
    }

    void armarx::RobotUnit::checkNJointControllerClassName(const std::string& className) const
    {
        if (!NJointControllerRegistry::has(className))
        {
            std::stringstream ss;
            ss << "Requested controller class '" << className << "' unknown! Known classes:" << NJointControllerRegistry::getKeys()
               << " (If this class exists in a different lib then load it via loadLibFromPath(path) or loadLibFromPackage(package, lib))";
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
    }

    void armarx::RobotUnit::addControlDevice(const ControlDevicePtr& cd)
    {
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingDevices, __FUNCTION__);
        controlDevices.add(cd->getDeviceName(), cd);
        ARMARX_INFO << "added ControlDevice " << cd->getDeviceName();
    }

    void armarx::RobotUnit::addSensorDevice(const SensorDevicePtr& sd)
    {
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingDevices, __FUNCTION__);
        if (!sd)
        {
            std::stringstream ss;
            ss << "armarx::RobotUnit::addSensorDevice: SensorDevice is nullptr";
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
        if (!sd->getSensorValue())
        {
            std::stringstream ss;
            ss << "armarx::RobotUnit::addSensorDevice: SensorDevice " << sd->getDeviceName()
               << " has null SensorValue (this is not allowed)";
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
        if (sd->getDeviceName() == rtThreadTimingsSensorDeviceName)
        {
            if (!std::dynamic_pointer_cast<RTThreadTimingsSensorDevice>(sd))
            {
                throw InvalidArgumentException
                {
                    "You tried to add a SensorDevice with the name " + sd->getDeviceName() +
                    " which does not derive from RTThreadTimingsSensorDevice. (Don't do this)"
                };
            }
            //this checks if we already added such a device (do this before setting timingSensorDevice)
            sensorDevices.add(sd->getDeviceName(), sd);
            rtThreadTimingsSensorDevice = std::dynamic_pointer_cast<RTThreadTimingsSensorDevice>(sd);
        }
        else
        {
            sensorDevices.add(sd->getDeviceName(), sd);
        }
        ARMARX_INFO << "added SensorDevice " << sd->getDeviceName() << " (valuetype = " << sd->getSensorValueType() << ")";
    }

    void armarx::RobotUnit::finishDeviceInitialization()
    {
        auto beg = MicroNow();
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingDevices, __FUNCTION__);
        if (!sensorDevices.has(rtThreadTimingsSensorDeviceName))
        {
            addSensorDevice(createRTThreadTimingSensorDevice());
        }
        state = RobotUnitState::InitializingUnits;
        //check added devices
        ARMARX_DEBUG << "checking ControlDevices:";
        for (const ControlDevicePtr& controlDevice : controlDevices.values())
        {
            ARMARX_CHECK_EXPRESSION(controlDevice);
            ARMARX_DEBUG << "----" << controlDevice->getDeviceName();
            if (!controlDevice->hasJointController(ControlModes::EmergencyStop))
            {
                std::stringstream s;
                s << "ControlDevice " << controlDevice->getDeviceName()
                  << " has no JointController with ControlMode " << ControlModes::EmergencyStop
                  << " (to fix this, add a JointController with this ControlMode to the ControlDevice).\nAvailable controllers: "
                  <<  controlDevice->getControlModes();
                ARMARX_ERROR << "--------" << s.str();
                throw LogicError {s.str()};

            }
            if (!controlDevice->hasJointController(ControlModes::StopMovement))
            {
                std::stringstream s;
                s << "ControlDevice " << controlDevice->getDeviceName()
                  << " has no JointController with ControlMode \"" << ControlModes::StopMovement
                  << "\" (to fix this, add a JointController with this ControlMode to the ControlDevice) \nAvailable controllers: "
                  << controlDevice->getControlModes();
                ARMARX_ERROR << "--------" << s.str();
                throw LogicError {s.str()};
            }
        }
        ARMARX_DEBUG << "checking sensor devices:";
        for (const SensorDevicePtr& sensorDevice : sensorDevices.values())
        {
            ARMARX_CHECK_EXPRESSION(sensorDevice);
            ARMARX_DEBUG << "----" << sensorDevice->getDeviceName();
            if (!sensorDevice->getSensorValue())
            {
                std::stringstream s;
                s << "SensorDevice " << sensorDevice->getSensorValue() << " has null SensorValue";
                ARMARX_ERROR << "--------" << s.str();
                throw LogicError {s.str()};
            }
            ARMARX_CHECK_EXPRESSION(sensorDevice);
        }

        ARMARX_DEBUG << "initializing buffers:";
        {
            ARMARX_DEBUG << "----initializing controller buffers";
            controllersActivated.reinitAllBuffers(JointAndNJointControllers {getNumberOfControlDevices()});
            JointAndNJointControllers ctrlinitReq(getNumberOfControlDevices());
            for (std::size_t i = 0 ; i < getNumberOfControlDevices(); ++i)
            {
                ctrlinitReq.jointControllers.at(i) = controlDevices.at(i)->rtGetJointStopMovementController();
            }
            controllersRequested.reinitAllBuffers(ctrlinitReq);
            controllersRequested.commitWrite();//to make sure the diry bit is activated (to trigger rt switch)
        }
        ARMARX_DEBUG << "copying device ptrs to const ptr map";
        {
            for (const auto& dev : controlDevices.values())
            {
                controlDevicesConstPtr[dev->getDeviceName()] = dev;
            }
            for (const auto& dev : sensorDevices.values())
            {
                sensorDevicesConstPtr[dev->getDeviceName()] = dev;
                const SensorValue1DoFActuatorPosition* sv = dev->getSensorValue()->asA<SensorValue1DoFActuatorPosition>();
                if (sv && rtRobot->hasRobotNode(dev->getDeviceName()))
                {
                    nodePositionSensorMap[rtRobot->getRobotNode(dev->getDeviceName())] = sv;
                }
            }
        }

        ARMARX_VERBOSE << "ControlDevices:\n" << controlDevices.keys();
        ARMARX_VERBOSE << "SensorDevices:\n" << sensorDevices.keys();
        ctrlModeGroups.groupIndices.assign(getNumberOfControlDevices(), IndexSentinel());

        if (!ctrlModeGroups.groupsMerged.empty())
        {
            ARMARX_DEBUG << "Remove control devs from ControlDeviceHardwareControlModeGroups ("
                         << ctrlModeGroups.groups.size() << ")";
            const auto groupsMerged = ctrlModeGroups.groupsMerged;
            for (const auto& dev : groupsMerged)
            {
                if (controlDevices.has(dev))
                {
                    continue;
                }
                ctrlModeGroups.groupsMerged.erase(dev);
                for (auto& group : ctrlModeGroups.groups)
                {
                    group.erase(dev);
                }
                ARMARX_DEBUG << "----removing nonexistent device: " << dev;
            }
            //remove empty groups
            std::vector<std::set<std::string>> cleanedGroups;
            cleanedGroups.reserve(ctrlModeGroups.groups.size());
            for (auto& group : ctrlModeGroups.groups)
            {
                const auto sz = group.size();
                if (sz == 1)
                {
                    ARMARX_DEBUG << "----removing group with one dev: " << *group.begin();
                }
                else if (sz > 1)
                {
                    cleanedGroups.emplace_back(std::move(group));
                }
            }
            ctrlModeGroups.groups = cleanedGroups;
            ARMARX_DEBUG << "----number of groups left: " << ctrlModeGroups.groups.size();
        }
        if (!ctrlModeGroups.groupsMerged.empty())
        {
            ARMARX_DEBUG << "Checking control modes for ControlDeviceHardwareControlModeGroups ("
                         << ctrlModeGroups.groups.size() << ")";
            ctrlModeGroups.deviceIndices.resize(ctrlModeGroups.groups.size());
            //iterate over groups
            for (std::size_t groupIdx = 0; groupIdx < ctrlModeGroups.groups.size(); ++groupIdx)
            {
                const std::set<std::string>& group = ctrlModeGroups.groups.at(groupIdx);
                ctrlModeGroups.deviceIndices.at(groupIdx).reserve(group.size());
                ARMARX_CHECK_EXPRESSION(!group.empty());
                ARMARX_DEBUG << "----Group " << groupIdx << " size: " << group.size();
                //gets a map of ControlMode->HardwareControlMode for the given device
                const auto getControlModeToHWControlMode = [&](const std::string & devname)
                {
                    std::map<std::string, std::string> controlModeToHWControlMode;
                    const ControlDevicePtr& cd = controlDevices.at(devname);
                    for (const auto& jointCtrl : cd->getJointControllers())
                    {
                        controlModeToHWControlMode[jointCtrl->getControlMode()] = jointCtrl->getHardwareControlMode();
                    }
                    return controlModeToHWControlMode;
                };
                //get modes of first dev
                const auto controlModeToHWControlMode = getControlModeToHWControlMode(*group.begin());
                //check other devs
                for (const auto& devname : group)
                {
                    ARMARX_CHECK_EXPRESSION_W_HINT(
                        controlDevices.has(devname),
                        "The ControlDeviceHardwareControlModeGroups property contains device names not existent in the robot: "
                        << devname << "\navailable:\n" << controlDevices.keys()
                    );
                    //Assert all devices in a group have the same control modes with the same hw controle modes
                    const auto controlModeToHWControlModeForDevice = getControlModeToHWControlMode(devname);
                    ARMARX_CHECK_EXPRESSION_W_HINT(
                        controlModeToHWControlModeForDevice == controlModeToHWControlMode,
                        "Error for control modes of device '" << devname << "'\n"
                        << "it has the modes: " << controlModeToHWControlModeForDevice
                        << "\n but should have the modes: " << controlModeToHWControlMode
                    );
                    //insert the device index into the device indices
                    const auto devIdx = controlDevices.index(devname);
                    ctrlModeGroups.deviceIndices.at(groupIdx).emplace_back(devIdx);
                    //insert the group index into the group indices (+ check the current group index is the sentinel
                    ARMARX_CHECK_EXPRESSION(ctrlModeGroups.groupIndices.size() > devIdx);
                    ARMARX_CHECK_EXPRESSION(ctrlModeGroups.groupIndices.at(devIdx) == IndexSentinel());
                    ctrlModeGroups.groupIndices.at(devIdx) = groupIdx;
                    ARMARX_DEBUG << "------- " << devname;
                }
            }
        }
        ARMARX_INFO << "finishDeviceInitialization...done! " << (MicroNow() - beg).count() << " us";
    }

    ManagedIceObjectPtr RobotUnit::createKinematicSubUnit(const Ice::PropertiesPtr& properties, const std::string& positionControlMode,
            const std::string& velocityControlMode, const std::string& torqueControlMode)
    {
        using UnitT = KinematicSubUnit;



        //add ctrl et al
        std::map<std::string, UnitT::ActuatorData> devs;
        for (const ControlDevicePtr& controlDevice : controlDevices.values())
        {
            ARMARX_CHECK_EXPRESSION(controlDevice);
            const auto& controlDeviceName = controlDevice->getDeviceName();
            if (!controlDevice->hasTag(ControlDeviceTags::CreateNoDefaultController))
            {
                UnitT::ActuatorData ad;
                ad.name = controlDeviceName;
                ad.sensorDeviceIndex =
                    sensorDevices.has(controlDeviceName) ?
                    sensorDevices.index(controlDeviceName) :
                    std::numeric_limits<std::size_t>::max();
                /// TODO use better kinematic unit controllers (posThroughVel + ramps)
#define initializeKinematicUnit_tmp_helper(mode, CtargT, ctrl)                                                          \
    {                                                                                                                   \
        const auto& controlMode = mode;                                                                                 \
        NJointKinematicUnitPassThroughControllerPtr& pt = ctrl;                                                           \
        JointController* jointCtrl = controlDevice->getJointController(controlMode);                                           \
        if (jointCtrl && jointCtrl->getControlTarget()->isA<CtargT>())                                                            \
        {                                                                                                               \
            NJointKinematicUnitPassThroughControllerConfigPtr config = new NJointKinematicUnitPassThroughControllerConfig;  \
            config->controlMode=controlMode;                                                                            \
            config->deviceName=controlDeviceName;                                                                       \
            const NJointControllerPtr& nJointCtrl = createNJointController(                                                       \
                                                    "NJointKinematicUnitPassThroughController",                                   \
                                                    "NJointKU_PTCtrl_"+controlDeviceName+"_"+controlMode,                         \
                                                    config,                                                                     \
                                                    false, true);                                                               \
            pt = NJointKinematicUnitPassThroughControllerPtr::dynamicCast(nJointCtrl);                                          \
            ARMARX_CHECK_EXPRESSION(pt);                                                                                \
        }                                                                                                               \
    }
                initializeKinematicUnit_tmp_helper(positionControlMode, ControlTarget1DoFActuatorPosition, ad.ctrlPos)
                initializeKinematicUnit_tmp_helper(velocityControlMode, ControlTarget1DoFActuatorVelocity, ad.ctrlVel)
                initializeKinematicUnit_tmp_helper(torqueControlMode, ControlTarget1DoFActuatorTorque, ad.ctrlTor)
#undef initializeKinematicUnit_tmp_helper


                if (ad.ctrlPos || ad.ctrlTor || ad.ctrlVel)
                {
                    devs[controlDeviceName] = std::move(ad);
                }
            }
        }
        if (devs.empty())
        {
            ARMARX_IMPORTANT << "No joint devices found - skipping adding of KinematicUnit";
            return NULL;
        }
        ARMARX_CHECK_EXPRESSION_W_HINT(!devs.empty(), "no 1DoF ControlDevice found with matching SensorDevice");
        //add it

        const std::string configName = getProperty<std::string>("KinematicUnitName");
        const std::string confPre = getConfigDomain() + "." + configName + ".";
        //properties->setProperty(confPre + "MinimumLoggingLevel", getProperty<std::string>("MinimumLoggingLevel").getValue());
        //fill properties
        properties->setProperty(confPre + "RobotNodeSetName", robotNodeSetName);
        properties->setProperty(confPre + "RobotFileName", robotFileName);
        properties->setProperty(confPre + "RobotFileNameProject", robotProjectName);
        ARMARX_DEBUG << "creating unit " << configName << " using these properties: " << properties->getPropertiesForPrefix("");
        IceInternal::Handle<UnitT> unit = Component::create<UnitT>(properties, configName, getConfigDomain());

        //fill devices (sensor + controller)
        unit->setupData(
            getProperty<std::string>("RobotFileName").getValue(),
            robot->clone(),
            std::move(devs),
            ctrlModeGroups.groups,
            ctrlModeGroups.groupsMerged,
            this
        );
        return unit;
    }

    void armarx::RobotUnit::initializeDefaultUnits()
    {
        auto beg = MicroNow();
        {
            auto guard = getGuard();
            throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
            ARMARX_INFO << "initializing default units";
            initializeKinematicUnit();
            ARMARX_DEBUG << "KinematicUnit initialized";
            initializePlatformUnit();
            ARMARX_DEBUG << "PlatformUnit initialized";
            initializeForceTorqueUnit();
            ARMARX_DEBUG << "ForceTorqueUnit initialized";
            initializeInertialMeasurementUnit();
            ARMARX_DEBUG << "InertialMeasurementUnit initialized";
            initializeTrajectoryControllerUnit();
            ARMARX_DEBUG << "TrajectoryControllerUnit initialized";
            initializeTcpControllerUnit();
            ARMARX_DEBUG << "TcpControllerUnit initialized";
        }
        ARMARX_INFO << "initializing default units...done! " << (MicroNow() - beg).count() << " us";
    }

    void armarx::RobotUnit::initializeKinematicUnit()
    {
        using IfaceT = KinematicUnitInterface;

        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        //check if unit is already added
        if (getUnit(IfaceT::ice_staticId()))
        {
            return;
        }

        auto unit = createKinematicSubUnit(getIceProperties()->clone());
        //add
        if (unit)
        {
            addUnit(std::move(unit));
        }
    }

    void armarx::RobotUnit::initializePlatformUnit()
    {
        using UnitT = PlatformSubUnit;
        using IfaceT = PlatformUnitInterface;

        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        //check if unit is already added
        if (getUnit(IfaceT::ice_staticId()))
        {
            return;
        }
        //is there a platform dev?
        if (getRobotPlatformName().empty())
        {
            ARMARX_INFO << "no platform unit created since no platform name was given";
            return;
        }
        if (!controlDevices.has(getRobotPlatformName()))
        {
            ARMARX_WARNING << "no platform unit created since the platform control device with name '"
                           << getRobotPlatformName() << "' was not found";
            return;
        }
        const ControlDevicePtr& controlDevice = controlDevices.at(robotPlatformName);
        const SensorDevicePtr&  sensorDevice = sensorDevices.at(robotPlatformName);
        JointController* jointVel = controlDevice->getJointController(ControlModes::HolonomicPlatformVelocity);
        ARMARX_CHECK_EXPRESSION(jointVel);
        ARMARX_CHECK_EXPRESSION(sensorDevice->getSensorValue()->asA<SensorValueHolonomicPlatform>());
        //add it
        const std::string configName = getProperty<std::string>("PlatformUnitName");
        const std::string confPre = getConfigDomain() + "." + configName + ".";
        Ice::PropertiesPtr properties = getIceProperties()->clone();
        //properties->setProperty(confPre + "MinimumLoggingLevel", getProperty<std::string>("MinimumLoggingLevel").getValue());
        //fill properties
        properties->setProperty(confPre + "PlatformName", robotPlatformName);
        ARMARX_DEBUG << "creating unit " << configName << " using these properties: " << properties->getPropertiesForPrefix("");
        IceInternal::Handle<UnitT> unit = Component::create<UnitT>(properties, configName, getConfigDomain());
        //config
        NJointHolonomicPlatformUnitVelocityPassThroughControllerConfigPtr config = new NJointHolonomicPlatformUnitVelocityPassThroughControllerConfig;
        config->initialVelocityX = 0;
        config->initialVelocityY = 0;
        config->initialVelocityRotation = 0;
        config->platformName = robotPlatformName;
        auto ctrl = NJointHolonomicPlatformUnitVelocityPassThroughControllerPtr::dynamicCast(
                        createNJointController(
                            "NJointHolonomicPlatformUnitVelocityPassThroughController", configName + "_VelPTContoller",
                            config, false, true
                        )
                    );
        ARMARX_CHECK_EXPRESSION(ctrl);
        unit->pt = ctrl;

        NJointHolonomicPlatformRelativePositionControllerConfigPtr configRelativePositionCtrlCfg = new NJointHolonomicPlatformRelativePositionControllerConfig;
        configRelativePositionCtrlCfg->platformName = robotPlatformName;
        auto ctrlRelativePosition = NJointHolonomicPlatformRelativePositionControllerPtr::dynamicCast(
                                        createNJointController(
                                            "NJointHolonomicPlatformRelativePositionController", configName + "_RelativePositionContoller",
                                            configRelativePositionCtrlCfg, false, true
                                        )
                                    );
        ARMARX_CHECK_EXPRESSION(ctrlRelativePosition);
        unit->pt = ctrl;
        unit->relativePosCtrl = ctrlRelativePosition;

        unit->platformSensorIndex = sensorDevices.index(robotPlatformName);
        //add
        addUnit(std::move(unit));
    }

    void armarx::RobotUnit::initializeForceTorqueUnit()
    {
        using UnitT = ForceTorqueSubUnit;
        using IfaceT = ForceTorqueUnitInterface;

        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        //check if unit is already added
        if (getUnit(IfaceT::ice_staticId()))
        {
            return;
        }
        //check if it is required
        std::vector<ForceTorqueSubUnit::DeviceData> ftdevs;
        for (const SensorDevicePtr& ftSensorDevice : sensorDevices.values())
        {
            if (ftSensorDevice->getSensorValue()->isA<SensorValueForceTorque>())
            {
                ForceTorqueSubUnit::DeviceData ftSensorData;
                ftSensorData.sensorIndex = sensorDevices.index(ftSensorDevice->getDeviceName());
                ftSensorData.deviceName = ftSensorDevice->getDeviceName();
                ftSensorData.frame = ftSensorDevice->getReportingFrame();
                ARMARX_CHECK_EXPRESSION_W_HINT(
                    !ftSensorData.frame.empty(),
                    "The sensor device '" << ftSensorData.deviceName << "' (index: " << ftSensorData.sensorIndex
                    << ") reports force torque values but returns an empty string as reporting frame"
                );
                ARMARX_CHECK_EXPRESSION_W_HINT(
                    getRobot()->hasRobotNode(ftSensorData.frame),
                    "The sensor device '" << ftSensorData.deviceName << "' (index: " << ftSensorData.sensorIndex
                    << ") reports force torque values but returns a reporting frame '" << ftSensorData.frame
                    << "' which is not present in the robot '" << getRobot()->getName() << "'"
                );
                ftdevs.emplace_back(std::move(ftSensorData));
            }
        }
        if (ftdevs.empty())
        {
            ARMARX_IMPORTANT << "no force torque unit created since there are no force torque sensor devices";
            return;
        }
        //add it
        const std::string configName = getProperty<std::string>("ForceTorqueUnitName");
        const std::string confPre = getConfigDomain() + "." + configName + ".";
        Ice::PropertiesPtr properties = getIceProperties()->clone();
        //properties->setProperty(confPre + "MinimumLoggingLevel", getProperty<std::string>("MinimumLoggingLevel").getValue());
        //fill properties
        properties->setProperty(confPre + "AgentName", robot->getName());
        properties->setProperty(confPre + "ForceTorqueTopicName", getProperty<std::string>("ForceTorqueTopicName").getValue());
        ARMARX_DEBUG << "creating unit " << configName << " using these properties: " << properties->getPropertiesForPrefix("");
        IceInternal::Handle<UnitT> unit = Component::create<UnitT>(properties, configName, getConfigDomain());
        unit->devs = ftdevs;
        //add
        addUnit(std::move(unit));
    }

    void RobotUnit::initializeTrajectoryControllerUnit()
    {
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        using UnitT = TrajectoryControllerSubUnit;
        if (!getProperty<bool>("CreateTrajectoryPlayer").getValue())
        {
            return;
        }

        //check if unit is already added
        if (getUnit(UnitT::ice_staticId()) || !trajectoryControllerSubUnit)
        {
            return;
        }
        (TrajectoryControllerSubUnitPtr::dynamicCast(trajectoryControllerSubUnit))->setup(this);
        addUnit(trajectoryControllerSubUnit);
    }

    void RobotUnit::initializeTcpControllerUnit()
    {
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);

        if (!getProperty<bool>("CreateTCPControlUnit").getValue())
        {
            return;
        }
        using UnitT = TCPControllerSubUnit;

        //check if unit is already added
        if (getUnit(UnitT::ice_staticId()))
        {
            return;
        }
        (TCPControllerSubUnitPtr::dynamicCast(tcpControllerSubUnit))->setup(this, robot->clone());
        addUnit(tcpControllerSubUnit);
    }
    void armarx::RobotUnit::initializeInertialMeasurementUnit()
    {
        using UnitT = InertialMeasurementSubUnit;
        using IfaceT = InertialMeasurementUnitInterface;

        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        //check if unit is already added
        if (getUnit(IfaceT::ice_staticId()))
        {
            return;
        }
        //check if it is required
        std::map<std::string, std::size_t> imudevs;
        for (auto i : getIndices(sensorDevices.keys()))
        {
            const SensorDevicePtr& sensorDevice = sensorDevices.at(i);
            if (sensorDevice->getSensorValue()->isA<SensorValueIMU>())
            {
                imudevs[sensorDevice->getDeviceName()] = i;
            }
        }
        if (imudevs.empty())
        {
            ARMARX_IMPORTANT << "no inertial measurement unit created since there are no imu sensor devices";
            return;
        }
        //add it
        const std::string configName = getProperty<std::string>("InertialMeasurementUnitName");
        const std::string confPre = getConfigDomain() + "." + configName + ".";
        Ice::PropertiesPtr properties = getIceProperties()->clone();
        //properties->setProperty(confPre + "MinimumLoggingLevel", getProperty<std::string>("MinimumLoggingLevel").getValue());
        //fill properties
        properties->setProperty(confPre + "IMUTopicName", getProperty<std::string>("IMUTopicName").getValue());
        IceInternal::Handle<UnitT> unit = Component::create<UnitT>(properties, configName, getConfigDomain());
        unit->devs = imudevs;
        //add
        addUnit(std::move(unit));
    }

    void armarx::RobotUnit::addUnit(ManagedIceObjectPtr unit)
    {
        ARMARX_CHECK_EXPRESSION(unit);
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        getArmarXManager()->addObjectAsync(unit, "", true, false);
        //maybe add it to the sub units
        RobotUnitSubUnitPtr rsu = RobotUnitSubUnitPtr::dynamicCast(unit);
        if (rsu)
        {
            subUnits.emplace_back(std::move(rsu));
        }
        units.emplace_back(std::move(unit));
    }

    void armarx::RobotUnit::finishUnitInitialization()
    {

        auto beg = MicroNow();
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        state = RobotUnitState::WaitingForRTThreadInitialization;
        //start publisher
        std::size_t pubtimestep = getProperty<std::size_t>("PublishPeriodMs");
        if (!pubtimestep)
        {
            pubtimestep = 1;
        }
        publishNewSensorDataTime = TimeUtil::GetTime();
        publisherTask = new PublisherTaskT([&] {publish({});}, pubtimestep, false, getName() + "_PublisherTask");
        ARMARX_INFO << "starting publisher with timestep " << pubtimestep;
        publisherTask->setDelayWarningTolerance(100);
        publisherTask->start();

        std::size_t rtLoggingTimestep = getProperty<std::size_t>("RTLoggingPeriodMs");
        rtLogging.enabled = rtLoggingTimestep;
        rtLogging.loggingTaskTimestep = rtLoggingTimestep;
        ARMARX_CHECK_GREATER(getControlThreadTargetPeriod().toMicroSeconds(), 0);
        if (isRtLoggingEnabled())
        {
            //init buffer
            {
                std::size_t messageBufferSize = getProperty<std::size_t>("RTLoggingMessageBufferSize");
                std::size_t messageBufferNumberEntries = getProperty<std::size_t>("RTLoggingMessageNumber");
                std::size_t messageBufferMaxSize = getProperty<std::size_t>("RTLoggingMaxMessageBufferSize");
                std::size_t messageBufferMaxNumberEntries = getProperty<std::size_t>("RTLoggingMaxMessageNumber");

                rtLogging.keepBacklogFor = IceUtil::Time::milliSeconds(getProperty<std::size_t>("RTLoggingKeepIterationsForMs"));

                ARMARX_CHECK_LESS_EQUAL(messageBufferSize, messageBufferMaxSize);
                ARMARX_CHECK_LESS_EQUAL(messageBufferNumberEntries, messageBufferMaxNumberEntries);

                std::size_t cThreadPeriodUs = static_cast<std::size_t>(getControlThreadTargetPeriod().toMicroSeconds());
                std::size_t lThreadPeriodUs = rtLoggingTimestep * 1000;

                std::size_t nBuffers = (lThreadPeriodUs / cThreadPeriodUs + 1) * 100;
                const auto bufferSize = controlThreadOutputBuffer.initialize(
                                            nBuffers, controlDevices, sensorDevices,
                                            messageBufferSize, messageBufferNumberEntries,
                                            messageBufferMaxSize, messageBufferMaxNumberEntries);
                ARMARX_INFO << "RTLogging activated! using "
                            << nBuffers << "buffers "
                            << "(buffersize = " << bufferSize << " bytes)";
            }
            //init logging names
            {
                //sensorDevices
                rtLogging.controlDeviceValueLoggingNames.reserve(controlDevices.size());
                for (const auto& cd : controlDevices.values())
                {
                    rtLogging.controlDeviceValueLoggingNames.emplace_back();
                    auto& namesForDev = rtLogging.controlDeviceValueLoggingNames.back();
                    namesForDev.reserve(cd->getJointControllers().size());
                    for (auto jointC : cd->getJointControllers())
                    {
                        const auto names = jointC->getControlTarget()->getDataFieldNames();
                        namesForDev.emplace_back();
                        auto& fullNames = namesForDev.back();
                        fullNames.reserve(names.size());
                        for (const auto& name :  names)
                        {
                            fullNames.emplace_back(
                                "ctrl." +
                                cd->getDeviceName() + "." +
                                jointC->getControlMode() + "." +
                                name);
                        }
                    }
                }
                //sensorDevices
                rtLogging.sensorDeviceValueLoggingNames.reserve(sensorDevices.size());
                for (const auto& sd : sensorDevices.values())
                {
                    const auto names = sd->getSensorValue()->getDataFieldNames();
                    rtLogging.sensorDeviceValueLoggingNames.emplace_back();
                    auto& fullNames = rtLogging.sensorDeviceValueLoggingNames.back();
                    fullNames.reserve(names.size());
                    for (const auto& name :  names)
                    {
                        fullNames.emplace_back(
                            "sens." +
                            sd->getDeviceName() + "." +
                            name);
                    }
                }
            }
            //start logging thread is done in rtinit
            //maybe add the default log
            {
                const auto loggingpath = getProperty<std::string>("RTLoggingDefaultLog").getValue();
                if (!loggingpath.empty())
                {
                    rtLogging.defaultLogHandle = startRtLogging(loggingpath, getLoggingNames());
                }
            }
        }
        else
        {
            //no rt logging
            ARMARX_INFO << "RTLogging deactivated!";
            controlThreadOutputBuffer.initialize(0, controlDevices, sensorDevices, 0, 0, 0, 0);
        }

        startSelfCollisionAvoidance();
        if (isRtLoggingEnabled())
        {
            rtLogging.task = new RTLoggingTaskT([&] {rtLogging.doLogging();}, rtLogging.loggingTaskTimestep, false, getName() + "_RTLoggingTask");
            ARMARX_INFO << "starting rt logging with timestep " << rtLogging.loggingTaskTimestep;
            rtLogging.task->setDelayWarningTolerance(rtLogging.loggingTaskTimestep * 10);
            rtLogging.task->start();
        }

        ARMARX_INFO << "finishUnitInitialization...done! " << (MicroNow() - beg).count() << " us";
    }

    void armarx::RobotUnit::initRTThread()
    {
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::WaitingForRTThreadInitialization, __FUNCTION__);
        rtThreadId = std::this_thread::get_id();
        rtLogging.rtThreadId = LogSender::getThreadId();
        state = RobotUnitState::Running;
        ControlThreadOutputBuffer::RtLoggingInstance = &controlThreadOutputBuffer;
    }

    bool armarx::RobotUnit::rtSwitchControllerSetup()
    {
        rtThreadTimingsSensorDevice->rtMarkRtSwitchControllerSetupStart();
        ARMARX_ON_SCOPE_EXIT { rtThreadTimingsSensorDevice->rtMarkRtSwitchControllerSetupEnd(); };
        // !emergencyStop && !rtIsInEmergencyStop -> normal control flow
        // !emergencyStop &&  rtIsInEmergencyStop -> force switch to reactivate old ( + reset flag)
        //  emergencyStop && !rtIsInEmergencyStop -> deactivate all + set flag
        //  emergencyStop &&  rtIsInEmergencyStop -> nothing to do

        if (emergencyStop)
        {
            if (rtIsInEmergencyStop)
            {
                return false;
            }
            rtIsInEmergencyStop = true;
            bool updatedActive = false;
            //deactivate all nJointCtrl
            for (auto& nJointCtrl : rtGetActivatedNJointControllers())
            {
                if (nJointCtrl)
                {
                    RtSetNJointControllerDeactivatedStatus(nJointCtrl);
                    nJointCtrl = nullptr;
                    updatedActive = true;
                }
            }
            //set all JointCtrl to emergency stop (except stop movement)
            for (std::size_t i = 0; i < rtGetNumberOfControlDevices(); ++i)
            {
                ControlDevicePtr& controlDevice = controlDevices.at(i);
                const auto active = controlDevice->rtGetActiveJointController();
                const auto stopMov = controlDevice->rtGetJointStopMovementController();
                const auto emergency = controlDevice->rtGetJointEmergencyStopController();
                if (active != stopMov && active != emergency)
                {
                    controlDevice->rtSetActiveJointController(emergency);
                    rtGetActivatedJointControllers().at(i) = emergency;
                    updatedActive = true;
                }
            }
            if (updatedActive)
            {
                rtGetActivatedControllers().resetAssignement();
                rtWriteActivatedControllers();
            }
            return updatedActive;
        }

        if (!rtRequestedControllersChanged() && !rtIsInEmergencyStop)
        {
            return false;
        }
        rtIsInEmergencyStop = false;

        //handle nJointCtrl
        {
            //"merge"
            std::size_t n = rtGetNumberOfControlDevices();
            std::size_t idxAct = 0;
            std::size_t idxReq = 0;
            for (std::size_t i = 0; i < 2 * n; ++i)
            {
                //skip nullptrs in act
                while (idxAct < n && !rtGetActivatedNJointControllers().at(idxAct))
                {
                    ++idxAct;
                }
                const NJointControllerPtr& req = idxReq < n ? rtGetRequestedNJointControllers().at(idxReq) : nullptr; //may be null
                const NJointControllerPtr& act = idxAct < n ? rtGetActivatedNJointControllers().at(idxAct) : nullptr; //may be null if it is the last
                std::size_t reqId = req ? req->rtGetId() : std::numeric_limits<std::size_t>::max();
                std::size_t actId = act ? act->rtGetId() : std::numeric_limits<std::size_t>::max();

                if (reqId < actId)
                {
                    //new ctrl
                    RtSetNJointControllerActivatedStatus(req);
                    ++idxReq;
                }
                else if (reqId > actId)
                {
                    RtSetNJointControllerDeactivatedStatus(act);
                    ++idxAct;
                }
                else //if(reqId == actId)
                {
                    //same ctrl or both null ctrl
                    ++idxReq;
                    ++idxAct;
                }
                if (idxAct >= n && !req)
                {
                    break;
                }
            }
            rtGetActivatedNJointControllers() = rtGetRequestedNJointControllers();
        }

        //handle Joint Ctrl
        {
            for (std::size_t i = 0; i < rtGetNumberOfControlDevices(); ++i)
            {
                auto& controlDevice = controlDevices.at(i);
                const auto requestedJointCtrl = rtGetRequestedJointControllers().at(i);
                controlDevice->rtSetActiveJointController(requestedJointCtrl);
                rtGetActivatedJointControllers().at(i) = requestedJointCtrl;
            }
        }
        rtGetActivatedJointToNJointControllerAssignement() = rtGetRequestedControllers().jointToNJointControllerAssignement;
        rtWriteActivatedControllers();
        return true;
    }

    void armarx::RobotUnit::rtDeactivateAssignedNJointControllerBecauseOfError(std::size_t ctrlDevIndex, bool writeActivatedControlers)
    {
        std::size_t nJointCtrlIndex = rtGetActivatedJointToNJointControllerAssignement().at(ctrlDevIndex);
        ARMARX_CHECK_EXPRESSION_W_HINT(
            nJointCtrlIndex < rtGetNumberOfControlDevices(),
            nJointCtrlIndex << " < " << rtGetNumberOfControlDevices() <<
            ": no NJoint controller controls this device (name = "
            << controlDevices.at(ctrlDevIndex)->getDeviceName()
            << ", ControlMode = " << rtGetActivatedJointControllers().at(ctrlDevIndex)->getControlMode() << ")!" << "\n"
            << VAROUT(rtGetActivatedJointControllers()) << "\n"
            << VAROUT(rtGetActivatedNJointControllers())
        );

        rtDeactivateNJointControllerBecauseOfError(nJointCtrlIndex, false);

        if (writeActivatedControlers)
        {
            rtWriteActivatedControllers();
        }
    }

    void armarx::RobotUnit::rtDeactivateNJointControllerBecauseOfError(std::size_t nJointCtrlIndex, bool writeActivatedControlers)
    {
        const NJointControllerPtr& nJointCtrl = rtGetActivatedNJointControllers().at(nJointCtrlIndex);
        RtSetNJointControllerDeactivatedBecauseOfErrorStatus(nJointCtrl);
        for (auto ctrlDevIdx : nJointCtrl->rtGetControlDeviceUsedIndices())
        {
            const ControlDevicePtr& controlDevice = controlDevices.at(ctrlDevIdx);
            JointController* es = controlDevice->rtGetJointEmergencyStopController();
            controlDevice->rtSetActiveJointController(es);
            rtGetActivatedJointControllers().at(ctrlDevIdx) = es;
            ARMARX_CHECK_EXPRESSION(rtGetActivatedJointToNJointControllerAssignement().at(ctrlDevIdx) == nJointCtrlIndex);
            rtGetActivatedJointToNJointControllerAssignement().at(ctrlDevIdx) = IndexSentinel();
        }
        rtGetActivatedNJointControllers().at(nJointCtrlIndex) = nullptr;
        //check ControlDeviceHardwareControlModeGroups
        {
            for (auto ctrlDevIdx : nJointCtrl->rtGetControlDeviceUsedIndices())
            {
                const auto groupIdx = ctrlModeGroups.groupIndices.at(ctrlDevIdx);
                if (groupIdx == IndexSentinel())
                {
                    continue;
                }
                ControlDevice*  const dev  = controlDevices.at(ctrlDevIdx).get();
                JointController* const jointCtrl = dev->rtGetActiveJointController();
                const auto hwModeHash = jointCtrl->rtGetHardwareControlModeHash();
                //this device is in a group!
                // -> check all other devices
                for (const auto otherIdx : ctrlModeGroups.deviceIndices.at(groupIdx))
                {
                    ControlDevice*  const otherDev  = controlDevices.at(otherIdx).get();
                    JointController* const otherJointCtrl = otherDev->rtGetActiveJointController();
                    const auto otherHwModeHash = otherJointCtrl->rtGetHardwareControlModeHash();
                    if (hwModeHash == otherHwModeHash)
                    {
                        //the assigend ctrl has the same hwMode -> don't do anything
                        continue;
                    }
                    const auto otherNJointCtrl1Idx = rtGetActivatedJointToNJointControllerAssignement().at(otherIdx);
                    if (otherNJointCtrl1Idx == IndexSentinel())
                    {
                        //the hwmodes are different! (hence the other ctrl must be in stop movement
                        ARMARX_CHECK_EXPRESSION(otherJointCtrl == otherDev->rtGetJointStopMovementController());
                        //we need to activate the es contoller
                        JointController* const es = otherDev->rtGetJointEmergencyStopController();
                        otherDev->rtSetActiveJointController(es);
                        ARMARX_CHECK_EXPRESSION(es->rtGetHardwareControlModeHash() == hwModeHash);
                        rtGetActivatedJointControllers().at(otherIdx) = es;
                        continue;
                    }
                    //the devs NJoint controller needs to be deactivated
                    rtDeactivateAssignedNJointControllerBecauseOfError(otherIdx, false);
                }
            }
        }
        rtDeactivatedNJointControllerBecauseOfError(nJointCtrl);

        if (writeActivatedControlers)
        {
            rtWriteActivatedControllers();
        }
    }

    void armarx::RobotUnit::rtRunNJointControllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtThreadTimingsSensorDevice->rtMarkRtRunNJointControllersStart();
        for (std::size_t nJointCtrlIndex = 0; nJointCtrlIndex < rtGetActivatedNJointControllers().size(); ++nJointCtrlIndex)
        {
            NJointController* nJointCtrl = rtGetActivatedNJointControllers().at(nJointCtrlIndex);
            try
            {
                if (nJointCtrl)
                {
                    auto start = MicroNow();
                    nJointCtrl->rtSwapBufferAndRun(sensorValuesTimestamp, timeSinceLastIteration);
                    auto duration = MicroNow() - start;
                    if (checkControllerExecutionTimings)
                    {
                        if (duration.count() > 500)
                        {
                            ARMARX_ERROR << deactivateSpam(5) << "The NJointController " << nJointCtrl->getClassName() << " took " << duration.count() << " µs to run!";
                        }
                        else if (duration.count() > 50)
                        {
                            ARMARX_RT_LOGF_WARN("The NJointController with ID %d took %d µs to run!", nJointCtrl->getId(), duration.count()).deactivateSpam(5);
                        }
                    }
                }
            }
            catch (...)
            {
                ARMARX_ERROR << "NJoint Controller " << nJointCtrl->getInstanceName()
                             << " threw an exception and is now deactivated: "
                             << GetHandledExceptionString();
                rtDeactivateNJointControllerBecauseOfError(nJointCtrlIndex, false);
            }
        }
        rtWriteActivatedControllers();
        rtThreadTimingsSensorDevice->rtMarkRtRunNJointControllersEnd();
    }

    void armarx::RobotUnit::rtRunJointControllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtThreadTimingsSensorDevice->rtMarkRtRunJointControllersStart();
        for (const ControlDevicePtr& device : this->controlDevices.values())
        {
            device->rtRun(sensorValuesTimestamp, timeSinceLastIteration);
        }
        rtThreadTimingsSensorDevice->rtMarkRtRunJointControllersEnd();
    }

    void armarx::RobotUnit::rtReadSensorDeviceValues(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtThreadTimingsSensorDevice->rtMarkRtReadSensorDeviceValuesStart();
        for (const SensorDevicePtr& device : this->sensorDevices.values())
        {
            device->rtReadSensorValues(sensorValuesTimestamp, timeSinceLastIteration);
        }
        rtThreadTimingsSensorDevice->rtMarkRtReadSensorDeviceValuesEnd();
    }

    void armarx::RobotUnit::rtHandleInvalidTargets()
    {
        rtThreadTimingsSensorDevice->rtMarkRtHandleInvalidTargetsStart();
        bool oneIsInvalid = 0;
        for (std::size_t i = 0; i < rtGetNumberOfControlDevices(); ++i)
        {
            if (!rtGetActivatedJointControllers().at(i)->rtIsTargetVaild())
            {
                ARMARX_ERROR << "INVALID TARGET: " << controlDevices.at(i)->getDeviceName();
                rtDeactivateAssignedNJointControllerBecauseOfError(i, false);
                oneIsInvalid = true;
            }
        }
        if (oneIsInvalid)
        {
            rtWriteActivatedControllers();
        }
        rtThreadTimingsSensorDevice->rtMarkRtHandleInvalidTargetsEnd();
    }

    void armarx::RobotUnit::rtResetAllTargets()
    {
        rtThreadTimingsSensorDevice->rtMarkRtResetAllTargetsStart();
        for (const ControlDevicePtr& controlDev : controlDevices.values())
        {
            controlDev->rtGetActiveJointController()->rtResetTarget();
        }
        rtThreadTimingsSensorDevice->rtMarkRtResetAllTargetsEnd();
    }

    ConstSensorDevicePtr armarx::RobotUnit::getSensorDevice(const std::string& sensorDeviceName) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return sensorDevices.at(sensorDeviceName, SensorDevice::NullPtr);
    }

    ConstControlDevicePtr armarx::RobotUnit::getControlDevice(const std::string& deviceName) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controlDevices.at(deviceName, ControlDevice::NullPtr);
    }

    ControlTargetBase* armarx::RobotUnit::getControlTarget(const std::string& deviceName, const std::string& controlMode)
    {
        static const std::string fname = __FUNCTION__;
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (std::this_thread::get_id() != initNJointThreadId)
        {
            throw std::logic_error
            {
                fname + ": Must be only called during instantiation of a NJoint controller from the instantiated controler " +
                "(Don't try to call this function outside of a NJointController's  constructor)"
            };
        }
        if (initNJointControllerUsedJointController.find(deviceName) != initNJointControllerUsedJointController.end())
        {
            throw std::logic_error
            {
                fname + ": Must not request two ControlTargets for the same device. (got = " +
                initNJointControllerUsedJointController.at(deviceName)->getControlMode() + ", requested " + controlMode + ") " +
                "(You can only have a ControlDevice in one ControlMode. " +
                "To check all available ControlModes for this device, get the ControlDevice via getControlDevice(" + deviceName + ") and querry it)"
            };
        }
        // we should allow these modes for debugging reasons + there is no real reason to forbid them
        //if (controlMode == ControlModes::EmergencyStop)
        //{
        //    throw std::invalid_argument {"armarx::RobotUnit::getControlTarget: Can't get target for the control mode " + ControlModes::EmergencyStop};
        //}
        //if (controlMode == ControlModes::StopMovement)
        //{
        //    throw std::invalid_argument {"armarx::RobotUnit::getControlTarget: Can't get target for the control mode " + ControlModes::StopMovement};
        //}
        ControlDevicePtr controlDevice = controlDevices.at(deviceName, ControlDevice::NullPtr);
        if (!controlDevice)
        {
            return nullptr;
        }
        if (!controlDevice->hasJointController(controlMode))
        {
            return nullptr;
        }
        //there is a device and a target as was requested:
        JointController* jointCtrl = controlDevice->getJointController(controlMode);
        initNJointControllerUsedJointController[deviceName] = jointCtrl;
        return jointCtrl->getControlTarget();
    }

    NJointControllerInterfacePrx armarx::RobotUnit::createNJointControllerFromVariantConfig(
        const std::string& className, const std::string& instanceName, const StringVariantBaseMap& variants, const Ice::Current&)
    {
        //no lock required
        checkNJointControllerClassName(className);
        if (!NJointControllerRegistry::get(className)->hasRemoteConfiguration())
        {
            std::stringstream ss;
            ss << "Requested controller class '" << className << "' allows no remote configuration" << NJointControllerRegistry::getKeys()
               << " (Implement 'static WidgetDescription::WidgetPtr " << className << "::GenerateConfigDescription()'"
               << " and 'static NJointControllerConfigPtr " << className
               << "::GenerateConfigFromVariants(const StringVariantBaseMap&)' to allow remote configuration";
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
        return createNJointController(
                   className, instanceName,
                   NJointControllerRegistry::get(className)->GenerateConfigFromVariants(variants),
                   GlobalIceCurrent/*to select ice overload*/);
    }

    const NJointControllerPtr& armarx::RobotUnit::createNJointController(
        const std::string& className,
        const std::string& instanceName,
        const NJointControllerConfigPtr& config,
        bool deletable,
        bool internal)
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (instanceName.empty())
        {
            ARMARX_ERROR << "The instance name is empty! (give a unique name)";
            throw InvalidArgumentException {"The instance name is empty! (give a unique name)"};
        }
        //check if we would be able to create the class
        checkNJointControllerClassName(className);
        auto& factory = NJointControllerRegistry::get(className);

        //check if the instance name is already in use
        if (nJointControllers.count(instanceName))
        {
            std::stringstream ss;
            ss << "There already is a controller instance with the name '" << instanceName << "'. Use a different instance name instead."
               << " Other used instance names are " << getNJointControllerNames();
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }

        //create the controller
        ARMARX_CHECK_EXPRESSION(initNJointControllerUsedJointController.empty());//in case someone broke things
        initNJointThreadId = std::this_thread::get_id();
        ARMARX_ON_SCOPE_EXIT
        {
            initNJointControllerUsedJointController.clear();
            initNJointThreadId = std::thread::id{};
        };
        ARMARX_CHECK_EXPRESSION(factory);
        NJointControllerPtr nJointCtrl {factory->create(NJointControllerDescriptionProviderInterfacePtr::dynamicCast(RobotUnitPtr {this}), config, robot)};

        std::vector<char> ctrlDeviceUsedBitmap(getNumberOfControlDevices(), false);
        std::vector<std::size_t> ctrlDeviceUsedIndices;
        ctrlDeviceUsedIndices.reserve(getNumberOfControlDevices());
        for (const auto& cd2targ : initNJointControllerUsedJointController)
        {
            auto idx = controlDevices.index(cd2targ.first);
            ctrlDeviceUsedBitmap.at(idx) = true;
            ctrlDeviceUsedIndices.emplace_back(idx);
        }
        InitNJointControler(
            nJointCtrl,
            this,
            nJointControllerNextId++,
            initNJointControllerUsedJointController,
            std::move(ctrlDeviceUsedBitmap),
            std::move(ctrlDeviceUsedIndices),
            deletable, internal);
        getArmarXManager()->addObject(nJointCtrl, instanceName, false, false);
        nJointControllers[instanceName] = std::move(nJointCtrl);
        ARMARX_CHECK_EXPRESSION(listenerPrx);
        listenerPrx->nJointControllerCreated(instanceName);
        return nJointControllers.at(instanceName);
    }

    void armarx::RobotUnit::setActivateControllersRequest(std::set<NJointControllerPtr, NJointController::IdComp>&& ctrls)
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        //erase nullptr
        ctrls.erase(nullptr);
        ARMARX_CHECK_EXPRESSION(ctrls.size() <= getNumberOfControlDevices());
        const std::set<NJointControllerPtr, NJointController::IdComp> setOfRequestedCtrls
        {
            getRequestedNJointControllers().begin(),
            getRequestedNJointControllers().end()
        };
        if (setOfRequestedCtrls == ctrls)
        {
            //redirty the flag to swap in the same set again
            controllersRequested.commitWrite();
            return;
        }
        JointAndNJointControllers request {getNumberOfControlDevices()};
        std::size_t idx = 0;
        for (const NJointControllerPtr& nJointCtrl : ctrls)
        {
            request.nJointControllers.at(idx++) = nJointCtrl.get();
            //add Joint for this ctrl
            for (const auto& cd2cm : nJointCtrl->getControlDeviceUsedControlModeMap())
            {
                std::size_t jointCtrlIndex = controlDevices.index(cd2cm.first);
                if (request.jointControllers.at(jointCtrlIndex))
                {
                    std::stringstream ss;
                    ss << "RobotUnit:setActivateControllersRequest controllers to activate are in conflict!\ncontrollers:";
                    for (const auto& ctrl : ctrls)
                    {
                        ss << "\n" << ctrl->getInstanceName();
                    }
                    ARMARX_ERROR << ss.str();
                    throw InvalidArgumentException {ss.str()};
                }
                request.jointControllers.at(jointCtrlIndex) = controlDevices.at(jointCtrlIndex)->getJointController(cd2cm.second);
            }
        }
        for (auto i : getIndices(request.jointControllers))
        {
            JointController*& jointCtrl = request.jointControllers.at(i);
            if (!jointCtrl)
            {
                JointController* jointCtrlOld = getRequestedJointControllers().at(i);
                if (jointCtrlOld == controlDevices.at(i)->getJointStopMovementController())
                {
                    //don't replace this controller with emergency stop
                    jointCtrl = jointCtrlOld;
                }
                else
                {
                    //no one controls this device -> emergency stop
                    jointCtrl = controlDevices.at(i)->getJointEmergencyStopController();
                }
            }
        }
        writeRequestedControllers(std::move(request));
        ARMARX_INFO << "requested controllers:\n" << getRequestedNJointControllerNames();
    }

    void armarx::RobotUnit::deleteNJointControllers(const Ice::StringSeq& names, const Ice::Current&)
    {
        ARMARX_VERBOSE << "requesting controller deletion for:\n" << names;
        if (names.empty())
        {
            return;
        }
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        auto ctrlsToActVec = getNJointControllersNotNull(names); //also checks if these controllers exist
        ARMARX_DEBUG << "all controllers requested to delete exist" << std::flush;
        //check if all can be deleted
        for (const auto& nJointCtrl : ctrlsToActVec)
        {
            if (!nJointCtrl->isDeletable())
            {
                throw LogicError
                {
                    "The NJointController '" + nJointCtrl->getInstanceName() +
                    "' can't be deleted since this operation is not allowed for this controller! (no NJointController was deleted)"
                };
            }
        }
        for (const auto& nJointCtrl : ctrlsToActVec)
        {
            if (nJointCtrl->isControllerActive() || nJointCtrl->isControllerRequested())
            {
                throw LogicError
                {
                    "The NJointController '" + nJointCtrl->getInstanceName() +
                    "' can't be deleted since it is active or requested! (no NJointController was deleted)"
                };
            }
        }
        for (const auto& nJointCtrl : ctrlsToActVec)
        {
            const auto name = nJointCtrl->getInstanceName();
            //deletion is done in a different thread since this call may be done by the controller (protection against use after free)
            nJointControllersToBeDeleted[name] = std::move(nJointCtrl);
            nJointControllers.erase(name);
            ARMARX_VERBOSE << "added NJointController '" << name << "' to be deleted";
        }
    }

    void armarx::RobotUnit::activateNJointControllers(const Ice::StringSeq& names, const Ice::Current&)
    {
        ARMARX_VERBOSE << "requesting controller activation for:\n" << names;
        if (names.empty())
        {
            return;
        }
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        auto ctrlsToActVec = getNJointControllersNotNull(names); //also checks if these controllers exist
        ARMARX_DEBUG << "all requested controllers exist" << std::flush;
        std::set<NJointControllerPtr, NJointController::IdComp> ctrlsToAct {ctrlsToActVec.begin(), ctrlsToActVec.end()};
        ARMARX_CHECK_EXPRESSION(!ctrlsToAct.count(nullptr));
        std::set<NJointControllerPtr, NJointController::IdComp> ctrls {getRequestedNJointControllers().begin(), getRequestedNJointControllers().end()};
        ctrls.erase(nullptr);
        //check for conflict
        std::vector<char> inuse;
        //check requested controllers
        {
            auto r = NJointController::AreNotInConflict(ctrlsToAct.begin(), ctrlsToAct.end());
            if (!r)
            {
                std::stringstream ss;
                ss << "armarx::RobotUnit::activateNJointControllers: requested controllers are in conflict!\ncontrollers:\n" << names;
                ARMARX_ERROR << ss.str();
                throw InvalidArgumentException {ss.str()};
            }
            inuse = std::move(r.get());
        }
        ARMARX_DEBUG << "all requested controllers are conflict free" << std::flush;
        auto printInUse = ARMARX_STREAM_PRINTER
        {
            for (const auto c : inuse)
            {
                out << (c ? 1 : 0);
            }
        };
        ARMARX_DEBUG << "inuse field (request)\n" << printInUse;
        //add already active controllers if they are conflict free
        {
            if (ctrls.empty())
            {
                ARMARX_DEBUG << "no already requested NJointControllers";
            }
            for (const NJointControllerPtr& nJointCtrl : ctrls)
            {
                if (ctrlsToAct.count(nJointCtrl))
                {
                    continue;
                }
                auto r = nJointCtrl->isNotInConflictWith(inuse);
                if (r)
                {
                    ARMARX_DEBUG << "keeping  already requested NJointController '"
                                 << nJointCtrl->getInstanceName() << "' in   list of requested controllers";
                    ctrlsToAct.insert(nJointCtrl);
                    inuse = std::move(r.get());
                }
                else
                {
                    ARMARX_INFO << "removing already requested NJointController '"
                                << nJointCtrl->getInstanceName() << "' from list of requested controllers";
                }
            }
            ARMARX_DEBUG << "inuse field (all)\n" << printInUse;
        }
        setActivateControllersRequest(std::move(ctrlsToAct));
    }

    void armarx::RobotUnit::deactivateNJointControllers(const Ice::StringSeq& names, const Ice::Current&)
    {
        ARMARX_VERBOSE << "requesting controller deactivation for:\n" << names;
        if (names.empty())
        {
            return;
        }
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        auto ctrlsDeacVec = getNJointControllersNotNull(names); //also checks if these controllers exist
        std::set<NJointControllerPtr, NJointController::IdComp> ctrls {getRequestedNJointControllers().begin(), getRequestedNJointControllers().end()};
        for (const auto& nJointCtrlToDeactivate : ctrlsDeacVec)
        {
            ctrls.erase(nJointCtrlToDeactivate);
        }
        setActivateControllersRequest(std::move(ctrls));
    }

    void armarx::RobotUnit::switchNJointControllerSetup(const Ice::StringSeq& newSetup, const Ice::Current&)
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        auto ctrlsToActVec = getNJointControllersNotNull(newSetup); //also checks if these controllers exist
        setActivateControllersRequest({ctrlsToActVec.begin(), ctrlsToActVec.end()});
    }

    void armarx::RobotUnit::publish(StringVariantBaseMap additionalMap, StringVariantBaseMap timingMap)
    {
        auto begPublish = Now();
        static const int spamdelay = 30;

        if (state != RobotUnitState::Running)
        {
            ARMARX_DEBUG << deactivateSpam(spamdelay) << "skipping publishing in state " << state;
            return;
        }
        auto guard = getGuard();

        throwIfStateIsNot(RobotUnitState::Running, __FUNCTION__);

        //get batch proxies
        auto debugObserverBatchPrx = debugObserverPrx->ice_batchOneway();
        auto listenerBatchPrx = listenerPrx->ice_batchOneway();
        //delete NJoint queued for deletion
        {
            for (auto& pair : nJointControllersToBeDeleted)
            {
                listenerBatchPrx->nJointControllerDeleted(pair.first);
                getArmarXManager()->removeObjectNonBlocking(pair.second);
                ARMARX_VERBOSE << "deleted NJointController " << pair.first;
            }
            nJointControllersToBeDeleted.clear();
        }
        //swap buffers in
        const bool haveActivatedControllersChanged = activatedControllersChanged();
        const bool haveSensorAndControlValsChanged = sensorAndControlBufferChanged();
        //setup datastructures
        const auto& controlThreadOutputBuffer =  getSensorAndControlBuffer();
        const auto& activatedControllers   = getActivatedControllers();
        const auto& requestedControllers   = getRequestedControllers();

        const auto timestamp = controlThreadOutputBuffer.sensorValuesTimestamp;

        const auto numSensorDevices  = getNumberOfSensorDevices();
        const auto numControlDevices = getNumberOfControlDevices();

        const bool publishToObserver = !(publishIterationCount % debugObserverSkipIterations);
        const bool debugObserverPublishControlTargetsThisIteration = ObserverPublishControlTargets;
        const bool debugObserverPublishSensorValuesThisIteration = ObserverPublishSensorValues;
        StringVariantBaseMap ctrlDevMap, sensorDevMap;
        //publish publishing meta state
        additionalMap["haveActivatedControllersChanged"] = new TimedVariant {haveActivatedControllersChanged, timestamp};
        additionalMap["haveSensorAndControlValsChanged"] = new TimedVariant {haveSensorAndControlValsChanged, timestamp};
        additionalMap["publishToObserver"              ] = new TimedVariant {publishToObserver, timestamp};
        //update
        if (haveSensorAndControlValsChanged)
        {
            //update units
            timingMap["publishTimings_UnitUpdate"] = new TimedVariant
            {
                ARMARX_STOPWATCH(TimestampVariant)
                {
                    ARMARX_DEBUG << deactivateSpam(spamdelay) << "updating units with new sensor values";
                    publishNewSensorDataTime = TimeUtil::GetTime();
                    for (RobotUnitSubUnitPtr& rsu : subUnits)
                    {
                        if (rsu && rsu->getObjectScheduler() && rsu->getProxy())
                        {
                            timingMap["publishTimings_UnitUpdate_" + rsu->getName()] = new TimedVariant
                            {
                                ARMARX_STOPWATCH(TimestampVariant){rsu->update(controlThreadOutputBuffer, activatedControllers);},
                                timestamp
                            };
                        }
                    }
                },
                timestamp
            };
            //publish sensor updates

            timingMap["publishTimings_SensorUpdates"] = new TimedVariant
            {
                ARMARX_STOPWATCH(TimestampVariant)
                {
                    for (std::size_t sensidx = 0 ; sensidx < numSensorDevices; ++sensidx)
                    {
                        const SensorDevice& sensDev = *(sensorDevices.at(sensidx));
                        const SensorValueBase& sensVal = *(controlThreadOutputBuffer.sensors.at(sensidx));
                        auto variants = sensVal.toVariants(timestamp);

                        if (!variants.empty())
                        {
                            if (debugObserverPublishSensorValuesThisIteration && publishToObserver)
                            {
                                transformMapKeys(
                                    variants, sensorDevMap,
                                    detail::makeDebugObserverNameFixer(sensDev.getDeviceName() + "_" + sensVal.getSensorValueType(true) + "_")
                                );
                            }
                            SensorDeviceStatus status;
                            status.deviceName = sensDev.getDeviceName();
                            status.sensorValue = std::move(variants);
                            status.timestampUSec = MicroNow().count();
                            listenerBatchPrx->sensorDeviceStatusChanged(status);
                        }
                    }
                },
                timestamp
            };
        }
        else
        {
            const double sensSecondsAgo = (TimeUtil::GetTime() - publishNewSensorDataTime).toSecondsDouble();
            if (sensSecondsAgo > 1)
            {
                ARMARX_WARNING << deactivateSpam(spamdelay) << "armarx::RobotUnit::publish: Last sensor value update is "
                               << sensSecondsAgo << " seconds ago!";
            }
        }
        if (haveSensorAndControlValsChanged || haveActivatedControllersChanged)
        {
            timingMap["publishTimings_ControlUpdates"] = new TimedVariant
            {
                ARMARX_STOPWATCH(TimestampVariant)
                {
                    for (std::size_t ctrlidx = 0 ; ctrlidx < numControlDevices; ++ctrlidx)
                    {
                        const ControlDevice& ctrlDev = *(controlDevices.at(ctrlidx));

                        StringToStringVariantBaseMapMap variants;
                        for (const auto& ctrlVal : controlThreadOutputBuffer.control.at(ctrlidx))
                        {
                            const auto& controlMode = ctrlVal->getControlMode();
                            variants[controlMode] = ctrlVal->toVariants(timestamp);
                            if (
                                haveSensorAndControlValsChanged &&
                                !variants[controlMode].empty() &&
                                debugObserverPublishControlTargetsThisIteration  &&
                                publishToObserver
                            )
                            {
                                transformMapKeys(
                                    variants[controlMode], ctrlDevMap,
                                    detail::makeDebugObserverNameFixer(ctrlDev.getDeviceName() + "_" + controlMode + "_")
                                );
                            }
                        }
                        ControlDeviceStatus status;
                        const auto activeJointCtrl = activatedControllers.jointControllers.at(ctrlidx);
                        status.activeControlMode = activeJointCtrl ? activeJointCtrl->getControlMode() : std::string {"!!JointController is nullptr!!"};
                        status.deviceName = ctrlDev.getDeviceName();
                        ARMARX_CHECK_EXPRESSION(requestedControllers.jointControllers.at(ctrlidx));
                        status.requestedControlMode = requestedControllers.jointControllers.at(ctrlidx)->getControlMode();
                        status.controlTargetValues = std::move(variants);
                        status.timestampUSec = MicroNow().count();
                        listenerBatchPrx->controlDeviceStatusChanged(status);
                    }
                },
                timestamp
            };
        }
        //call publish hook + publish NJointController changes
        timingMap["publishTimings_NJointControllerUpdates"] = new TimedVariant
        {
            ARMARX_STOPWATCH(TimestampVariant)
            {
                for (const auto& pair : nJointControllers)
                {
                    const NJointControllerPtr& nJointCtrl = pair.second;
                    timingMap["publishTimings_NJointControllerUpdates_" + nJointCtrl->getInstanceName()] = new TimedVariant
                    {
                        ARMARX_STOPWATCH(TimestampVariant)
                        {
                            //run some hook for active (used for visu)
                            PublishNJointController(nJointCtrl, controlThreadOutputBuffer, debugDrawerPrx, debugObserverBatchPrx);
                            if (
                                GetNJointControllerStatusReportedActive(nJointCtrl) != nJointCtrl->isControllerActive() ||
                                GetNJointControllerStatusReportedRequested(nJointCtrl) != nJointCtrl->isControllerRequested()
                            )
                            {
                                UpdateNJointControllerStatusReported(nJointCtrl);
                                listenerBatchPrx->nJointControllerStatusChanged(nJointCtrl->getControllerStatus());
                            }
                        },
                        timestamp
                    };
                }
            },
            timestamp
        };

        //report new class names
        timingMap["publishTimings_ClassNameUpdates"] = new TimedVariant
        {
            ARMARX_STOPWATCH(TimestampVariant)
            {
                const auto classNames = NJointControllerRegistry::getKeys();
                if (lastReportedClasses.size() != classNames.size())
                {
                    for (const auto& name : classNames)
                    {
                        if (!lastReportedClasses.count(name))
                        {
                            listenerBatchPrx->nJointControllerClassAdded(name);
                            lastReportedClasses.emplace(name);
                        }
                    }
                }
            },
            timestamp
        };
        timingMap["publishTimings_RobotUnitListenerFlush"] = new TimedVariant
        {
            ARMARX_STOPWATCH(TimestampVariant){listenerBatchPrx->ice_flushBatchRequests();}, timestamp
        };

        if (publishToObserver)
        {
            timingMap["publishTimings_LastRobotUnitObserverFlush"] =
                new TimedVariant {TimestampVariant{lastRobotUnitObserverFlush.count()}, timestamp};
            timingMap["publishTimings_LastPublishLoop"] = new TimedVariant {TimestampVariant{lastPublishLoop.count()}, timestamp};
            lastRobotUnitObserverFlush = ARMARX_STOPWATCH()
            {
                if (robotUnitObs->getState() >= eManagedIceObjectStarted)
                {
                    robotUnitObs->offerOrUpdateDataFieldsFlatCopy(robotUnitObs->additionalChannel, additionalMap);
                    robotUnitObs->updateChannel(robotUnitObs->additionalChannel);
                    robotUnitObs->offerOrUpdateDataFieldsFlatCopy(robotUnitObs->timingChannel, timingMap);
                    robotUnitObs->updateChannel(robotUnitObs->timingChannel);
                    robotUnitObs->offerOrUpdateDataFieldsFlatCopy(robotUnitObs->controlDevicesChannel, ctrlDevMap);
                    robotUnitObs->updateChannel(robotUnitObs->controlDevicesChannel);
                    robotUnitObs->offerOrUpdateDataFieldsFlatCopy(robotUnitObs->sensorDevicesChannel, sensorDevMap);
                    robotUnitObs->updateChannel(robotUnitObs->sensorDevicesChannel);

                }
                debugObserverBatchPrx->ice_flushBatchRequests();
            };
        }
        else
        {
            lastRobotUnitObserverFlush = std::chrono::microseconds {0};
        }

        //warn if there are unset jointCtrl controllers
        {
            const auto& jointCtrls = getActivatedJointControllers();
            if (std::any_of(jointCtrls.begin(), jointCtrls.end(), [](const JointController * jointCtrl)
        {
            return !jointCtrl;
        }))
            {
                ARMARX_WARNING << "armarx::RobotUnit::publish: Some activated JointControllers are reported to be nullptr! "
                               << "(did you forget to call rtSwitchControllerSetup()?)";
            }
        }
        ++publishIterationCount;
        lastPublishLoop = MicroToNow(begPublish);
    }

    void armarx::RobotUnit::removeAllUnits()
    {
        subUnits.clear();
        for (ManagedIceObjectPtr& unit : units)
        {
            getArmarXManager()->removeObjectBlocking(unit->getName());
        }
        units.clear();
    }

    void armarx::RobotUnit::finishRunning()
    {
        ARMARX_DEBUG << "RobotUnit::finishRunning()";
        if (state == RobotUnitState::Exiting)
        {
            ARMARX_ERROR << "armarx::RobotUnit::finishRunning called multiple times!";
        }

        rtLogging.defaultLogHandle = nullptr;
        if (rtLogging.task)
        {
            ARMARX_DEBUG << "shutting down rt logging task";
            rtLogging.task->stop();
            while (rtLogging.task->isFunctionExecuting() || rtLogging.task->isRunning())
            {
                ARMARX_FATAL << deactivateSpam(0.1) << "RT LOGGING TASK IS RUNNING AFTER IT WAS STOPPED!";
            }
            rtLogging.task = nullptr;
            ARMARX_DEBUG << "shutting down rt logging task done";
        }

        if (publisherTask)
        {
            ARMARX_DEBUG << "shutting down publisher task";
            publisherTask->stop();
            auto tPublisherStopped = Now();
            while (publisherTask->isFunctionExecuting() || publisherTask->isRunning())
            {
                ARMARX_FATAL << deactivateSpam(0.1) << "PUBLISHER TASK IS RUNNING AFTER IT WAS STOPPED!";
            }
            publisherTask = nullptr;
            //since the drawer queues draw events and we want to clear the layers, we have to sleep here
            std::this_thread::sleep_until(tPublisherStopped + std::chrono::milliseconds {100});
            ARMARX_DEBUG << "shutting down publisher task done";
        }

        auto guard = getGuard();
        try
        {
            throwIfStateIsNot(RobotUnitState::Running, __FUNCTION__, true);
            state = RobotUnitState::Exiting;
            //const ptr maps
            controlDevicesConstPtr.clear();
            sensorDevicesConstPtr.clear();
            //publisher


            //RT
            ARMARX_DEBUG << "joining rt thread";
            joinRTThread();
            ARMARX_DEBUG << "joining rt thread...done";
            //call deac publishing of all active NJoint
            for (const auto& pair : nJointControllers)
            {
                ARMARX_DEBUG << "forcing deactivation of publishing for " << pair.first;
                DeactivateNJointControllerPublishing(pair.second, debugDrawerPrx, debugObserverPrx);
            }
            //units
            ARMARX_DEBUG << "removing units";
            removeAllUnits();
            ARMARX_DEBUG << "removing units...done";
            //NJoint queued for deletion (some could still be in the queue)
            {
                ARMARX_DEBUG << "remove NJointControllers queued for deletion";
                for (auto& n2NJointCtrl : nJointControllersToBeDeleted)
                {
                    NJointControllerPtr& nJointCtrl = n2NJointCtrl.second;
                    getArmarXManager()->removeObjectBlocking(nJointCtrl->getName());
                }
                nJointControllersToBeDeleted.clear();
                ARMARX_DEBUG << "remove NJointControllers queued for deletion...done";
            }
            //NJoint
            {
                ARMARX_DEBUG << "remove NJointControllers";
                for (auto& n2NJointCtrl : nJointControllers)
                {
                    NJointControllerPtr& nJointCtrl = n2NJointCtrl.second;
                    getArmarXManager()->removeObjectBlocking(nJointCtrl->getName());
                }
                nJointControllers.clear();
                ARMARX_DEBUG << "remove NJointControllers...done";
            }
            //clear devices
            sensorDevices.clear();
            controlDevices.clear();
            //robot
            if (robot.use_count() != 1)
            {
                ARMARX_WARNING << "The robot's smart pointer use count is " << robot.use_count();
            }
            robot.reset();
            ARMARX_INFO << "RobotUnit finished running";
        }
        catch (Ice::Exception& e)
        {
            ARMARX_ERROR << "exception in RobotUnit::finishRunning!\nwhat:\n"
                         << e.what()
                         << "\n\tname: " << e.ice_name()
                         << "\n\tfile: " << e.ice_file()
                         << "\n\tline: " << e.ice_line()
                         << "\n\tstack: " << e.ice_stackTrace();
        }
        catch (std::exception& e)
        {
            ARMARX_ERROR << "exception in RobotUnit::finishRunning!\nwhat:\n" << e.what();
        }
        catch (...)
        {
            ARMARX_ERROR << "exception in RobotUnit::finishRunning!";
        }
        ARMARX_DEBUG << "RobotUnit::finishRunning()...done";
    }

    NJointControllerClassDescription armarx::RobotUnit::getNJointControllerClassDescription(
        const std::string& className, const Ice::Current&) const
    {
        while (state == RobotUnitState::PreComponentInitialization || state == RobotUnitState::InitializingDevices)
        {
            //this phase should only last short so busy waiting is ok
            std::this_thread::sleep_for(std::chrono::milliseconds(50));
        }
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        checkNJointControllerClassName(className);
        NJointControllerClassDescription data;
        data.className = className;
        if (NJointControllerRegistry::get(className)->hasRemoteConfiguration())
        {
            data.configDescription =
                NJointControllerRegistry::get(className)->GenerateConfigDescription(robot, controlDevicesConstPtr, sensorDevicesConstPtr);
        }
        return data;
    }

    NJointControllerClassDescriptionSeq armarx::RobotUnit::getNJointControllerClassDescriptions(const Ice::Current&) const
    {
        std::size_t tries = 200;
        while (state == RobotUnitState::PreComponentInitialization || state == RobotUnitState::InitializingDevices)
        {
            //this phase should only last short so busy waiting is ok
            std::this_thread::sleep_for(std::chrono::milliseconds(50));
            if (! --tries)
            {
                throw RuntimeError {"RobotUnit::getNJointControllerClassDescriptions: it took too long to for the unit to get in a valid state"};
            }
        }
        auto guard = getGuard();
        NJointControllerClassDescriptionSeq r;
        r.reserve(NJointControllerRegistry::getKeys().size());
        for (const auto& key : NJointControllerRegistry::getKeys())
        {
            r.emplace_back(getNJointControllerClassDescription(key));
        }
        return r;
    }

    NJointControllerInterfacePrx armarx::RobotUnit::getNJointController(const std::string& name, const Ice::Current&) const
    {
        NJointControllerPtr ctrl;
        {
            auto guard = getGuard();
            auto it = nJointControllers.find(name);
            if (it == nJointControllers.end())
            {
                return nullptr;
            }
            ctrl = it->second;
        }
        return NJointControllerInterfacePrx::uncheckedCast(ctrl->getProxy(-1, true));
    }

    NJointControllerStatus armarx::RobotUnit::getNJointControllerStatus(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getNJointControllerNotNull(name)->getControllerStatus();
    }

    NJointControllerStatusSeq armarx::RobotUnit::getNJointControllerStatuses(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        NJointControllerStatusSeq r;
        r.reserve(nJointControllers.size());
        for (const auto& nJointCtrl : nJointControllers)
        {
            r.emplace_back(nJointCtrl.second->getControllerStatus());
        }
        return r;
    }

    NJointControllerDescription armarx::RobotUnit::getNJointControllerDescription(const std::string& name, const Ice::Current&) const
    {
        NJointControllerPtr nJointCtrl;
        {
            auto guard = getGuard();
            throwIfDevicesNotReady(__FUNCTION__);
            nJointCtrl = getNJointControllerNotNull(name);
        }
        return nJointCtrl->getControllerDescription();
    }

    NJointControllerDescriptionSeq armarx::RobotUnit::getNJointControllerDescriptions(const Ice::Current&) const
    {
        std::map<std::string, NJointControllerPtr> nJointControllersCopy;
        {
            auto guard = getGuard();
            if (!areDevicesReady())
            {
                return {};
            }
            nJointControllersCopy = nJointControllers;
        }
        NJointControllerDescriptionSeq r;
        r.reserve(nJointControllersCopy.size());
        for (const auto& nJointCtrl : nJointControllersCopy)
        {
            r.emplace_back(nJointCtrl.second->getControllerDescription());
        }
        return r;
    }

    NJointControllerDescriptionWithStatus armarx::RobotUnit::getNJointControllerDescriptionWithStatus(
        const std::string& name, const Ice::Current&) const
    {
        NJointControllerPtr nJointCtrl;
        {
            auto guard = getGuard();
            throwIfDevicesNotReady(__FUNCTION__);
            nJointCtrl = getNJointControllerNotNull(name);
        }
        return nJointCtrl->getControllerDescriptionWithStatus();
    }

    NJointControllerDescriptionWithStatusSeq armarx::RobotUnit::getNJointControllerDescriptionsWithStatuses(const Ice::Current&) const
    {
        std::map<std::string, NJointControllerPtr> nJointControllersCopy;
        {
            auto guard = getGuard();
            if (!areDevicesReady())
            {
                return {};
            }
            nJointControllersCopy = nJointControllers;
        }
        NJointControllerDescriptionWithStatusSeq r;
        r.reserve(nJointControllersCopy.size());
        for (const auto& nJointCtrl : nJointControllersCopy)
        {
            r.emplace_back(nJointCtrl.second->getControllerDescriptionWithStatus());
        }
        return r;
    }

    Ice::StringSeq armarx::RobotUnit::getControlDeviceNames(const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controlDevices.keys();
    }

    ControlDeviceDescription armarx::RobotUnit::getControlDeviceDescription(std::size_t idx) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        const ControlDevicePtr& controlDevice = controlDevices.at(idx);
        ControlDeviceDescription data;
        data.deviceName = controlDevice->getDeviceName();
        data.tags.assign(controlDevice->getTags().begin(), controlDevice->getTags().end());
        for (const auto& jointCtrl : controlDevice->getJointControllers())
        {
            data.contolModeToTargetType[jointCtrl->getControlMode()].targetType = jointCtrl->getControlTarget()->getControlTargetType();
            data.contolModeToTargetType[jointCtrl->getControlMode()].hardwareControlMode = jointCtrl->getHardwareControlMode();
        }
        return data;
    }

    ControlDeviceDescription armarx::RobotUnit::getControlDeviceDescription(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (!controlDevices.has(name))
        {
            std::stringstream ss;
            ss << "armarx::RobotUnit::getControlDeviceDescription: There is no ControlDevice '" << name
               << "'. There are these ControlDevices: " << controlDevices.keys();
            throw InvalidArgumentException {ss.str()};
        }
        return getControlDeviceDescription(controlDevices.index(name));
    }

    ControlDeviceDescriptionSeq armarx::RobotUnit::getControlDeviceDescriptions(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        ControlDeviceDescriptionSeq r;
        r.reserve(getNumberOfControlDevices());
        for (auto idx : getIndices(controlDevices.values()))
        {
            r.emplace_back(getControlDeviceDescription(idx));
        }
        return r;
    }

    ControlDeviceStatus armarx::RobotUnit::getControlDeviceStatus(std::size_t idx) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        const ControlDevicePtr& controlDevice = controlDevices.at(idx);
        ControlDeviceStatus status;
        const auto activeJointCtrl = getActivatedJointControllers().at(idx);
        status.activeControlMode = activeJointCtrl ? activeJointCtrl->getControlMode() : std::string {"!!JointController is nullptr!!"};
        status.deviceName = controlDevice->getDeviceName();
        ARMARX_CHECK_EXPRESSION(getRequestedJointControllers().at(idx));
        status.requestedControlMode = getRequestedJointControllers().at(idx)->getControlMode();
        for (const auto& targ : getSensorAndControlBuffer().control.at(idx))
        {
            status.controlTargetValues[targ->getControlMode()] = targ->toVariants(getSensorAndControlBuffer().sensorValuesTimestamp);
        }
        status.timestampUSec = MicroNow().count();
        return status;
    }

    ControlDeviceStatus armarx::RobotUnit::getControlDeviceStatus(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (!controlDevices.has(name))
        {
            std::stringstream ss;
            ss << "armarx::RobotUnit::getControlDeviceStatus: There is no ControlDevice '" << name
               << "'. There are these ControlDevices: " << controlDevices.keys();
            throw InvalidArgumentException {ss.str()};
        }
        return getControlDeviceStatus(controlDevices.index(name));
    }

    ControlDeviceStatusSeq armarx::RobotUnit::getControlDeviceStatuses(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        ControlDeviceStatusSeq r;
        r.reserve(getNumberOfControlDevices());
        for (auto idx : getIndices(controlDevices.values()))
        {
            r.emplace_back(getControlDeviceStatus(idx));
        }
        return r;
    }

    Ice::StringSeq armarx::RobotUnit::getSensorDeviceNames(const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return sensorDevices.keys();
    }

    SensorDeviceDescription armarx::RobotUnit::getSensorDeviceDescription(std::size_t idx) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        const SensorDevicePtr& sensorDevice = sensorDevices.at(idx);
        SensorDeviceDescription data;
        data.deviceName = sensorDevice->getDeviceName();
        data.tags.assign(sensorDevice->getTags().begin(), sensorDevice->getTags().end());
        data.sensorValueType = sensorDevice->getSensorValueType();
        return data;
    }

    SensorDeviceDescription armarx::RobotUnit::getSensorDeviceDescription(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (!sensorDevices.has(name))
        {
            std::stringstream ss;
            ss << "armarx::RobotUnit::getSensorDeviceDescription: There is no SensorDevice '" << name
               << "'. There are these SensorDevices: " << sensorDevices.keys();
            throw InvalidArgumentException {ss.str()};
        }
        return getSensorDeviceDescription(sensorDevices.index(name));
    }

    SensorDeviceDescriptionSeq armarx::RobotUnit::getSensorDeviceDescriptions(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        SensorDeviceDescriptionSeq r;
        r.reserve(getNumberOfSensorDevices());
        for (auto idx : getIndices(sensorDevices.values()))
        {
            r.emplace_back(getSensorDeviceDescription(idx));
        }
        return r;
    }

    SensorDeviceStatus armarx::RobotUnit::getSensorDeviceStatus(std::size_t idx) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        const SensorDevicePtr& sensorDevice = sensorDevices.at(idx);
        SensorDeviceStatus status;
        status.deviceName = sensorDevice->getDeviceName();
        status.sensorValue = getSensorAndControlBuffer().sensors.at(idx)->toVariants(getSensorAndControlBuffer().sensorValuesTimestamp);
        status.timestampUSec = MicroNow().count();
        return status;
    }

    SensorDeviceStatus armarx::RobotUnit::getSensorDeviceStatus(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (!sensorDevices.has(name))
        {
            std::stringstream ss;
            ss << "armarx::RobotUnit::getSensorDeviceStatus: There is no SensorDevice '" << name
               << "'. There are these SensorDevices: " << sensorDevices.keys();
            throw InvalidArgumentException {ss.str()};
        }
        return getSensorDeviceStatus(sensorDevices.index(name));
    }

    SensorDeviceStatusSeq armarx::RobotUnit::getSensorDeviceStatuses(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        SensorDeviceStatusSeq r;
        r.reserve(getNumberOfSensorDevices());
        for (auto idx : getIndices(sensorDevices.values()))
        {
            r.emplace_back(getSensorDeviceStatus(idx));
        }
        return r;
    }

    std::vector<armarx::NJointControllerPtr> armarx::RobotUnit::getNJointControllersNotNull(const std::vector<std::string>& names) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        std::vector<NJointControllerPtr> ctrl;
        ctrl.reserve(names.size());
        for (const auto& name : names)
        {
            ctrl.emplace_back(getNJointControllerNotNull(name));
        }
        return ctrl;
    }
    const NJointControllerPtr& armarx::RobotUnit::getNJointControllerNotNull(const std::string& name) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        auto it = nJointControllers.find(name);
        if (it == nJointControllers.end())
        {
            std::stringstream ss;
            ss << "RobotUnit: there is no NJointController with name '" << name
               << "'. Existing NJointControllers are: " << getNJointControllerNames();
            throw InvalidArgumentException {ss.str()};
        }
        if (!it->second)
        {
            std::stringstream ss;
            ss << "RobotUnit: The NJointController with name '" << name
               << "'. Is a nullptr! This should never be the case (invariant)!  \nMap:\n" << nJointControllers;
            ARMARX_FATAL << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
        return it->second;
    }

    StringNJointControllerPrxDictionary armarx::RobotUnit::getAllNJointControllers(const Ice::Current&) const
    {
        std::map<std::string, NJointControllerPtr> nJointControllersCopy;
        {
            auto guard = getGuard();
            //copy to keep lock retention time low
            nJointControllersCopy = nJointControllers;
        }
        StringNJointControllerPrxDictionary result;
        for (const auto& pair : nJointControllersCopy)
        {
            result[pair.first] = NJointControllerInterfacePrx::uncheckedCast(pair.second->getProxy(-1, true));
        }
        return result;
    }

    Ice::ObjectProxySeq armarx::RobotUnit::getUnits(const Ice::Current&) const
    {
        std::vector<ManagedIceObjectPtr> unitsCopy;
        {
            auto guard = getGuard();
            //copy to keep lock retention time low
            unitsCopy = units;
        }
        Ice::ObjectProxySeq r;
        r.reserve(unitsCopy.size());
        for (const ManagedIceObjectPtr& unit : unitsCopy)
        {
            r.emplace_back(unit->getProxy(-1, true));
        }
        return r;
    }

    const ManagedIceObjectPtr& armarx::RobotUnit::getUnit(const std::string& staticIceId) const
    {
        auto guard = getGuard();
        for (const ManagedIceObjectPtr& unit : units)
        {
            if (unit->ice_isA(staticIceId))
            {
                return unit;
            }
        }
        return ManagedIceObject::NullPtr;
    }

    Ice::ObjectPrx armarx::RobotUnit::getUnit(const std::string& staticIceId, const Ice::Current&) const
    {
        //no lock required
        ManagedIceObjectPtr unit = getUnit(staticIceId);
        if (unit)
        {
            return unit->getProxy(-1, true);
        }
        return nullptr;
    }

    const VirtualRobot::RobotPtr& armarx::RobotUnit::getRobot() const
    {
        auto guard = getGuard();
        throwIfStateIsNot(
        {
            RobotUnitState::InitializingDevices,
            RobotUnitState::InitializingUnits,
            RobotUnitState::WaitingForRTThreadInitialization,
            RobotUnitState::Running
        },
        __FUNCTION__
        );
        ARMARX_CHECK_EXPRESSION(robot);
        return robot;
    }

    std::string armarx::RobotUnit::getRobotName() const
    {
        auto guard = getGuard();
        throwIfStateIsNot(
        {
            RobotUnitState::InitializingDevices,
            RobotUnitState::InitializingUnits,
            RobotUnitState::WaitingForRTThreadInitialization,
            RobotUnitState::Running
        },
        __FUNCTION__
        );
        return getRobot()->getName();
    }

    std::size_t armarx::RobotUnit::getNumberOfControlDevices() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controlDevices.size();
    }

    std::size_t armarx::RobotUnit::getNumberOfSensorDevices() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return sensorDevices.size();
    }

    const DebugDrawerInterfacePrx& armarx::RobotUnit::getDebugDrawerProxy() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return debugDrawerPrx;
    }

    const RobotUnitListenerPrx& armarx::RobotUnit::getListenerProxy() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return listenerPrx;
    }

    void armarx::RobotUnit::icePropertiesInitialized()
    {
        robotUnitObs = Component::create<RobotUnitObserver>(getIceProperties(), "", getConfigDomain());
        addPropertyUser(PropertyUserPtr::dynamicCast(robotUnitObs));

        Ice::PropertiesPtr properties = getIceProperties()->clone();
        const std::string& configDomain = "ArmarX";
        // create TCPControlSubUnit

        {
            const std::string configNameTCPControlUnit = getProperty<std::string>("TCPControlUnitName");
            tcpControllerSubUnit = Component::create<TCPControllerSubUnit>(properties, configNameTCPControlUnit, configDomain);
            addPropertyUser(PropertyUserPtr::dynamicCast(tcpControllerSubUnit));
        }

        // create TrajectoryControllerSubUnit

        {
            const std::string configNameTrajectoryControllerUnit = getProperty<std::string>("TrajectoryControllerUnitName");
            const std::string confPre = configDomain + "." + configNameTrajectoryControllerUnit + ".";
            properties->setProperty(confPre + "KinematicUnitName", getProperty<std::string>("KinematicUnitName").getValue());
            trajectoryControllerSubUnit = Component::create<TrajectoryControllerSubUnit>(properties, configNameTrajectoryControllerUnit, configDomain);
            addPropertyUser(PropertyUserPtr::dynamicCast(trajectoryControllerSubUnit));
        }
    }


    RemoteReferenceCounterBasePtr armarx::RobotUnit::startRtLoggingWithAliasNames(const std::string& formatString, const StringStringDictionary& aliasNames, const Ice::Current&)
    {
        if (!isRtLoggingEnabled())
        {
            throw LogicError {"RtLogging is disabled in the config! (set RTLoggingPeriodMs to a non zero value"};
        }
        FileSystemPathBuilder pb {formatString};
        std::lock_guard<std::mutex> guard {rtLogging.mutex};
        if (rtLogging.entries.count(pb.getPath()))
        {
            throw InvalidArgumentException {"There already is a logger for the path '" + pb.getPath() + "'"};
        }
        auto ptr = std::make_shared<RtLogging::Entry>();
        rtLogging.entries[pb.getPath()] = ptr;
        RtLogging::Entry& e = *ptr;
        e.filename = pb.getPath();
        pb.createParentDirectories();
        e.stream.open(e.filename);
        if (!e.stream)
        {
            rtLogging.entries.erase(pb.getPath());
            throw LogicError {"RtLogging could not open filestream for '" + pb.getPath() + "'"};
        }

        std::stringstream header;
        header << "marker;iteration;timestamp";
        auto logDev = [&](const std::string & dev)
        {
            for (const auto& pair : aliasNames)
            {
                std::string name = pair.first;
                if (boost::starts_with(dev + "$", name))
                {
                    if ((name.size() > 0 && name.back() == '$'))
                    {
                        name.resize(name.size() - 1);
                    }
                    const std::string& alias = pair.second.empty() ? name : pair.second;
                    header << ";" << alias << dev.substr(name.size());
                    return true;
                }
            }
            return false;
        };

        //get logged sensor device values
        {
            e.loggedSensorDeviceValues.reserve(rtLogging.sensorDeviceValueLoggingNames.size());
            for (const auto& svfields : rtLogging.sensorDeviceValueLoggingNames)
            {
                e.loggedSensorDeviceValues.emplace_back();
                auto& svfieldsFlags = e.loggedSensorDeviceValues.back(); //vv
                svfieldsFlags.reserve(svfields.size());
                for (const auto& field : svfields)
                {
                    svfieldsFlags.emplace_back(logDev(field));
                }
            }
        }
        //get logged control device values
        {
            e.loggedControlDeviceValues.reserve(rtLogging.controlDeviceValueLoggingNames.size());
            for (const auto& sjctrls : rtLogging.controlDeviceValueLoggingNames)
            {
                e.loggedControlDeviceValues.emplace_back();
                auto& sjctrlsFlags = e.loggedControlDeviceValues.back(); //vv
                sjctrlsFlags.reserve(sjctrls.size());
                for (const auto& ctargfields : sjctrls)
                {

                    sjctrlsFlags.emplace_back();
                    auto& ctargfieldsFlags = sjctrlsFlags.back(); //v
                    ctargfieldsFlags.reserve(ctargfields.size());

                    for (const auto& field : ctargfields)
                    {
                        ctargfieldsFlags.emplace_back(logDev(field));
                    }
                }
            }
        }

        //write header
        e.stream << header.str() << std::flush; // newline is written at the beginning of each log line
        //create and return handle
        auto block = getArmarXManager()->createRemoteReferenceCount(
                         [ptr]()
        {
            ARMARX_DEBUG_S << "RobotUnit: stop RtLogging for file " << ptr->filename;
            ptr->stopLogging = true;
        }, e.filename);
        auto counter = block->getReferenceCounter();
        block->activateCounting();
        ARMARX_DEBUG_S << "RobotUnit: start RtLogging for file " << ptr->filename;
        return counter;
    }

    RemoteReferenceCounterBasePtr armarx::RobotUnit::startRtLogging(const std::string& formatString, const Ice::StringSeq& loggingNames, const Ice::Current&)
    {
        StringStringDictionary alias;
        for (const auto& name : loggingNames)
        {
            alias.emplace(name, "");
        }
        return startRtLoggingWithAliasNames(formatString, alias);
    }

    void armarx::RobotUnit::addMarkerToRtLog(const RemoteReferenceCounterBasePtr& token, const std::string& marker, const Ice::Current&)
    {
        if (!isRtLoggingEnabled())
        {
            throw LogicError {"RtLogging is disabled in the config! (set RTLoggingPeriodMs to a non zwero value"};
        }
        std::lock_guard<std::mutex> guard {rtLogging.mutex};
        if (!rtLogging.entries.count(token->getId()))
        {
            throw InvalidArgumentException {"addMarkerToRtLog called for a nonexistent log"};
        }
        rtLogging.entries.at(token->getId())->marker = marker;
    }

    void armarx::RobotUnit::stopRtLogging(const armarx::RemoteReferenceCounterBasePtr& token, const Ice::Current&)
    {
        if (!isRtLoggingEnabled())
        {
            throw LogicError {"RtLogging is disabled in the config! (set RTLoggingPeriodMs to a non zwero value"};
        }
        std::lock_guard<std::mutex> guard {rtLogging.mutex};
        if (!rtLogging.entries.count(token->getId()))
        {
            throw InvalidArgumentException {"addMarkerToRtLog called for a nonexistent log"};
        }
        ARMARX_DEBUG_S << "RobotUnit: stop RtLogging for file " << rtLogging.entries.at(token->getId())->filename;
        rtLogging.entries.at(token->getId())->stopLogging = true;
    }

    Ice::StringSeq armarx::RobotUnit::getLoggingNames(const Ice::Current&) const
    {
        Ice::StringSeq result;
        for (const auto& strs : rtLogging.sensorDeviceValueLoggingNames)
        {
            result.insert(result.end(), strs.begin(), strs.end());
        }
        for (const auto& strvecss : rtLogging.controlDeviceValueLoggingNames)
        {
            for (const auto& strs : strvecss)
            {
                result.insert(result.end(), strs.begin(), strs.end());
            }
        }
        return result;
    }

    void armarx::RobotUnit::writeRecentIterationsToFile(const std::string& formatString, const Ice::Current&) const
    {
        if (!isRtLoggingEnabled())
        {
            throw LogicError {"RtLogging is disabled in the config! (set RTLoggingPeriodMs to a non zero value"};
        }
        std::lock_guard<std::mutex> guard {rtLogging.mutex};
        FileSystemPathBuilder pb {formatString};
        pb.createParentDirectories();
        std::ofstream outCSV {pb.getPath() + ".csv"};
        if (!outCSV)
        {
            throw LogicError {"writeRecentIterationsToFile could not open filestream for '" + pb.getPath() + ".csv'"};
        }
        std::ofstream outMsg {pb.getPath() + ".messages"};
        if (!outMsg)
        {
            throw LogicError {"writeRecentIterationsToFile could not open filestream for '" + pb.getPath() + ".messages'"};
        }
        ARMARX_INFO << "writing the last " << rtLogging.backlog.size() << " iterations to " << pb.getPath() << ".{csv, messages}";
        //write csv header
        {
            outCSV << "iteration;timestamp";
            for (const auto& vs : rtLogging.sensorDeviceValueLoggingNames)
            {
                for (const auto& s : vs)
                {
                    outCSV << ";" << s;
                }
            }
            for (const auto& vvs : rtLogging.controlDeviceValueLoggingNames)
            {
                for (const auto& vs : vvs)
                {
                    for (const auto& s : vs)
                    {
                        outCSV << ";" << s;
                    }
                }
            }
            outCSV << std::endl;
        }

        for (const detail::ControlThreadOutputBufferEntry& iteration : rtLogging.backlog)
        {
            //write csv data
            {
                outCSV  << iteration.iteration << ";" << iteration.sensorValuesTimestamp;
                //sens
                {
                    for (const SensorValueBase* val : iteration.sensors)
                    {
                        for (std::size_t idxField = 0; idxField < val->getNumberOfDataFields(); ++ idxField)
                        {
                            outCSV  << ";" << val->getDataFieldAsString(idxField);
                        }
                    }
                }
                //ctrl
                {
                    for (const auto& vals : iteration.control)
                    {
                        for (const ControlTargetBase* val : vals)
                        {
                            for (std::size_t idxField = 0; idxField < val->getNumberOfDataFields(); ++ idxField)
                            {
                                outCSV  << ";" << val->getDataFieldAsString(idxField);
                            }
                        }
                    }
                }
                outCSV << std::endl;
            }
            //write message data
            {
                for (const detail::RtMessageLogEntryBase* msg : iteration.messages.getEntries())
                {
                    if (!msg)
                    {
                        break;
                    }
                    outMsg << "[" << msg->getTime().toDateTime() << "] iteration "
                           << iteration.iteration << ":\n"
                           << msg->format() << std::endl;
                }
                outMsg << "\nmessages lost: " << iteration.messages.messagesLost
                       << " (required additional "
                       << iteration.messages.requiredAdditionalBufferSpace << " bytes, "
                       << iteration.messages.requiredAdditionalEntries << " message entries)\n" << std::endl;
            }
        }
    }

    const RobotUnit::SelfCollisionAvoidanceData& RobotUnit::getSelfCollisionAvoidanceData() const
    {
        return selfCollisionAvoidanceData;
    }

    const VirtualRobot::RobotPtr& RobotUnit::getRtRobot() const
    {
        return rtRobot;
    }

    void armarx::RobotUnit::startSelfCollisionAvoidance()
    {
        // fill self collision avoidance data
        // clone robot
        //        VirtualRobot::CollisionCheckerPtr cc(new VirtualRobot::CollisionChecker());
        selfCollisionAvoidanceData.robot = this->robot->clone(robot->getName() + "_Clone_SelfCollisionAvoidance");

        selfCollisionAvoidanceData.minDistance = getProperty<float>("MinSelfDistance").getValue();

        // set collision models
        std::map<std::string, VirtualRobot::SceneObjectSetPtr> colModels;
        const std::string colModelsString = getProperty<std::string>("SelfCollisionModelPairs").getValue();
        std::vector<std::string> entries;
        if (!colModelsString.empty())
        {
            entries = armarx::Split(colModelsString, ";", true, true);
        }
        for (std::string entry : entries)
        {
            // Removing parentheses
            boost::trim_if(entry, boost::is_any_of("{}"));
            std::string first = armarx::Split(entry, ",", true, true)[0];
            std::string second = armarx::Split(entry, ",", true, true)[1];

            if (!selfCollisionAvoidanceData.robot->hasRobotNodeSet(first) && first != "FLOOR")
            {
                ARMARX_WARNING << "No RobotNodeSet with name '" << first << "' defined in " << robotFileName << ". Skipping.";
                continue;
            }
            if (!selfCollisionAvoidanceData.robot->hasRobotNodeSet(second) && second != "FLOOR")
            {
                ARMARX_WARNING << "No RobotNodeSet with name '" << second << "' defined in " << robotFileName << ". Skipping.";
                continue;
            }

            selfCollisionAvoidanceData.modelPairs.push_back(std::make_pair(first, second));

            auto inflateNodeSet = [&](VirtualRobot::RobotNodeSetPtr nodeset)
            {
                for (int i = 0; i < static_cast<int>(nodeset->getSize()); ++i)
                {
                    if (nodeset->getNode(i)->getCollisionModel())
                    {
                        nodeset->getNode(i)->getCollisionModel()->inflateModel(selfCollisionAvoidanceData.minDistance / 2.f);
                    }
                    else
                    {
                        ARMARX_WARNING << "Self Collision Avoidance: No collision model found for '" << nodeset->getNode(i)->getName() << "'";
                    }
                }
            };

            if (colModels.find(first) == colModels.end() && first != "FLOOR")
            {
                inflateNodeSet(selfCollisionAvoidanceData.robot->getRobotNodeSet(first));
                colModels[first] = selfCollisionAvoidanceData.robot->getRobotNodeSet(first);
            }
            if (colModels.find(second) == colModels.end() && second != "FLOOR")
            {
                inflateNodeSet(selfCollisionAvoidanceData.robot->getRobotNodeSet(second));
                colModels[second] = selfCollisionAvoidanceData.robot->getRobotNodeSet(second);
            }
        }

        // Adding scene-object representing the floor, to avoid collisions with the floor
        VirtualRobot::SceneObjectSetPtr boxSet(new VirtualRobot::SceneObjectSet("FLOOR"));
        float floorHeight = 10.0f;
        VirtualRobot::ObstaclePtr boxOb = VirtualRobot::Obstacle::createBox(20000, 20000, floorHeight);
        boxOb->getCollisionModel()->inflateModel(selfCollisionAvoidanceData.minDistance / 2);
        Eigen::Matrix4f globalPose = Eigen::Matrix4f::Identity();

        boxOb->setGlobalPose(globalPose);
        boxSet->addSceneObject(boxOb);
        colModels["FLOOR"] = boxSet;

        selfCollisionAvoidanceData.collisionModels = colModels;

        // Setup robot synchronization
        this->selfCollisionAvoidanceRobotSynchronization.reset(new RobotSynchronization(this, selfCollisionAvoidanceData.robot, selfCollisionAvoidanceData.robot->getRobotNodeNames()));

        // start the task
        int colChecksPerSecond = getProperty<int>("SelfCollisionChecksPerSecond").getValue();
        int intervalMs = (1.0 / (float) colChecksPerSecond) * 1000;
        selfCollisionAvoidanceTask = new PeriodicTask<RobotUnit>(this, &RobotUnit::updateSelfCollisionAvoidance, intervalMs, false, "SelfCollisionAvoidance");
        selfCollisionAvoidanceTask->setDelayWarningTolerance(200);
        selfCollisionAvoidanceTask->start();

        // logging
        std::string validCollisionModelsString;
        for (auto it = selfCollisionAvoidanceData.collisionModels.begin(); it != selfCollisionAvoidanceData.collisionModels.end(); ++it)
        {
            validCollisionModelsString += it->first;
            validCollisionModelsString += " ";
        }
        ARMARX_IMPORTANT << "Started Self Collision Avoidance:" <<
                         "\nRobot: " << selfCollisionAvoidanceData.robot->getName() <<
                         "\nMin. allowed distance: " << selfCollisionAvoidanceData.minDistance << " mm" <<
                         "\nCollision models: " << validCollisionModelsString <<
                         "\nCollision checks per second: " << colChecksPerSecond << " (every " << intervalMs << " ms)";
    }

    void armarx::RobotUnit::stopSelfCollisionAvoidance()
    {
        ARMARX_IMPORTANT << "Stopping Self Collision Avoidance.";
        if (selfCollisionAvoidanceTask)
        {
            selfCollisionAvoidanceTask->stop();
        }
    }

    void armarx::RobotUnit::updateSelfCollisionAvoidance()
    {
        if (emergencyStop || !getProperty<bool>("EnableSelfCollisionAvoidance").getValue())
        {
            return;
        }
        if (!devicesReadyStates.count(state) || state != RobotUnitState::Running)
        {
            ARMARX_VERBOSE << deactivateSpam(1) << "Self Distance Check: Waiting for device initialization.";
            return;
        }

        auto startTime = std::chrono::high_resolution_clock::now();

        // update joint values
        selfCollisionAvoidanceRobotSynchronization->sync();
        float absSum = 0;
        VirtualRobot::RobotPtr tmpRobot = selfCollisionAvoidanceData.robot;
        for (auto& node : tmpRobot->getRobotNodes())
        {
            absSum += std::abs(node->getJointValue());
        }
        // only check collision if joint values are not completely zero
        if (absSum < 0.00001f)
        {
            ARMARX_INFO << deactivateSpam() << "All joint values are zero - skipping collision check for now";
            return;
        }


        // check distances between all collision models
        std::vector<VirtualRobot::SceneObjectSetPtr> conflictingColModels;
        std::string distancesString("All collision states:\n");
        bool anyCollision = false;
        for (auto it = selfCollisionAvoidanceData.modelPairs.begin(); it != selfCollisionAvoidanceData.modelPairs.end(); ++it)
        {
            bool collision = selfCollisionAvoidanceData.robot->getCollisionChecker()->checkCollision(selfCollisionAvoidanceData.collisionModels[it->first], selfCollisionAvoidanceData.collisionModels[it->second]);
            anyCollision |= collision;
            distancesString += "---- " + it->first + " <-> " + it->second + " in Collision: " + (collision ? "True" : "False") +  "\n";
            if (collision)
            {
                conflictingColModels.push_back(selfCollisionAvoidanceData.collisionModels[it->first]);
                conflictingColModels.push_back(selfCollisionAvoidanceData.collisionModels[it->second]);
            }
        }

        auto endTime = std::chrono::high_resolution_clock::now();
        std::chrono::duration<double> duration = std::chrono::duration_cast<std::chrono::duration<double>>(endTime - startTime);

        if (anyCollision)
        {
            // collision
            std::string conflictingColModelsString = "";
            for (std::size_t i = 0; i < conflictingColModels.size(); i++)
            {
                conflictingColModelsString += conflictingColModels.at(i)->getName() + ((i % 2) ? ";" : "<->");
            }

            ARMARX_WARNING << "Self Distance Check: Collision. Entering EmergencyStop." <<
                           "\nmin. allowed distance: " << selfCollisionAvoidanceData.minDistance << " mm" <<
                           "\nConflicting collision models: " << conflictingColModelsString <<
                           "\n" << distancesString <<
                           "\nComputation time: " << (duration.count() * 1000) << " ms (max. allowed duration (as specified): " << (1000.0 / getProperty<int>("SelfCollisionChecksPerSecond").getValue()) << " ms)";;

            setEmergencyStopState(EmergencyStopState::eEmergencyStopActive);
            getEmergencyStopMaster()->setEmergencyStopState(EmergencyStopState::eEmergencyStopActive);
            // since at least one of the NJoint-Controllers is going to cause a collision, we just kick them all.
            setActivateControllersRequest({});
        }
        else
        {
            // no collision
            ARMARX_DEBUG << deactivateSpam(5) << "Self Distance Check: OK." <<
                         "\nmin. allowed distance: " << selfCollisionAvoidanceData.minDistance << " mm" <<
                         "\n" << distancesString <<
                         "\nComputation time: " << (duration.count() * 1000) << " ms (max. allowed duration (as specified): " << (1000.0 / getProperty<int>("SelfCollisionChecksPerSecond").getValue()) << " ms)";
        }

    }

    void armarx::RobotUnit::componentPropertiesUpdated(const std::set<std::string>& changedProperties)
    {
        if (changedProperties.count("ObserverPublishSensorValues"))
        {
            ObserverPublishSensorValues = getProperty<bool>("ObserverPublishSensorValues").getValue();
        }
        if (changedProperties.count("ObserverPublishControlTargets"))
        {
            ObserverPublishControlTargets = getProperty<bool>("ObserverPublishControlTargets").getValue();
        }
        if (changedProperties.count("ObserverPrintEveryNthIterations"))
        {
            debugObserverSkipIterations = getProperty<std::uint64_t>("ObserverPrintEveryNthIterations").getValue();
        }
    }

    void armarx::RobotUnit::onInitComponent()
    {
        ARMARX_DEBUG << "armarx::RobotUnit::onInitComponent()";
        //        EmergencyStopMaster::onInitComponent();
        {
            auto guard = getGuard();
            getArmarXManager()->increaseSchedulers(static_cast<int>(getProperty<std::uint64_t>("AdditionalObjectSchedulerCount").getValue()));

            throwIfStateIsNot(RobotUnitState::PreComponentInitialization, __FUNCTION__);
            robotUnitListenerTopicName = getProperty<std::string>("RobotUnitListenerTopicName").getValue();
            offeringTopic(getListenerTopicName());
            offeringTopic(getProperty<std::string>("DebugDrawerUpdatesTopicName").getValue());

            ObserverPublishSensorValues = getProperty<bool>("ObserverPublishSensorValues").getValue();
            ObserverPublishControlTargets = getProperty<bool>("ObserverPublishControlTargets").getValue();

            checkControllerExecutionTimings = getProperty<bool>("CheckControllerExecutionTimings");

            //load robot
            {
                robotNodeSetName            = getProperty<std::string>("RobotNodeSetName").getValue();
                robotProjectName            = getProperty<std::string>("RobotFileNameProject").getValue();
                robotFileName               = getProperty<std::string>("RobotFileName").getValue();
                robotPlatformName           = getProperty<std::string>("PlatformName").getValue();
                debugObserverSkipIterations = getProperty<std::uint64_t>("ObserverPrintEveryNthIterations").getValue();
                if (debugObserverSkipIterations < 1)
                {
                    debugObserverSkipIterations = 1;
                }
                Ice::StringSeq includePaths;
                if (!robotProjectName.empty())
                {
                    CMakePackageFinder finder(robotProjectName);
                    Ice::StringSeq projectIncludePaths;
                    auto pathsString = finder.getDataDir();
                    boost::split(projectIncludePaths,
                                 pathsString,
                                 boost::is_any_of(";,"),
                                 boost::token_compress_on);
                    includePaths.insert(includePaths.end(), projectIncludePaths.begin(), projectIncludePaths.end());
                }
                if (!ArmarXDataPath::getAbsolutePath(robotFileName, robotFileName, includePaths))
                {
                    throw UserException("Could not find robot file " + robotFileName);
                }
                // read the robot
                try
                {
                    robot = VirtualRobot::RobotIO::loadRobot(robotFileName, VirtualRobot::RobotIO::eFull);
                    rtRobot = VirtualRobot::RobotIO::loadRobot(robotFileName, VirtualRobot::RobotIO::eStructure);
                    rtRobot->setUpdateCollisionModel(false);
                    rtRobot->setUpdateVisualization(false);
                    rtRobot->setThreadsafe(false);
                }
                catch (VirtualRobot::VirtualRobotException& e)
                {
                    throw UserException(e.what());
                }
                ARMARX_INFO << "Loaded robot:"
                            << "\n\tProject      : " << robotProjectName
                            << "\n\tRobot file   : " << robotFileName
                            << "\n\tRobotNodeSet : " << robotNodeSetName
                            << "\n\tNodeNames    : " << robot->getRobotNodeSet(robotNodeSetName)->getNodeNames();
            }
            //ControlDeviceHardwareControlModeGroups
            const std::string controlDeviceHardwareControlModeGroupsStr = getProperty<std::string>("ControlDeviceHardwareControlModeGroups").getValue();
            if (!controlDeviceHardwareControlModeGroupsStr.empty())
            {
                const auto numGroups = std::count(
                                           controlDeviceHardwareControlModeGroupsStr.begin(),
                                           controlDeviceHardwareControlModeGroupsStr.end(), ';'
                                       ) + 1;
                ctrlModeGroups.groups.reserve(numGroups);
                std::vector<std::string> strGroups;
                strGroups.reserve(numGroups);
                boost::algorithm::split(strGroups, controlDeviceHardwareControlModeGroupsStr, boost::is_any_of(";"));
                for (const auto& gstr : strGroups)
                {
                    std::vector<std::string> strElems;
                    boost::algorithm::split(strElems, gstr, boost::is_any_of(","));
                    std::set<std::string> group;
                    for (auto& device : strElems)
                    {
                        boost::algorithm::trim(device);
                        ARMARX_CHECK_EXPRESSION_W_HINT(
                            !device.empty(),
                            "The ControlDeviceHardwareControlModeGroups property contains empty device names"
                        );
                        ARMARX_CHECK_EXPRESSION_W_HINT(
                            !ctrlModeGroups.groupsMerged.count(device),
                            "The ControlDeviceHardwareControlModeGroups property contains duplicate device names: " << device
                        );
                        ctrlModeGroups.groupsMerged.emplace(device);
                        group.emplace(std::move(device));
                    }
                    if (!group.empty())
                    {
                        ARMARX_DEBUG << "adding device group:\n"
                                     << ARMARX_STREAM_PRINTER
                        {
                            for (const auto& elem : group)
                            {
                                out << "    " << elem << "\n";
                            }
                        };
                        ctrlModeGroups.groups.emplace_back(std::move(group));
                    }
                }
            }
            ARMARX_CHECK_EXPRESSION(robot);
            state = RobotUnitState::InitializingDevices;
        }
        ARMARX_DEBUG << "armarx::RobotUnit::onInitComponent(): onInitRobotUnit()";
        onInitRobotUnit();
        getArmarXManager()->addObject(robotUnitObs);
        ARMARX_DEBUG << "armarx::RobotUnit::onInitComponent(): add emergency stop master";
        {
            emergencyStopMaster = new RobotUnitEmergencyStopMaster {this,  getProperty<std::string>("EmergencyStopTopic").getValue()};
            ManagedIceObjectPtr obj = ManagedIceObjectPtr::dynamicCast(emergencyStopMaster);
            ARMARX_CHECK_EXPRESSION_W_HINT(obj, "RobotUnitEmergencyStopMaster");
            getArmarXManager()->addObject(obj, getProperty<std::string>("EmergencyStopMasterName").getValue(), true);
            auto prx = obj->getProxy(-1);
            try
            {
                getArmarXManager()->getIceManager()->getIceGridSession()->getAdmin()->addObject(prx);
            }
            catch (const IceGrid::ObjectExistsException&)
            {
                getArmarXManager()->getIceManager()->getIceGridSession()->getAdmin()->updateObject(prx);
            }
            catch (std::exception& e)
            {
                ARMARX_WARNING << "add the robot unit as emergency stop master to the ice grid admin!\nThere was an exception:\n" << e.what();
            }
        }
        ARMARX_DEBUG << "armarx::RobotUnit::onInitComponent()...done!";
    }

    void armarx::RobotUnit::onConnectComponent()
    {
        ARMARX_DEBUG << "RobotUnit::onConnectComponent()";
        listenerPrx = getTopic<RobotUnitListenerPrx>(getListenerTopicName());
        debugDrawerPrx = getTopic<DebugDrawerInterfacePrx>(getProperty<std::string>("DebugDrawerUpdatesTopicName").getValue());
        debugObserverPrx = getTopic<DebugObserverInterfacePrx>(getProperty<std::string>("DebugObserverTopicName").getValue());
        onConnectRobotUnit();
        ARMARX_DEBUG << "RobotUnit::onConnectComponent()...done!";
    }

    void armarx::RobotUnit::onDisconnectComponent()
    {
        ARMARX_DEBUG << "RobotUnit::onDisconnectComponent()";
        stopSelfCollisionAvoidance();

        // Disconnecting TCPController and TrajectoryController

        onDisconnectRobotUnit();
        ARMARX_DEBUG << "RobotUnit::onDisconnectComponent()...done!";
    }

    void armarx::RobotUnit::onExitComponent()
    {
        ARMARX_DEBUG << "RobotUnit::onExitComponent()";
        ARMARX_DEBUG << "remove EmergencyStopMaster";
        try
        {
            getArmarXManager()->removeObjectBlocking(ManagedIceObjectPtr::dynamicCast(emergencyStopMaster));
            getArmarXManager()->getIceManager()->removeObject(getProperty<std::string>("EmergencyStopMasterName").getValue());
        }
        catch (...) {}
        ARMARX_DEBUG << "onExitRobotUnit()";
        onExitRobotUnit();
        ARMARX_DEBUG << "finishRunning()";
        finishRunning();
        ARMARX_DEBUG << "RobotUnit::onExitComponent()...done!";
    }

    void armarx::RobotUnit::rtUpdateSensorAndControlBuffer(
        const IceUtil::Time& sensorValuesTimestamp,
        const IceUtil::Time& timeSinceLastIteration)
    {
        rtThreadTimingsSensorDevice->rtMarkRtUpdateSensorAndControlBufferStart();
        SensorAndControl& sc = controlThreadOutputBuffer.getWriteBuffer();
        sc.writeTimestamp = IceUtil::Time::now(); // this has to be in real time
        sc.sensorValuesTimestamp = sensorValuesTimestamp;
        sc.timeSinceLastIteration = timeSinceLastIteration;
        for (std::size_t sensIdx = 0; sensIdx < rtGetNumberOfSensorDevices(); ++sensIdx)
        {
            sensorDevices.at(sensIdx)->getSensorValue()->_copyTo(sc.sensors.at(sensIdx));
        }
        for (std::size_t ctrlIdx = 0; ctrlIdx < controlDevices.size(); ++ctrlIdx)
        {
            ControlDevice& controlDevice = *controlDevices.at(ctrlIdx);
            auto& tv = sc.control.at(ctrlIdx);
            for (std::size_t targIdx = 0; targIdx < controlDevice.rtGetJointControllers().size(); ++targIdx)
            {
                JointController& jointCtrl = *controlDevice.rtGetJointControllers().at(targIdx);
                jointCtrl.getControlTarget()->_copyTo(tv.at(targIdx));
            }
        }
        controlThreadOutputBuffer.commitWrite();

        for (auto& node : nodePositionSensorMap)
        {
            node.first->setJointValueNoUpdate(node.second->position);
        }
        rtRobot->applyJointValues();
        rtThreadTimingsSensorDevice->rtMarkRtUpdateSensorAndControlBufferEnd();
    }

    void armarx::RobotUnit::writeRequestedControllers(JointAndNJointControllers&& setOfControllers)
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        //check NJoint
        const auto& nJointCtrls = setOfControllers.nJointControllers;
        std::set<NJointControllerPtr> nJointSet {nJointCtrls.begin(), nJointCtrls.end()};
        nJointSet.erase(nullptr);
        //# NJoint
        const std::size_t nNJointCtrls = std::count_if(nJointCtrls.begin(), nJointCtrls.end(), [](const NJointControllerPtr & p)
        {
            return p;
        });

        ARMARX_CHECK_EXPRESSION(nNJointCtrls == nJointSet.size());
        ARMARX_CHECK_EXPRESSION(nJointCtrls.size() == getNumberOfControlDevices());
        //first nNJointCtrls not null
        ARMARX_CHECK_EXPRESSION(std::all_of(nJointCtrls.begin(), nJointCtrls.begin() + nNJointCtrls, [](NJointController * p)
        {
            return p;
        }));
        //last getNumberOfControlDevices()-nNJointCtrls null
        ARMARX_CHECK_EXPRESSION(std::all_of(nJointCtrls.begin() + nNJointCtrls, nJointCtrls.end(), [](NJointController * p)
        {
            return !p;
        }));
        //conflict free and sorted
        ARMARX_CHECK_EXPRESSION(std::is_sorted(nJointCtrls.begin(), nJointCtrls.end(), NJointController::IdComp {}));
        ARMARX_CHECK_EXPRESSION(NJointController::AreNotInConflict(nJointCtrls.begin(), nJointCtrls.begin() + nNJointCtrls));

        //check JointCtrl
        const auto& jointCtrls = setOfControllers.jointControllers;
        ARMARX_CHECK_EXPRESSION(jointCtrls.size() == getNumberOfControlDevices());
        ARMARX_CHECK_EXPRESSION(std::all_of(jointCtrls.begin(), jointCtrls.end(), [](JointController * p)
        {
            return p;
        }));

        //check groups
        {
            ARMARX_DEBUG << "check groups";
            std::size_t grpIdx = 0;
            for (const auto& group : ctrlModeGroups.groups)
            {
                ARMARX_CHECK_EXPRESSION(!group.empty());
                const auto hwMode = setOfControllers.jointControllers.at(controlDevices.index(*group.begin()))->getHardwareControlMode();
                ARMARX_DEBUG << "---- group " << grpIdx << " mode '" << hwMode << "'";
                for (const auto& other : group)
                {
                    const auto& dev = controlDevices.index(other);
                    ARMARX_CHECK_EXPRESSION(dev);
                    const auto otherHwMode = setOfControllers.jointControllers.at(dev)->getHardwareControlMode();
                    ARMARX_DEBUG << "-------- '" << other << "'' mode '" << otherHwMode << "'";
                    ARMARX_CHECK_EXPRESSION(otherHwMode == hwMode);
                }
                ++grpIdx;
            }
        }
        //setup assignement
        {
            ARMARX_DEBUG << "setup assignement index";
            setOfControllers.resetAssignement();
            for (std::size_t nJointCtrlIndex = 0; nJointCtrlIndex < setOfControllers.nJointControllers.size(); ++nJointCtrlIndex)
            {
                const NJointController* nJoint = setOfControllers.nJointControllers.at(nJointCtrlIndex);
                if (!nJoint)
                {
                    break;
                }
                ARMARX_DEBUG << "---- " << nJoint->getInstanceName();
                for (std::size_t jointIndex : nJoint->getControlDeviceUsedIndices())
                {
                    ARMARX_CHECK_EXPRESSION(setOfControllers.jointToNJointControllerAssignement.at(jointIndex) == IndexSentinel());
                    setOfControllers.jointToNJointControllerAssignement.at(jointIndex) = nJointCtrlIndex;
                }
            }
        }

        {
            ARMARX_DEBUG << "call checkSetOfControllersToActivate";
            checkSetOfControllersToActivate(setOfControllers);
        }
        //set requested state
        {
            ARMARX_DEBUG << "set requested state for NJoint controllers";
            for (const auto& name2NJoint : nJointControllers)
            {
                const NJointControllerPtr& nJoint = name2NJoint.second;
                SetNJointControllerRequested(nJoint, nJointSet.count(nJoint));
            }
        }
        controllersRequested.getWriteBuffer() = std::move(setOfControllers);
        controllersRequested.commitWrite();
    }

    const JointAndNJointControllers& armarx::RobotUnit::getRequestedControllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controllersRequested.getWriteBuffer();
    }

    const std::vector<armarx::JointController*>& armarx::RobotUnit::getRequestedJointControllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getRequestedControllers().jointControllers;
    }

    const std::vector<armarx::NJointController*>& armarx::RobotUnit::getRequestedNJointControllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getRequestedControllers().nJointControllers;
    }

    bool armarx::RobotUnit::activatedControllersChanged() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controllersActivated.updateReadBuffer();
    }

    const JointAndNJointControllers& armarx::RobotUnit::getActivatedControllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controllersActivated.getReadBuffer();
    }

    const std::vector<armarx::JointController*>& armarx::RobotUnit::getActivatedJointControllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getActivatedControllers().jointControllers;
    }

    const std::vector<armarx::NJointController*>& armarx::RobotUnit::getActivatedNJointControllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getActivatedControllers().nJointControllers;
    }

    bool armarx::RobotUnit::sensorAndControlBufferChanged() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controlThreadOutputBuffer.updateReadBuffer();
    }

    const SensorAndControl& armarx::RobotUnit::getSensorAndControlBuffer() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controlThreadOutputBuffer.getReadBuffer();
    }

    std::ostream& operator<<(std::ostream& o, RobotUnitState s)
    {
        o << std::to_string(s);
        return o;
    }
}

void armarx::RobotUnit::RtLogging::doLogging()
{
    ARMARX_CHECK_EXPRESSION(enabled);
    std::lock_guard<std::mutex> guard {mutex};
    const auto now = IceUtil::Time::now();
    // entries are removed last

    //remove backlog entries
    {
        while (!backlog.empty() && backlog.front().writeTimestamp + keepBacklogFor < now)
        {
            backlog.pop_front();
        }
    }
    //log all
    {
        ru->controlThreadOutputBuffer.foreachNewLoggingEntry([&](const ControlThreadOutputBuffer::Entry & data)
        {
            //base (marker;iteration;timestamp)
            {
                for (auto& pair : entries)
                {
                    Entry& e = *pair.second;
                    e.stream << "\n"
                             << e.marker << ";"
                             << data.iteration << ";"
                             << data.sensorValuesTimestamp.toMicroSeconds();
                    e.marker.clear();
                }
            }
            //sens
            {
                for (std::size_t idxDev = 0; idxDev < data.sensors.size(); ++ idxDev)
                {
                    const SensorValueBase* val = data.sensors.at(idxDev);
                    for (std::size_t idxField = 0; idxField < val->getNumberOfDataFields(); ++ idxField)
                    {
                        const auto str = val->getDataFieldAsString(idxField);
                        for (auto& pair : entries)
                        {
                            Entry& e = *pair.second;
                            if (e.loggedSensorDeviceValues.at(idxDev).at(idxField))
                            {
                                pair.second->stream  << ";" << str;
                            }
                        }
                    }
                }
            }
            //ctrl
            {
                for (std::size_t idxDev = 0; idxDev < data.control.size(); ++ idxDev)
                {
                    const auto& vals = data.control.at(idxDev);
                    for (std::size_t idxVal = 0; idxVal < vals.size(); ++idxVal)
                    {
                        const ControlTargetBase* val = vals.at(idxVal);
                        for (std::size_t idxField = 0; idxField < val->getNumberOfDataFields(); ++ idxField)
                        {
                            const auto str = val->getDataFieldAsString(idxField);
                            for (auto& pair : entries)
                            {
                                Entry& e = *pair.second;
                                if (e.loggedControlDeviceValues.at(idxDev).at(idxVal).at(idxField))
                                {
                                    pair.second->stream  << ";" << str;
                                }
                            }
                        }
                    }
                }
            }
            //store data to backlog
            {
                if (data.writeTimestamp + keepBacklogFor >= now)
                {
                    backlog.emplace_back(data, true); //true for minimal copy
                }
            }
            //print + reset messages
            {
                for (const detail::RtMessageLogEntryBase* ptr : data.messages.getEntries())
                {
                    if (!ptr)
                    {
                        break;
                    }
                    ptr->print(rtThreadId);
                }
            }
        });
    }
    ARMARX_DEBUG_S << ::deactivateSpam() << "the last " << backlog.size() << " iterations are stored";
    //flush all files
    {
        for (auto& pair : entries)
        {
            pair.second->stream << std::flush;
        }
    }

    //remove entries
    {
        std::vector<std::string> toRemove;
        toRemove.reserve(entries.size());
        for (auto& pair : entries)
        {
            if (pair.second->stopLogging)
            {
                //can't remove the current elemet
                //(this would invalidate the current iterator)
                toRemove.emplace_back(pair.first);
            }
        }
        for (const auto& rem : toRemove)
        {
            entries.erase(rem);
        }
    }
}

