/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::RobotUnit
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <thread>
#include <atomic>
#include <chrono>
#include <mutex>
#include <unordered_map>
#include <memory>
#include <fstream>

#include <ArmarXCore/core/util/OnScopeExit.h>
#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/ArmarXManager.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/util/algorithm.h>
#include <ArmarXCore/core/util/TripleBuffer.h>
#include <ArmarXCore/core/services/tasks/TaskUtil.h>
#include <ArmarXCore/observers/DebugObserver.h>
#include <ArmarXCore/interface/components/EmergencyStopInterface.h>

#include <RobotAPI/interface/units/RobotUnit/RobotUnitInterface.h>

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/SceneObjectSet.h>

#include "Units/RobotUnitSubUnit.h"
#include "JointControllers/JointController.h"
#include "NJointControllers/NJointController.h"
#include "Devices/ControlDevice.h"
#include "Devices/SensorDevice.h"
#include "Devices/RTThreadTimingsSensorDevice.h"
#include "util.h"
#include "util/JointAndNJointControllers.h"
#include "RobotUnitObserver.h"
#include "RobotSynchronization.h"

namespace armarx
{
    TYPEDEF_PTRS_HANDLE(RobotUnit);
    class RobotUnitObserver;
    typedef IceInternal::Handle<RobotUnitObserver> RobotUnitObserverPtr;

    class SensorValue1DoFActuatorPosition;
    /**
     * @class RobotUnitPropertyDefinitions
     * @brief
     */
    class RobotUnitPropertyDefinitions:
        public ComponentPropertyDefinitions
    {
    public:
        RobotUnitPropertyDefinitions(std::string prefix):
            ComponentPropertyDefinitions(prefix)
        {
            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            // ////////////////////////////////////////////////////////// General Config /////////////////////////////////////////////////////////// //
            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            defineOptionalProperty<std::uint64_t>(
                "AdditionalObjectSchedulerCount", 10,
                "Number of ObjectSchedulers to be added");

            defineRequiredProperty<std::string>(
                "RobotFileName", "Robot file name, e.g. robot_model.xml");
            defineOptionalProperty<std::string>(
                "RobotFileNameProject", "",
                "Project in which the robot filename is located (if robot is loaded from an external project)");

            defineOptionalProperty<std::string>(
                "RobotNodeSetName", "Robot",
                "Robot node set name as defined in robot xml file, e.g. 'LeftArm'");
            defineOptionalProperty<std::string>(
                "PlatformName", "Platform",
                "Name of the platform (will publish values on PlatformName + 'State')");

            defineOptionalProperty<std::string>(
                "ControlDeviceHardwareControlModeGroups", "",
                "Groups of control devices for which the same hardware control mode has to be set. "
                "For a control mode, the JointControllers of all devices in a group must have the identical hardware control mode. "
                "The kinematic unit will change the control mode of all joints in a group,"
                " if their current hw control mode does not match the new hw control mode. "
                "This can be used if the robot has a multi DOF joint which is represented by multiple  1DOF control devices "
                "(in such a case all of these 1DOF joints should have the same hw control mode, "
                "except he multi DOF joint can handle different control modes at once). "
                "Syntax: semicolon separated groups , each group is CSV of joints "
                "e.g. j1,j2;j3,j4,j5 -> Group 1 = j1+j2 Group 2 = j3+j4+j5; No joint may be in two groups!"
            );

            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            // ////////////////////////////////////////////////////////// Emergency Stop /////////////////////////////////////////////////////////// //
            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            defineOptionalProperty<std::string>(
                "EmergencyStopMasterName", "EmergencyStopMaster",
                "The name used to register as an EmergencyStopMaster");

            defineOptionalProperty<std::string>(
                "EmergencyStopTopic", "EmergencyStop",
                "The name of the topic over which changes of the emergencyStopState are sent.");

            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            // //////////////////////////////////////////////////////////// Publishing ///////////////////////////////////////////////////////////// //
            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            defineOptionalProperty<std::string>(
                "RobotUnitListenerTopicName", "RobotUnitListenerTopic",
                "The topic receiving events for RobotUnitListener");
            defineOptionalProperty<std::string>(
                "DebugDrawerUpdatesTopicName", "DebugDrawerUpdates",
                "The topic receiving events for debug drawing");
            defineOptionalProperty<std::size_t>(
                "PublishPeriodMs", 10,
                "Milliseconds between each publish");

            defineOptionalProperty<bool>(
                "ObserverPublishSensorValues", true,
                "Whether sensor values are send to the observer", PropertyDefinitionBase::eModifiable);
            defineOptionalProperty<bool>(
                "ObserverPublishControlTargets", true,
                "Whether control targets are send to the observer", PropertyDefinitionBase::eModifiable);
            defineOptionalProperty<std::string>(
                "DebugObserverTopicName", "DebugObserver", "The topic where updates are send to");
            defineOptionalProperty<std::uint64_t>(
                "ObserverPrintEveryNthIterations", 1,
                "Only every nth iteration data is send to the debug observer", PropertyDefinitionBase::eModifiable);

            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            // /////////////////////////////////////////////////////////////// Units /////////////////////////////////////////////////////////////// //
            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            defineOptionalProperty<std::string>(
                "KinematicUnitName", "KinematicUnit",
                "The name of the created kinematic unit");

            defineOptionalProperty<std::string>(
                "PlatformUnitName", "PlatformUnit",
                "The name of the created platform unit");

            defineOptionalProperty<std::string>(
                "ForceTorqueUnitName", "ForceTorqueUnit",
                "The name of the created force troque unit (empty = default)");
            defineOptionalProperty<std::string>(
                "ForceTorqueTopicName", "ForceTorqueValues",
                "Name of the topic on which the force torque sensor values are provided");

            defineOptionalProperty<std::string>(
                "InertialMeasurementUnitName", "InertialMeasurementUnit",
                "The name of the created inertial measurement unit (empty = default)");
            defineOptionalProperty<std::string>(
                "IMUTopicName", "IMUValues",
                "Name of the IMU Topic.");

            defineOptionalProperty<bool>(
                "CreateTCPControlUnit", false,
                "If true the TCPControl SubUnit is created and added");
            defineOptionalProperty<std::string>(
                "TCPControlUnitName", "TCPControlUnit",
                "Name of the TCPControlUnit.");
            defineOptionalProperty<bool>(
                "CreateTrajectoryPlayer", true,
                "If true the TrajectoryPlayer SubUnit is created and added");
            defineOptionalProperty<std::string>(
                "TrajectoryControllerUnitName", "TrajectoryPlayer",
                "Name of the TrajectoryControllerUnit. The respective component outside of the RobotUnit is TrajectoryPlayer");

            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            // //////////////////////////////////////////////////////////// Rt Logging ///////////////////////////////////////////////////////////// //
            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            defineOptionalProperty<std::size_t>(
                "RTLoggingPeriodMs", 10,
                "Period of the rt-logging thread in milliseconds. "
                "A high period can cause spikes in disk activity if data is logged to disk. "
                "A low period causes more computation overhead and memory consumption. "
                "If set to 0, all rt-logging is deactivated");

            defineOptionalProperty<std::string>(
                "RTLoggingDefaultLog", "",
                "If rt logging is active and a file path is given, all data is logged to this file.");

            defineOptionalProperty<std::size_t>(
                "RTLoggingMessageNumber", 1000,
                "Number of messages that can be logged in the control thread");
            defineOptionalProperty<std::size_t>(
                "RTLoggingMessageBufferSize", 1024 * 1024,
                "Number of bytes that can be logged in the control thread");
            defineOptionalProperty<std::size_t>(
                "RTLoggingMaxMessageNumber", 16000,
                "Max number of messages that can be logged in the control thread");
            defineOptionalProperty<std::size_t>(
                "RTLoggingMaxMessageBufferSize", 16 * 1024 * 1024,
                "Max number of bytes that can be logged in the control thread");

            defineOptionalProperty<std::size_t>(
                "RTLoggingKeepIterationsForMs", 60 * 1000,
                "All logging data (SensorValues, ControlTargets, Messages) is keept for this duration "
                "and can be dunped in case of an error.");

            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            // ////////////////////////////////////////////////////// Self Collision Avoidance ///////////////////////////////////////////////////// //
            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            defineOptionalProperty<bool>(
                "EnableSelfCollisionAvoidance", true,
                "Whether self collision avoidance should be active or not.", PropertyDefinitionBase::eModifiable);
            defineOptionalProperty<int>(
                "SelfCollisionChecksPerSecond", 10,
                "Frequency [Hz] of self collision checking (default = 10).", PropertyDefinitionBase::eConstant);
            defineOptionalProperty<float>(
                "MinSelfDistance", 20,
                "Minimum allowed distance (mm) between each collision model before entering emergency stop (default = 20).",
                PropertyDefinitionBase::eConstant);
            defineOptionalProperty<std::string>(
                "SelfCollisionModelPairs", "",
                "Comma seperated list of pairs of collision models that should be checked against each other by collision avoidance \n"
                "(e.g. {rnsColModel1,rnsColModel2};{rnsColModel3,rnsColModel4}; .... "
                "says that rnsColModel1 will be only checked against rnsColModel2 and rnsColModel3 will only be checked against rnsColModel4). \n"
                "Following predefined descriptors can be used: 'FLOOR' which represents a virtual collision model of the floor.", PropertyDefinitionBase::eConstant);

            defineOptionalProperty<std::string>(
                "TCPControlUnitName", "TCPControlUnit",
                "Name of the TCPControlUnit.");
            defineOptionalProperty<std::string>(
                "TrajectoryControllerUnitName", "TrajectoryPlayer",
                "Name of the TrajectoryControllerUnit. The respective component outside of the RobotUnit is TrajectoryPlayer");

            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
            // ////////////////////////////////////////////////////// Misc. Properties ///////////////////////////////////////////////////////////// //
            // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //

            defineOptionalProperty<bool>(
                "CheckControllerExecutionTimings", false,
                "Check controller execution timings and print a warning if execution takes long.");
        }
    };

    /// @brief The current state of the multi step initialization of a RobotUnit
    enum class RobotUnitState : std::size_t
    {
        PreComponentInitialization,
        InitializingDevices,
        //ok to use devices/create controllers
        InitializingUnits,
        WaitingForRTThreadInitialization,
        Running,
        Exiting
    };
}
namespace std
{
    std::string to_string(armarx::RobotUnitState s);
}

namespace armarx
{
    std::ostream& operator<<(std::ostream& o, RobotUnitState s);

    /**
     * @brief This class is a friend of NJointController.
     *
     * It encapsulates all access to privates of a NJointController
     * to reduce the amount of code which could break invariants
     * in NJointController and to collect all such code in one place
     * to make this problem more manageable.
     *
     * Also the private functions of NJointController should not be made public to prevent
     * other classes (e.g. KinematicSubUnit) or derived NJointControllers
     * from accessing them
     *
     */
    class RobotUnitNJointControllerAccess
    {
    protected:
        static void RtSetNJointControllerActivatedStatus(const NJointControllerPtr& nJointCtrl);
        static void RtSetNJointControllerDeactivatedStatus(const NJointControllerPtr& nJointCtrl);
        static void RtSetNJointControllerDeactivatedBecauseOfErrorStatus(const NJointControllerPtr& nJointCtrl);
        static void InitNJointControler(
            const NJointControllerPtr& nJointCtrl,
            RobotUnit* ru,
            std::size_t ctrlId,
            std::map<std::string, const JointController*> ctrlDeviceUsedJointController,
            std::vector<char> ctrlDeviceUsedBitmap,
            std::vector<std::size_t> ctrlDeviceUsedIndices,
            bool deletable,
            bool internal
        ) noexcept;
        static void PublishNJointController(
            const NJointControllerPtr& nJointCtrl,
            const SensorAndControl& sac,
            const DebugDrawerInterfacePrx& draw,
            const DebugObserverInterfacePrx& observer
        );
        bool GetNJointControllerStatusReportedActive(const NJointControllerPtr& nJointCtrl);
        bool GetNJointControllerStatusReportedRequested(const NJointControllerPtr& nJointCtrl);
        void UpdateNJointControllerStatusReported(const NJointControllerPtr& nJointCtrl);
        void SetNJointControllerRequested(const NJointControllerPtr& nJointCtrl, bool requested);

        void DeactivateNJointControllerPublishing(
            const NJointControllerPtr& nJointCtrl,
            const DebugDrawerInterfacePrx& draw,
            const DebugObserverInterfacePrx& observer);
    };

    /**
     * @brief The RobotUnit class manages a robot and its controllers.
     *
     * The \ref RobotUnit manages \ref NJointController and \ref JointController for a robot.
     * Controllers are executed in a control thread.
     * \section Units
     * Units are created on an as needed basis.
     * (e.g. If there is a platform, a \ref PlatformSubUnit is created)
     *
     * Available units are
     *
     *  - \ref KinematicSubUnit
     *  - \ref PlatformSubUnit
     *  - \ref ForceTorqueSubUnit
     *
     * \section RtLogging
     * Since the control thread (Section \ref softrt) should adhere soft rt properties,
     * no function executed in it is allowed to block.
     * Hence logging of messages via armarx logging streams (e.g. \ref ARMARX_INFO) is forbidden.
     *
     * There are non-blocking macros to log messages in the control thread:
     *
     *  - \ref ARMARX_RT_LOGF_DEBUG
     *  - \ref ARMARX_RT_LOGF_VERBOSE
     *  - \ref ARMARX_RT_LOGF_INFO
     *  - \ref ARMARX_RT_LOGF_IMPORTANT
     *  - \ref ARMARX_RT_LOGF_WARN
     *  - \ref ARMARX_RT_LOGF_ERROR
     *  - \ref ARMARX_RT_LOGF_FATAL
     *
     * \warning These functions must only be called in the control thread after calling armarx::RobotUnit::initRTThread
     *
     * These macros work similar to printf and write their content to the corresponding
     * armarx logging stream.
     *
     * The stream returns an object of the type \ref armarx::detail::RtMessageLogEntryBase.
     * It is possible to deactivate spam on these streams by calling
     * \ref armarx::detail::RtMessageLogEntryBase::deactivateSpam .
     *
     * E.g.
     * \code
            // prints every 1.5 seconds the message "foo 42 bar 24" to ARMARX_IMPORTANT
            ARMARX_RT_LOGF_IMPORTANT("foo %d bar %d\n", 42, 24).deactivateSpam(1.5);
     * \endcode
     * \note It is not possible to print out strings this would need memory allocation.
     *
     * If an error occours, it is possible to print all recent messages (even the deactivated messages)
     * to a file by calling the ice interface function \ref RobotUnitInterface::writeRecentIterationsToFile.
     * The property "RTLoggingKeepIterationsForMs" decides how long all logging messages are keept (default 60 seconds).
     * The \ref RobotUnitGui has a button to call this function.
     *
     * \section Threads
     * This section explains the running threads and their responsibilities.
     * \subsection ctrlthrd Control Thread
     * This thread should be able to adhere soft rt if PREEMPT_RT is used.
     *
     * Its tasks are:
     *
     *  - Communication with the Robot
     *  - Execution of \ref NJointController and \ref JointController
     *  - Transferring data (\ref SensorValues, \ref ControlTargets) to other threads
     *    (publishing, units, logging)
     *
     * \subsubsection softrt Soft rt
     *
     * To adhere soft rt, the following actions are forbidden (an example and the reason are in parentheses):
     *
     *  - Blocking operations (e.g. lock on a mutex) (can block the thread for a long time)
     *  - Writing data to an armarx stream (e.g. \ref ARMARX_INFO << "foo") (is blocking)
     *  - Allocating / freeing memory on the heap (e.g. std::vector<int>(2,2) ) (is blocking)
     *  - Writing data to a stream (e.g. std::cout << "bar") (may allocate memory)
     *  - Making remote calls (e.g. via ICE) (is blocking)
     *
     * \subsection pubthrd Publish Thread
     *
     * Its tasks are:
     *
     *  - Update Units
     *  - Call publishing hooks of \ref NJointController
     *  - Publish data to \ref RobotUnitListener
     *
     * \subsection logthrd RtLogging Thread
     * If this thread is deactivated, no Rt Logging is possible and
     * no messages from the control Thread are printed.
     *
     * Its tasks are:
     *
     *  - Log SensorValues and ControlTargets to files
     *  - Print messages from the Control Thread
     *
     * \subsection icethrd ICE Threads
     *
     *  - Receive ICE calls
     */
    class RobotUnit :
        virtual public RobotUnitInterface,
        virtual public Component,
        virtual public NJointControllerDescriptionProviderInterface,
        virtual private RobotUnitNJointControllerAccess
    {
    protected:
        using PublisherTaskT = SimplePeriodicTask<std::function<void(void)>>;
        using RTLoggingTaskT = SimplePeriodicTask<std::function<void(void)>>;
    public:
        using MutexType = std::recursive_mutex;
        using GuardType = std::unique_lock<MutexType>;
        using ClockT = std::chrono::high_resolution_clock;
        using TimePointT = std::chrono::high_resolution_clock::time_point;
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // data
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        //devices
    private:
        /**
         * @brief ControlDevices added to this unit
         * data may only be added during and only be used after State::InitializingDevices
         */
        KeyValueVector<std::string, ControlDevicePtr> controlDevices;
        /**
         * @brief SensorDevices added to this unit
         * data may only be added during and only be used after State::InitializingDevices
         */
        KeyValueVector<std::string, SensorDevicePtr > sensorDevices;

        /// @brief const pointer to all ControlDevices (passed to GenerateConfigDescription of a NJointController)
        std::map<std::string, ConstControlDevicePtr> controlDevicesConstPtr;
        /// @brief const pointer to all SensorDevices (passed to GenerateConfigDescription of a NJointController)
        std::map<std::string, ConstSensorDevicePtr> sensorDevicesConstPtr;

        /// @brief a pointer to the RTThreadTimingsSensorDevice used to meassure timings in the rt loop
        RTThreadTimingsSensorDevicePtr rtThreadTimingsSensorDevice;
        /// @brief The device name of rtThreadTimingsSensorDevice
        static const std::string rtThreadTimingsSensorDeviceName;
        //[triple] buffers
    private:
        /// @brief Controllers the RT thread is supposed to activate (direction nonRT -> RT) (some NJoint controllers may be null)
        /// data may only be used after State::InitializingDevices
        /// all NJointController have to be sorted by id (null controllers are at the end)
        WriteBufferedTripleBuffer<JointAndNJointControllers> controllersRequested;
        /// @brief Controllers the RT thread is currently running (direction RT -> nonRT) (some NJoint controllers may be null)
        /// data may only be used after State::InitializingDevices
        /// all NJointController are sorted by id (null controllers may be anywhere)
        WriteBufferedTripleBuffer<JointAndNJointControllers> controllersActivated;
        /// @brief Transfers Data  (current sensor values and control targets) from the control thread to other threads (direction RT -> nonRT)
        /// \warning DO NOT ACCESS EXCEPT YOU KNOW WHAT YOU ARE DOING!
        ControlThreadOutputBuffer controlThreadOutputBuffer;
        //(instantiating) NJoint controllers
    private:
        /// data may only be used after State::InitializingDevices
        std::thread::id initNJointThreadId;
        /// data may only be used after State::InitializingDevices
        std::map<std::string, const JointController*> initNJointControllerUsedJointController;
        /// @brief Holds all currently loaded NJoint controllers (index: [instancename])(May not be accessed in rt.)
        std::map<std::string, NJointControllerPtr> nJointControllers;
        /// @brief These controllers will be deleted in the next iteration of publish
        std::map<std::string, NJointControllerPtr> nJointControllersToBeDeleted;
        /// @brief the id of the next NJointController instance that is created
        std::size_t nJointControllerNextId {0};

        //listeners
    private:
        std::string robotUnitListenerTopicName;
        RobotUnitListenerPrx listenerPrx;
        DebugDrawerInterfacePrx debugDrawerPrx;

        //publishing
    private:
        std::uint64_t publishIterationCount {0};

        IceUtil::Time publishNewSensorDataTime;
        PublisherTaskT::pointer_type publisherTask;
        std::set<std::string> lastReportedClasses;

        std::atomic_bool ObserverPublishSensorValues;
        std::atomic_bool ObserverPublishControlTargets;
        std::atomic<std::uint64_t> debugObserverSkipIterations;

        std::chrono::microseconds lastRobotUnitObserverFlush;
        std::chrono::microseconds lastPublishLoop;
        DebugObserverInterfacePrx debugObserverPrx;
        RobotUnitObserverPtr robotUnitObs;

        ManagedIceObjectPtr tcpControllerSubUnit;
        ManagedIceObjectPtr trajectoryControllerSubUnit;

        //rt logging
    private:
        struct RtLogging
        {
            RtLogging(RobotUnit* ru): ru {ru} {}
            struct Entry
            {
                std::atomic_bool stopLogging {false};
                std::vector<std::vector<std::vector<char>>> loggedControlDeviceValues;
                std::vector<std::vector<char>> loggedSensorDeviceValues;
                std::ofstream stream;
                std::string filename;
                std::string marker;
            };
            mutable std::mutex mutex;
            std::map<std::string, std::shared_ptr<Entry>> entries;
            bool enabled {false};
            std::vector<std::vector<std::vector<std::string>>> controlDeviceValueLoggingNames;
            std::vector<std::vector<            std::string >> sensorDeviceValueLoggingNames;

            RobotUnit* const ru;
            RTLoggingTaskT::pointer_type task;

            RemoteReferenceCounterBasePtr defaultLogHandle;

            std::deque<detail::ControlThreadOutputBufferEntry> backlog;
            IceUtil::Time keepBacklogFor;

            std::size_t loggingTaskTimestep {0};

            Ice::Int rtThreadId {0};
            void doLogging();
        };
        RtLogging rtLogging;

        //other
    private:
        /// @brief The RobotUnit's state
        std::atomic<RobotUnitState> state {RobotUnitState::PreComponentInitialization};
        /// @brief holds all units managed by the RobotUnit
        std::vector<ManagedIceObjectPtr> units;
        /// @brief holds a copy of all units managed by the RobotUnit derived from RobotUnitSubUnit
        /// this is done to prevent casting
        std::vector<RobotUnitSubUnitPtr> subUnits;
        std::thread::id rtThreadId;
        mutable MutexType dataMutex;
        static const std::set<RobotUnitState> devicesReadyStates;

        std::atomic_bool emergencyStop {false};
        std::atomic_bool rtIsInEmergencyStop {false};
        EmergencyStopMasterInterfacePtr emergencyStopMaster;

        struct ControlDeviceHardwareControlModeGroups
        {
            /// @brief contains a vector per group.
            /// Each vector contains the names of devices in this group.
            std::vector<std::set<std::string>> groups;
            /// @brief contains a vector per group.
            /// Each vector contains the indices of devices in this group.
            std::vector<std::vector<std::size_t>> deviceIndices;
            /// @brief contains the names of all devices in any group
            std::set<std::string> groupsMerged;
            /// @brief contains the group index for each device (or the sentinel)
            std::vector<std::size_t> groupIndices;
        };
        ControlDeviceHardwareControlModeGroups ctrlModeGroups;

        struct SelfCollisionAvoidanceData
        {
            VirtualRobot::RobotPtr robot;
            std::map<std::string, VirtualRobot::SceneObjectSetPtr> collisionModels;
            float minDistance {0}; // min allowed distance (mm) between each collision model
            std::vector<std::pair<std::string, std::string>> modelPairs; // list of pairs of collision models that should be checked agaisnt each other
        };

        //robot
    private:
        boost::shared_ptr<VirtualRobot::Robot> robot;
        std::string robotNodeSetName;
        std::string robotProjectName;
        std::string robotFileName;
        std::string robotPlatformName;

        VirtualRobot::RobotPtr rtRobot;
        VirtualRobot::RobotNodeSetPtr rtRobotJointSet;
        std::map<VirtualRobot::RobotNodePtr, const SensorValue1DoFActuatorPosition*> nodePositionSensorMap;

        SelfCollisionAvoidanceData selfCollisionAvoidanceData;
        PeriodicTask<RobotUnit>::pointer_type selfCollisionAvoidanceTask;
        void startSelfCollisionAvoidance();
        void stopSelfCollisionAvoidance();
        void updateSelfCollisionAvoidance();

        RobotSynchronizationPtr selfCollisionAvoidanceRobotSynchronization;

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // util
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        bool checkControllerExecutionTimings = false;
    protected:
        template<class ExceptT = LogicError>
        inline void throwIfStateIsNot(const std::set<RobotUnitState>& sinit, const std::string& fnc, bool onlyWarn = false) const;
        template<class ExceptT = LogicError>
        inline void throwIfStateIsNot(RobotUnitState s, const std::string& fnc, bool onlyWarn = false) const;
        template<class ExceptT = LogicError>
        inline void throwIfDevicesNotReady(const std::string& fnc) const;

        template<class Cont>
        static Ice::StringSeq GetNonNullNames(const Cont& c);

        GuardType getGuard() const;
        inline const KeyValueVector<std::string, SensorDevicePtr>& getSensorDevices() const;
        inline const KeyValueVector<std::string, ControlDevicePtr>& getControlDevices() const;

        void checkNJointControllerClassName(const std::string& className) const;

        //time
        static TimePointT Now();
        static std::chrono::microseconds MicroToNow(TimePointT last);
        static std::chrono::microseconds MicroNow();

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // other
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        RobotUnit(): rtLogging {this} {}
        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;
        std::string getDefaultName() const override;

        static constexpr std::size_t IndexSentinel();
    public:
        void componentPropertiesUpdated(const std::set<std::string>& changedProperties) override;
    protected:
        virtual void onInitRobotUnit();
        virtual void onConnectRobotUnit();
        virtual void onDisconnectRobotUnit();
        /// @brief called in RobotUnit::onExitComponent() before calling RobotUnit::finishRunning
        virtual void onExitRobotUnit();
        void onInitComponent()       final override;
        void onConnectComponent()    final override;
        void onDisconnectComponent() final override;
        void onExitComponent()       final override;
    private:
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // TripleBuffer
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    protected:
        //rt requested
        bool rtRequestedControllersChanged() const;
        const JointAndNJointControllers&       rtGetRequestedControllers()     const;
        const std::vector<JointController*>& rtGetRequestedJointControllers() const;
        const std::vector<NJointController*>& rtGetRequestedNJointControllers() const;
        // rt activated
        void rtWriteActivatedControllers();
        JointAndNJointControllers&       rtGetActivatedControllers();
        std::vector<JointController*>& rtGetActivatedJointControllers();
        std::vector<NJointController*>& rtGetActivatedNJointControllers();
        std::vector<std::size_t>&     rtGetActivatedJointToNJointControllerAssignement();
        // rt sensor and control
        //declaration further down
        //void rtUpdateSensorAndControlBuffer(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);

        // non rt requested
    private:
        void writeRequestedControllers(JointAndNJointControllers&& setOfControllers);
        const JointAndNJointControllers&       getRequestedControllers()     const;
        const std::vector<JointController*>& getRequestedJointControllers() const;
        const std::vector<NJointController*>& getRequestedNJointControllers() const;
        // non rt activated
    private:
        bool activatedControllersChanged() const;
        const JointAndNJointControllers&       getActivatedControllers()     const;
        const std::vector<JointController*>& getActivatedJointControllers() const;
        const std::vector<NJointController*>& getActivatedNJointControllers() const;
        // non rt sensor and control
    private:
        bool sensorAndControlBufferChanged() const;
        const SensorAndControl& getSensorAndControlBuffer() const;

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: InitializingDevices
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    protected:
        void addControlDevice(const ControlDevicePtr& cd);
        void addSensorDevice(const SensorDevicePtr& sd);

        virtual RTThreadTimingsSensorDevicePtr createRTThreadTimingSensorDevice() const;

        void finishDeviceInitialization();

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: InitializingUnits
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    protected:
        ManagedIceObjectPtr createKinematicSubUnit(const Ice::PropertiesPtr& properties, const std::string& positionControlMode = ControlModes::Position1DoF,
                const std::string& velocityControlMode = ControlModes::Velocity1DoF, const std::string& torqueControlMode = ControlModes::Torque1DoF
                                                  );


        virtual void initializeDefaultUnits();
        virtual void initializeKinematicUnit();
        virtual void initializePlatformUnit();
        virtual void initializeForceTorqueUnit();
        virtual void initializeInertialMeasurementUnit();
        virtual void initializeTrajectoryControllerUnit();
        virtual void initializeTcpControllerUnit();
        void addUnit(ManagedIceObjectPtr unit);
        void finishUnitInitialization();
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: Running (RT)
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        virtual void initRTThread();

        //switching helper (rt)
        void rtDeactivateAssignedNJointControllerBecauseOfError(std::size_t ctrlDevIndex, bool writeActivatedControlers = true);
        /**
         * @brief Deactivates a NJointController and sets all used ControlDevices to EmergencyStop
         * @param nJointCtrlIndex The NJointController to deactivate (index in controllersActivated)
         * @param writeActivatedControlers Whether controllersActivated should be committed
         */
        void rtDeactivateNJointControllerBecauseOfError(std::size_t nJointCtrlIndex, bool writeActivatedControlers = true);
        /**
         * @brief Hook for deriving classes (called AFTER a NJointController is deactivated due to an error)
         *
         * This is useful when two Control devices can be controlled by different NJointControllers but either both or none
         * have to be controlled.
         *
         * Since this hook is called after the controller with an error was deactivated an other call to
         * \ref rtDeactivateAssignedNJointControllerBecauseOfError can't lead to a cyclic execution.
         */
        virtual void rtDeactivatedNJointControllerBecauseOfError(const NJointControllerPtr& /*nJointCtrl*/);

        //run (these functions should be called in the control loop)
        bool rtSwitchControllerSetup();
        void rtRunNJointControllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
        void rtHandleInvalidTargets();
        /**
         * @brief Runs Joint controllers and writes target values to ControlDevices
         * @param sensorValuesTimestamp
         * @param timeSinceLastIteration
         */
        void rtRunJointControllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
        void rtReadSensorDeviceValues(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
        //rtBusSendReceive call this robot specific function here
        void rtUpdateSensorAndControlBuffer(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
        void rtResetAllTargets();

        RTThreadTimingsSensorDevice& rtGetRTThreadTimingsSensorDevice();

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: InitializingUnits + Running (NON RT)
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        //create NJoint controllers
        ConstSensorDevicePtr  getSensorDevice(const std::string& sensorDeviceName) const override;
        ConstControlDevicePtr getControlDevice(const std::string& deviceName) const override;
        ControlTargetBase* getControlTarget(const std::string& deviceName, const std::string& controlMode) override;
        NJointControllerInterfacePrx createNJointController(
            const std::string& className, const std::string& instanceName,
            const NJointControllerConfigPtr& config, const Ice::Current&) override;
        NJointControllerInterfacePrx createNJointControllerFromVariantConfig(
            const std::string& className, const std::string& instanceName,
            const StringVariantBaseMap& variants, const Ice::Current&) override;
        const NJointControllerPtr& createNJointController(
            const std::string& className, const std::string& instanceName,
            const NJointControllerConfigPtr& config, bool deletable, bool internal);
        //deleting NJoint controllers
        void deleteNJointController(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
        void deleteNJointControllers(const Ice::StringSeq& names, const Ice::Current& = GlobalIceCurrent) override;
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: Running (NON RT)
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        //switching
        void activateNJointController(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
        void activateNJointControllers(const Ice::StringSeq& names, const Ice::Current& = GlobalIceCurrent) override;
        void deactivateNJointController(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
        void deactivateNJointControllers(const Ice::StringSeq& names, const Ice::Current& = GlobalIceCurrent) override;
        void switchNJointControllerSetup(const Ice::StringSeq& newSetup, const Ice::Current& = GlobalIceCurrent) override;
    private:
        void setActivateControllersRequest(std::set<NJointControllerPtr, NJointController::IdComp>&& ctrls);
        /**
         * @brief This hook is called prior to writing the set of requested controllers.
         * A deriving class can override this function to perform some additional checks
         * (e.g. two controll devices may always need to have the same control mode, this can be checked here)
         * If any check fails, throw an exception with a descriptive message.
         * @param controllers The controllers to activate
         */
        virtual void checkSetOfControllersToActivate(const JointAndNJointControllers& /*controllers*/) const;
        // publishing
    protected:
        virtual void publish(StringVariantBaseMap additionalMap = {}, StringVariantBaseMap timingMap = {});

        //shutting down
    protected:
        ///@brief joint your thread in this hook. (used by RobotUnit::finishRunning())
        virtual void joinRTThread() = 0;
    private:
        ///@brief delete all units (used by RobotUnit::finishRunning())
        virtual void removeAllUnits();
        void finishRunning();

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // Lib loading
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        bool loadLibFromPath(const std::string& path, const Ice::Current& = GlobalIceCurrent) override
        {
            return getArmarXManager()->loadLibFromPath(path);
        }
        bool loadLibFromPackage(const std::string& package, const std::string& lib, const Ice::Current& = GlobalIceCurrent) override
        {
            return getArmarXManager()->loadLibFromPackage(package, lib);
        }

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // querry data
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        //NJoint
        Ice::StringSeq getNJointControllerNames(const Ice::Current& = GlobalIceCurrent) const override;
        Ice::StringSeq getRequestedNJointControllerNames(const Ice::Current& = GlobalIceCurrent) const override;
        Ice::StringSeq getActivatedNJointControllerNames(const Ice::Current& = GlobalIceCurrent) const override;

        NJointControllerInterfacePrx getNJointController(const std::string& name, const Ice::Current&) const override;
        StringNJointControllerPrxDictionary getAllNJointControllers(const Ice::Current&) const override;
        std::vector<NJointControllerPtr> getNJointControllersNotNull(const std::vector<std::string>& names) const;
        const NJointControllerPtr& getNJointControllerNotNull(const std::string& name) const;

        NJointControllerStatus getNJointControllerStatus(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        NJointControllerStatusSeq getNJointControllerStatuses(const Ice::Current&) const override;

        NJointControllerDescription getNJointControllerDescription(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        NJointControllerDescriptionSeq getNJointControllerDescriptions(const Ice::Current&) const override;

        NJointControllerDescriptionWithStatus getNJointControllerDescriptionWithStatus(
            const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        NJointControllerDescriptionWithStatusSeq getNJointControllerDescriptionsWithStatuses(const Ice::Current&) const override;
        //control devices
        Ice::StringSeq getControlDeviceNames(const Ice::Current& = GlobalIceCurrent) const override;
        ControlDeviceDescription getControlDeviceDescription(std::size_t cdidx) const;
        ControlDeviceDescription getControlDeviceDescription(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        ControlDeviceDescriptionSeq getControlDeviceDescriptions(const Ice::Current&) const override;
        ControlDeviceStatus getControlDeviceStatus(std::size_t cdidx) const;
        ControlDeviceStatus getControlDeviceStatus(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        ControlDeviceStatusSeq getControlDeviceStatuses(const Ice::Current&) const override;

        //sensor devices
        Ice::StringSeq getSensorDeviceNames(const Ice::Current&) const override;
        SensorDeviceDescription getSensorDeviceDescription(std::size_t cdidx) const;
        SensorDeviceDescription getSensorDeviceDescription(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        SensorDeviceDescriptionSeq getSensorDeviceDescriptions(const Ice::Current&) const override;
        SensorDeviceStatus getSensorDeviceStatus(std::size_t idx) const;
        SensorDeviceStatus getSensorDeviceStatus(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        SensorDeviceStatusSeq getSensorDeviceStatuses(const Ice::Current&) const override;

        //nJointController classes
        Ice::StringSeq getNJointControllerClassNames(const Ice::Current& = GlobalIceCurrent) const override;
        NJointControllerClassDescription getNJointControllerClassDescription(
            const std::string& className, const Ice::Current& = GlobalIceCurrent) const override;
        NJointControllerClassDescriptionSeq getNJointControllerClassDescriptions(const Ice::Current&) const override;

        //units
        Ice::ObjectProxySeq getUnits(const Ice::Current& = GlobalIceCurrent) const override;
        Ice::ObjectPrx getUnit(const std::string& staticIceId, const Ice::Current&) const override;
        const ManagedIceObjectPtr& getUnit(const std::string& staticIceId) const;
        KinematicUnitInterfacePrx getKinematicUnit(const Ice::Current& = GlobalIceCurrent) const override;
        ForceTorqueUnitInterfacePrx getForceTorqueUnit(const Ice::Current& = GlobalIceCurrent) const override;
        InertialMeasurementUnitInterfacePrx getInertialMeasurementUnit(const Ice::Current& = GlobalIceCurrent) const override;
        PlatformUnitInterfacePrx getPlatformUnit(const Ice::Current& = GlobalIceCurrent) const override;
        //other ice functions
        std::string getListenerTopicName(const Ice::Current& = GlobalIceCurrent) const override;
        bool isRunning(const Ice::Current& = GlobalIceCurrent) const override
        {
            return getRobotUnitState() == RobotUnitState::Running;
        }
        //other local functions
        const boost::shared_ptr<VirtualRobot::Robot>& getRobot() const;
        std::string getRobotName() const;
        const std::string& getRobotNodetSeName() const;
        const std::string& getRobotProjectName() const;
        const std::string& getRobotFileName() const;
        const std::string& getRobotPlatformName() const;

        std::size_t getNumberOfControlDevices() const;
        std::size_t rtGetNumberOfControlDevices() const;

        std::size_t getNumberOfSensorDevices() const;
        std::size_t rtGetNumberOfSensorDevices() const;

        const DebugDrawerInterfacePrx& getDebugDrawerProxy() const;
        bool areDevicesReady() const;
        RobotUnitState getRobotUnitState() const;
        std::string getName() const override;

        void setEmergencyStopState(EmergencyStopState state);

        EmergencyStopState getEmergencyStopState() const;
        const EmergencyStopMasterInterfacePtr& getEmergencyStopMaster() const
        {
            return emergencyStopMaster;
        }

        const SelfCollisionAvoidanceData& getSelfCollisionAvoidanceData() const;

        const VirtualRobot::RobotPtr& getRtRobot() const;


        virtual IceUtil::Time getControlThreadTargetPeriod() const = 0;

    protected:
        const RobotUnitListenerPrx& getListenerProxy() const;
        void icePropertiesInitialized() override;

        // RT-Logging
    public:
        bool isRtLoggingEnabled(const Ice::Current& = GlobalIceCurrent) const override
        {
            return rtLogging.enabled;
        }
        RemoteReferenceCounterBasePtr startRtLogging(const std::string& formatString, const Ice::StringSeq& loggingNames, const Ice::Current& = GlobalIceCurrent) override;
        RemoteReferenceCounterBasePtr startRtLoggingWithAliasNames(const std::string& formatString, const StringStringDictionary& aliasNames, const Ice::Current&   = GlobalIceCurrent) override;
        void addMarkerToRtLog(const RemoteReferenceCounterBasePtr& token, const std::string& marker, const Ice::Current& = GlobalIceCurrent) override;
        void stopRtLogging(const RemoteReferenceCounterBasePtr& token, const Ice::Current& = GlobalIceCurrent) override;

        Ice::StringSeq getLoggingNames(const Ice::Current& = GlobalIceCurrent) const override;

        void writeRecentIterationsToFile(const std::string& formatString, const Ice::Current& = GlobalIceCurrent) const override;

    };
}

#include "RobotUnit.ipp"
