/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::RobotUnit
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once
#include "../EigenForwardDeclarations.h"
#include <string>
#include <chrono>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/observers/variant/TimedVariant.h>
#include <ArmarXCore/interface/observers/VariantBase.h>
namespace armarx
{
    namespace introspection
    {
        template<class T, class = void> struct DataFieldsInfo;
        //    static std::size_t GetNumberOfFields();
        //    static std::string GetFieldAsString(std::size_t i, T field);
        //    static std::vector<std::string> GetFieldNames();
        //    static std::map<std::string, VariantBasePtr> ToVariants(
        //                  T value,
        //                  const std::string& name,
        //                  const IceUtil::Time& timestamp,
        //                  const std::string& frame = "",
        //                  const std::string& agent = "")

        template<class T>
        struct DataFieldsInfo<T, typename std::enable_if<std::is_fundamental<T>::value>::type>
        {
            static std::size_t GetNumberOfFields()
            {
                return 1;
            }
            static std::string GetFieldAsString(std::size_t i, T field)
            {
                ARMARX_CHECK_EQUAL(i, 0);
                return to_string(field);
            }
            static std::vector<std::string> GetFieldNames()
            {
                throw std::logic_error {"Fundamental types have no field names"};
            }
            static std::map<std::string, VariantBasePtr> ToVariants(
                T value,
                const std::string& name,
                const IceUtil::Time& timestamp,
                const std::string& frame = "",
                const std::string& agent = "")
            {
                ARMARX_CHECK_EXPRESSION_W_HINT(frame.empty() && agent.empty(), "There is no framed version for fundamental types");
                return {{name, {new TimedVariant(value, timestamp)}}};
            }
        };

        template<>
        struct DataFieldsInfo<Eigen::Vector3f, void>
        {
            static std::size_t GetNumberOfFields()
            {
                return 3;
            }
            static std::string GetFieldAsString(std::size_t i, const Eigen::Vector3f& field);
            static std::vector<std::string> GetFieldNames()
            {
                return {"x", "y", "z"};
            }
            static std::map<std::string, VariantBasePtr> ToVariants(
                const Eigen::Vector3f& value,
                const std::string& name,
                const IceUtil::Time& timestamp,
                const std::string& frame = "",
                const std::string& agent = "");
        };



#define make_DataFieldsInfo_for_eigen_vector(Type,Num)                                              \
    template<>                                                                                          \
    struct DataFieldsInfo<Eigen::Vector##Num##Type, void>                                               \
    {                                                                                                   \
        static std::size_t GetNumberOfFields()                                                          \
        {                                                                                               \
            return Num;                                                                                 \
        }                                                                                               \
        static std::string GetFieldAsString(std::size_t i, const Eigen::Vector##Num##Type& field);      \
        static std::vector<std::string> GetFieldNames();                                                \
        static std::map<std::string, VariantBasePtr> ToVariants(                                        \
                const Eigen::Vector##Num##Type& value,                                                      \
                const std::string& name,                                                                    \
                const IceUtil::Time& timestamp,                                                             \
                const std::string& frame = "",                                                              \
                const std::string& agent = "");                                                             \
    };

        make_DataFieldsInfo_for_eigen_vector(f, 2)
        make_DataFieldsInfo_for_eigen_vector(f, 4)
        make_DataFieldsInfo_for_eigen_vector(f, 5)
        make_DataFieldsInfo_for_eigen_vector(f, 6)

        make_DataFieldsInfo_for_eigen_vector(d, 2)
        make_DataFieldsInfo_for_eigen_vector(d, 3)
        make_DataFieldsInfo_for_eigen_vector(d, 4)
        make_DataFieldsInfo_for_eigen_vector(d, 5)
        make_DataFieldsInfo_for_eigen_vector(d, 6)

        make_DataFieldsInfo_for_eigen_vector(i, 2)
        make_DataFieldsInfo_for_eigen_vector(i, 3)
        make_DataFieldsInfo_for_eigen_vector(i, 4)
        make_DataFieldsInfo_for_eigen_vector(i, 5)
        make_DataFieldsInfo_for_eigen_vector(i, 6)
#undef make_DataFieldsInfo_for_eigen_vector

        template<>
        struct DataFieldsInfo<Eigen::Matrix4f, void>
        {
            static std::size_t GetNumberOfFields()
            {
                return 16;
            }
            static std::string GetFieldAsString(std::size_t i, const Eigen::Matrix4f& field);
            static std::vector<std::string> GetFieldNames()
            {
                return
                {
                    "00", "01", "02", "03",
                    "10", "11", "12", "13",
                    "20", "21", "22", "23",
                    "30", "31", "32", "33"
                };
            }
            static std::map<std::string, VariantBasePtr> ToVariants(
                const Eigen::Matrix4f& value,
                const std::string& name,
                const IceUtil::Time& timestamp,
                const std::string& frame = "",
                const std::string& agent = "");
        };

        template<>
        struct DataFieldsInfo<Eigen::Quaternionf, void>
        {
            static std::size_t GetNumberOfFields()
            {
                return 4;
            }
            static std::string GetFieldAsString(std::size_t i, const Eigen::Quaternionf& field);
            static std::vector<std::string> GetFieldNames()
            {
                return {"qw", "qx", "qy", "qz"};
            }
            static std::map<std::string, VariantBasePtr> ToVariants(
                const Eigen::Quaternionf& value,
                const std::string& name,
                const IceUtil::Time& timestamp,
                const std::string& frame = "",
                const std::string& agent = "");
        };

        template<>
        struct DataFieldsInfo<std::chrono::microseconds, void>
        {
            static std::size_t GetNumberOfFields()
            {
                return 1;
            }
            static std::string GetFieldAsString(std::size_t i, std::chrono::microseconds field)
            {
                return to_string(field.count());
            }
            static std::vector<std::string> GetFieldNames()
            {
                throw std::logic_error {"should never be called"};
            }
            static std::map<std::string, VariantBasePtr> ToVariants(
                std::chrono::microseconds value,
                const std::string& name,
                const IceUtil::Time& timestamp,
                const std::string& frame = "",
                const std::string& agent = "");
        };
        template<>
        struct DataFieldsInfo<IceUtil::Time, void>
        {
            static std::size_t GetNumberOfFields()
            {
                return 1;
            }
            static std::string GetFieldAsString(std::size_t i, IceUtil::Time field)
            {
                return to_string(field.toMicroSeconds());
            }
            static std::vector<std::string> GetFieldNames()
            {
                throw std::logic_error {"should never be called"};
            }
            static std::map<std::string, VariantBasePtr> ToVariants(
                IceUtil::Time value,
                const std::string& name,
                const IceUtil::Time& timestamp,
                const std::string& frame = "",
                const std::string& agent = "");
        };
    }
}

namespace armarx
{
    struct JointStatus;
    namespace introspection
    {
        template<>
        struct DataFieldsInfo<JointStatus, void>
        {
            static std::size_t GetNumberOfFields()
            {
                return 4;
            }
            static std::string GetFieldAsString(std::size_t i, const JointStatus& field);
            static std::vector<std::string> GetFieldNames()
            {
                return {"error", "operation", "enabled", "emergencyStop"};
            }

            static std::map<std::string, VariantBasePtr> ToVariants(
                const JointStatus& value,
                const std::string& name,
                const IceUtil::Time& timestamp,
                const std::string& frame,
                const std::string& agent);
        };
    }
}

