/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::RobotUnitPlugin
 * \author     Raphael ( ufdrv at student dot kit dot edu )
 * \date       2017
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include <thread>

#include "RobotUnitWidgetBase.h"
#include <ArmarXCore/core/exceptions/Exception.h>
#include <ArmarXCore/core/logging/Logging.h>

namespace armarx
{
    RobotUnitWidgetBase::RobotUnitWidgetBase(QString name, QWidget* parent) :
        QWidget(parent)
    {
        setObjectName(name);
    }

    void RobotUnitWidgetBase::updateContent()
    {
        if (isResetting)
        {
            return;
        }
        std::lock_guard<std::recursive_mutex> guard {mutex};
        resetCount = 0;
        getResettigLabel().setText("Resetting... " + QString::number(resetCount));
        getStackedWidget().setCurrentIndex(0);
        if (!robotUnit)
        {
            return;
        }
        doContentUpdate();
    }

    void RobotUnitWidgetBase::reset(RobotUnitInterfacePrx ru)
    {
        isResetting = true;
        gotResetData = false;
        std::lock_guard<std::recursive_mutex> guard {mutex};
        robotUnit = ru;
        QMetaObject::invokeMethod(this, "doReset", Qt::QueuedConnection);
    }

    void RobotUnitWidgetBase::doReset()
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        clearAll();
        getTreeWidget().clear();
        addFilter();
        if (robotUnit)
        {
            resetTimerId = startTimer(0);
            getStackedWidget().setCurrentIndex(1);
            std::thread
            {
                [&]
                {
                    ARMARX_INFO << "Resetting of widget '" << objectName().toStdString() << "' started";
                    try
                    {
                        getResetData();
                    }
                    catch (...)
                    {
                        ARMARX_WARNING << "Resetting of widget '" << objectName().toStdString() << "' failed: "
                                       << armarx::GetHandledExceptionString();
                    }
                    gotResetData = true;
                    QMetaObject::invokeMethod(this, "updateContent", Qt::QueuedConnection);
                    ARMARX_INFO << "Resetting of widget '" << objectName().toStdString() << "' done!";
                }
            } .detach();
        }
        else
        {
            gotResetData = true;
            QMetaObject::invokeMethod(this, "updateContent", Qt::QueuedConnection);
        }
    }

    void RobotUnitWidgetBase::timerEvent(QTimerEvent*)
    {
        if (isResetting && gotResetData)
        {
            getResettigLabel().setText("Resetting... " + QString::number(++resetCount));
            isResetting = ! addOneFromResetData();
            if (!isResetting)
            {
                killTimer(resetTimerId);
                resetTimerId = 0;
                QMetaObject::invokeMethod(this, "updateContent", Qt::QueuedConnection);
            }
        }
    }
}
