/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::RobotUnitPluginWidgetController
 * \author     Raphael ( ufdrv at student dot kit dot edu )
 * \date       2017
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "RobotUnitPluginWidgetController.h"
#include "../QWidgets/StyleSheets.h"

#include <string>

#include <boost/filesystem.hpp>

#include <QDir>
#include <QSortFilterProxyModel>
#include <QAction>

#include <ArmarXCore/core/system/cmake/CMakePackageFinder.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

using namespace armarx;

RobotUnitPluginWidgetController::RobotUnitPluginWidgetController()
{
    widget.setupUi(getWidget());

    //add subwidget ControlDevices
    {
        controlDevices = new ControlDevicesWidget;
        widget.groupBoxCDev->layout()->addWidget(controlDevices);
        widget.groupBoxCDev->setVisible(false);
        controlDevices->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Expanding);
    }
    //add subwidget SensorDevices
    {
        sensorDevices = new SensorDevicesWidget;
        widget.groupBoxSDev->layout()->addWidget(sensorDevices);
        widget.groupBoxSDev->setVisible(false);
        sensorDevices->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Expanding);
    }
    //add subwidget NJointControllers
    {
        nJointControllers = new NJointControllersWidget;
        widget.groupBoxNJointCtrl->layout()->addWidget(nJointControllers);
        widget.groupBoxNJointCtrl->setVisible(true);
        nJointControllers->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Expanding);
    }
    //add subwidget NJointControllerClasses
    {
        nJointControllerClasses = new NJointControllerClassesWidget;
        widget.groupBoxNJointCtrlClasses->layout()->addWidget(nJointControllerClasses);
        widget.groupBoxNJointCtrlClasses->setVisible(false);
        nJointControllerClasses->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Expanding);
    }
    updateToolBarActionCheckedState();
}


void RobotUnitPluginWidgetController::loadSettings(QSettings* settings)
{
    robotUnitProxyName = settings->value("robotUnitProxyName", QString::fromStdString(robotUnitProxyName)).toString().toStdString();
    nJointControllerClasses->loadSettings(settings);
    nJointControllers->loadSettings(settings);
    controlDevices->loadSettings(settings);
    sensorDevices->loadSettings(settings);
    widget.groupBoxCDev->setVisible(settings->value("cdevVisible", false).toBool());
    widget.groupBoxSDev->setVisible(settings->value("sdevVisible", false).toBool());
    widget.groupBoxNJointCtrl->setVisible(settings->value("ctrlVisible", true).toBool());
    widget.groupBoxNJointCtrlClasses->setVisible(settings->value("classVisible", false).toBool());
    updateToolBarActionCheckedState();
}

void RobotUnitPluginWidgetController::saveSettings(QSettings* settings)
{
    settings->setValue("robotUnitProxyName", QString::fromStdString(robotUnitProxyName));
    nJointControllerClasses->saveSettings(settings);
    nJointControllers->saveSettings(settings);
    controlDevices->saveSettings(settings);
    sensorDevices->saveSettings(settings);
    settings->setValue("cdevVisible", widget.groupBoxCDev->isVisible());
    settings->setValue("sdevVisible", widget.groupBoxSDev->isVisible());
    settings->setValue("ctrlVisible", widget.groupBoxNJointCtrl->isVisible());
    settings->setValue("classVisible", widget.groupBoxNJointCtrlClasses->isVisible());
}

void RobotUnitPluginWidgetController::onInitComponent()
{
    usingProxy(robotUnitProxyName);
    ARMARX_INFO << "RobotUnitPluginWidgetController::onInitComponent()" << std::flush;
}

void RobotUnitPluginWidgetController::onConnectComponent()
{
    robotUnitPrx = getProxy<RobotUnitInterfacePrx>(robotUnitProxyName);
    listenerTopicName = robotUnitPrx->getListenerTopicName();
    usingTopic(listenerTopicName);
    updateToolBarActionCheckedState();
    QMetaObject::invokeMethod(this, "startOnConnectTimer", Qt::QueuedConnection);
}

void RobotUnitPluginWidgetController::onDisconnectComponent()
{
    unsubscribeFromTopic(listenerTopicName);
    robotUnitPrx = nullptr;
    killTimer(onConnectTimerId);
    QMetaObject::invokeMethod(this, "refreshNJointControllersClicked", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "refreshNJointControllerClassesClicked", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "refreshControlDevicesClicked", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "refreshSensorDevicesClicked", Qt::QueuedConnection);
}

QPointer<QDialog> RobotUnitPluginWidgetController::getConfigDialog(QWidget* parent)
{
    if (!dialog)
    {
        dialog = new SimpleConfigDialog(parent);
        dialog->addProxyFinder<RobotUnitInterfacePrx>({"RobotUnit", "The RobotUnit", "*"});
    }
    return qobject_cast<SimpleConfigDialog*>(dialog);
}

void RobotUnitPluginWidgetController::configured()
{
    robotUnitProxyName = dialog->getProxyName("RobotUnit");
}

QPointer<QWidget> RobotUnitPluginWidgetController::getCustomTitlebarWidget(QWidget* parent)
{
    if (customToolbar)
    {
        customToolbar->setParent(parent);
    }
    else
    {
        customToolbar = new QToolBar(parent);
        customToolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
        customToolbar->setIconSize(QSize(16, 16));
        customToolbar->addAction(QIcon(":/icons/view-refresh-7.png"), "", this, SLOT(refreshControlDevicesClicked()))
        ->setToolTip("Update the list of Control Devices");
        showCDevs = new QAction {"Control Devices", customToolbar};
        showCDevs->setCheckable(true);
        showCDevs->setToolTip("Hide/Show the list of Control Devices");
        connect(showCDevs, SIGNAL(toggled(bool)), widget.groupBoxCDev, SLOT(setVisible(bool)));
        customToolbar->addAction(showCDevs);
        customToolbar->addSeparator();

        customToolbar->addAction(QIcon(":/icons/view-refresh-7.png"), "", this, SLOT(refreshSensorDevicesClicked()))
        ->setToolTip("Update the list of Sensor Devices");
        showSDevs = new QAction {"Sensor Devices", customToolbar};
        showSDevs->setCheckable(true);
        showSDevs->setToolTip("Hide/Show the list of Sensor Devices");
        connect(showSDevs, SIGNAL(toggled(bool)), widget.groupBoxSDev, SLOT(setVisible(bool)));
        customToolbar->addAction(showSDevs);
        customToolbar->addSeparator();

        customToolbar->addAction(QIcon(":/icons/view-refresh-7.png"), "", this, SLOT(refreshNJointControllersClicked()))
        ->setToolTip("Update the list of NJointControllers");
        showNJoint = new QAction {"NJointControllers", customToolbar};
        showNJoint->setCheckable(true);
        showNJoint->setToolTip("Hide/Show the list of NJointControllers");
        connect(showNJoint, SIGNAL(toggled(bool)), widget.groupBoxNJointCtrl, SLOT(setVisible(bool)));
        customToolbar->addAction(showNJoint);
        customToolbar->addSeparator();

        customToolbar->addAction(QIcon(":/icons/view-refresh-7.png"), "", this, SLOT(refreshNJointControllerClassesClicked()))
        ->setToolTip("Update the list of NJointController Classes");
        showNJointClasses = new QAction {"NJointController Classes", customToolbar};
        showNJointClasses->setCheckable(true);
        showNJointClasses->setToolTip("Hide/Show the list of NJointControllers Classes");
        connect(showNJointClasses, SIGNAL(toggled(bool)), widget.groupBoxNJointCtrlClasses, SLOT(setVisible(bool)));
        customToolbar->addAction(showNJointClasses);
        customToolbar->addSeparator();

        customToolbar->addAction(QIcon(":/icons/document-save.svg"), "Write log", this, SLOT(writeLogClicked()))
        ->setToolTip("Writes the log to /tmp/");
    }
    updateToolBarActionCheckedState();
    return customToolbar.data();
}

void RobotUnitPluginWidgetController::nJointControllerStatusChanged(const NJointControllerStatus& status, const Ice::Current&)
{
    nJointControllers->nJointControllerStatusChanged(status);
}

void RobotUnitPluginWidgetController::controlDeviceStatusChanged(const ControlDeviceStatus& status, const Ice::Current&)
{
    controlDevices->controlDeviceStatusChanged(status);
}

void RobotUnitPluginWidgetController::sensorDeviceStatusChanged(const SensorDeviceStatus& status, const Ice::Current&)
{
    sensorDevices->sensorDeviceStatusChanged(status);
}

void RobotUnitPluginWidgetController::nJointControllerClassAdded(const std::string& name, const Ice::Current&)
{
    nJointControllerClasses->nJointControllerClassAdded(name);
}

void RobotUnitPluginWidgetController::nJointControllerCreated(const std::string& name, const Ice::Current&)
{
    nJointControllers->nJointControllerCreated(name);
}

void RobotUnitPluginWidgetController::nJointControllerDeleted(const std::string& name, const Ice::Current&)
{
    nJointControllers->nJointControllerDeleted(name);
}

void RobotUnitPluginWidgetController::refreshNJointControllersClicked()
{
    nJointControllers->reset(robotUnitPrx);
}

void RobotUnitPluginWidgetController::refreshNJointControllerClassesClicked()
{
    nJointControllerClasses->reset(robotUnitPrx);
}

void RobotUnitPluginWidgetController::refreshControlDevicesClicked()
{
    controlDevices->reset(robotUnitPrx);
}

void RobotUnitPluginWidgetController::refreshSensorDevicesClicked()
{
    sensorDevices->reset(robotUnitPrx);
}

void RobotUnitPluginWidgetController::writeLogClicked()
{
    robotUnitPrx->writeRecentIterationsToFile("/tmp/RobotUnitLog-{DateTime}");
}

void RobotUnitPluginWidgetController::startOnConnectTimer()
{
    onConnectTimerId = startTimer(100);
}

void RobotUnitPluginWidgetController::updateToolBarActionCheckedState()
{
    if (customToolbar)
    {
        showCDevs->setChecked(widget.groupBoxCDev->isVisible());
        showSDevs->setChecked(widget.groupBoxSDev->isVisible());
        showNJoint->setChecked(widget.groupBoxNJointCtrl->isVisible());
        showNJointClasses->setChecked(widget.groupBoxNJointCtrlClasses->isVisible());
    }
}


void armarx::RobotUnitPluginWidgetController::timerEvent(QTimerEvent*)
{
    if (!robotUnitPrx)
    {
        killTimer(onConnectTimerId);
    }
    if (robotUnitPrx->isRunning())
    {
        refreshNJointControllersClicked();
        refreshNJointControllerClassesClicked();
        refreshControlDevicesClicked();
        refreshSensorDevicesClicked();
        killTimer(onConnectTimerId);
    }
}
