/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::RobotUnit
 * @author     Raphael Grimm ( raphael dot grimm at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <ArmarXCore/interface/core/UserException.ice>
#include <ArmarXCore/interface/core/BasicTypes.ice>
#include <ArmarXCore/interface/core/RemoteReferenceCount.ice>


#include <ArmarXGui/interface/WidgetDescription.ice>

#include <RobotAPI/interface/units/RobotUnit/NJointController.ice>

#include <RobotAPI/interface/units/KinematicUnitInterface.ice>
#include <RobotAPI/interface/units/ForceTorqueUnit.ice>
#include <RobotAPI/interface/units/InertialMeasurementUnit.ice>
#include <RobotAPI/interface/units/PlatformUnitInterface.ice>

module armarx
{
    dictionary<string, Ice::StringSeq> ControlDeviceNameToControlModesDictionary;

    dictionary<string, string> ControlDeviceNameToNJointControllerNameDictionary;

    struct HWControlModeAndTargetType
    {
        string hardwareControlMode;
        string targetType;
    };
    dictionary<string, HWControlModeAndTargetType> ControlModeToHWControlModeAndTargetTypeDictionary;

    struct ControlDeviceDescription
    {
        string deviceName;
        ControlModeToHWControlModeAndTargetTypeDictionary contolModeToTargetType;
        Ice::StringSeq tags;
    };
    sequence<ControlDeviceDescription> ControlDeviceDescriptionSeq;
    struct ControlDeviceStatus
    {
        string deviceName;
        string activeControlMode;
        string requestedControlMode;
        StringToStringVariantBaseMapMap controlTargetValues;
        long timestampUSec = 0;
    };
    sequence<ControlDeviceStatus> ControlDeviceStatusSeq;


    struct SensorDeviceDescription
    {
        string deviceName;
        string sensorValueType;
        Ice::StringSeq tags;
    };
    sequence<SensorDeviceDescription> SensorDeviceDescriptionSeq;
    struct SensorDeviceStatus
    {
        string deviceName;
        StringVariantBaseMap sensorValue;
        long timestampUSec = 0;
    };
    sequence<SensorDeviceStatus> SensorDeviceStatusSeq;


    struct NJointControllerClassDescription
    {
        string className;
        WidgetDescription::Widget configDescription;
    };
    sequence<NJointControllerClassDescription> NJointControllerClassDescriptionSeq;

    interface RobotUnitInterface
    {
        //controllers
        //names
        ["cpp:const"] idempotent Ice::StringSeq getNJointControllerNames();
        ["cpp:const"] idempotent Ice::StringSeq getRequestedNJointControllerNames();
        ["cpp:const"] idempotent Ice::StringSeq getActivatedNJointControllerNames();
        //proxy/information
        ["cpp:const"] idempotent NJointControllerInterface*               getNJointController(string name);
        ["cpp:const"] idempotent StringNJointControllerPrxDictionary      getAllNJointControllers();

        ["cpp:const"] idempotent NJointControllerStatus                   getNJointControllerStatus(string name) throws InvalidArgumentException;
        ["cpp:const"] idempotent NJointControllerStatusSeq                getNJointControllerStatuses();

        ["cpp:const"] idempotent NJointControllerDescription              getNJointControllerDescription(string name) throws InvalidArgumentException;
        ["cpp:const"] idempotent NJointControllerDescriptionSeq           getNJointControllerDescriptions();

        ["cpp:const"] idempotent NJointControllerDescriptionWithStatus    getNJointControllerDescriptionWithStatus(string name) throws InvalidArgumentException;
        ["cpp:const"] idempotent NJointControllerDescriptionWithStatusSeq getNJointControllerDescriptionsWithStatuses();

        //devices
        ["cpp:const"] idempotent Ice::StringSeq                           getControlDeviceNames() throws LogicError;
        ["cpp:const"] idempotent ControlDeviceDescription                 getControlDeviceDescription(string name) throws InvalidArgumentException, LogicError;
        ["cpp:const"] idempotent ControlDeviceDescriptionSeq              getControlDeviceDescriptions() throws LogicError;
        ["cpp:const"] idempotent ControlDeviceStatus                      getControlDeviceStatus(string name) throws InvalidArgumentException, LogicError;
        ["cpp:const"] idempotent ControlDeviceStatusSeq                   getControlDeviceStatuses() throws LogicError;

        ["cpp:const"] idempotent Ice::StringSeq                           getSensorDeviceNames() throws LogicError;
        ["cpp:const"] idempotent SensorDeviceDescription                  getSensorDeviceDescription(string name) throws InvalidArgumentException, LogicError;
        ["cpp:const"] idempotent SensorDeviceDescriptionSeq               getSensorDeviceDescriptions() throws LogicError;
        ["cpp:const"] idempotent SensorDeviceStatus                       getSensorDeviceStatus(string name) throws InvalidArgumentException, LogicError;
        ["cpp:const"] idempotent SensorDeviceStatusSeq                    getSensorDeviceStatuses() throws LogicError;
        //classes
        ["cpp:const"] idempotent Ice::StringSeq                           getNJointControllerClassNames();
        ["cpp:const"] idempotent NJointControllerClassDescription         getNJointControllerClassDescription(string name) throws InvalidArgumentException;
        ["cpp:const"] idempotent NJointControllerClassDescriptionSeq      getNJointControllerClassDescriptions();

        //switching
        void switchNJointControllerSetup(Ice::StringSeq newSetup) throws InvalidArgumentException, LogicError;

        void activateNJointController(string controllerInstanceName) throws InvalidArgumentException, LogicError;
        void activateNJointControllers(Ice::StringSeq controllerInstanceNames) throws InvalidArgumentException, LogicError;
        void deactivateNJointController(string controllerInstanceName)throws InvalidArgumentException, LogicError;
        void deactivateNJointControllers(Ice::StringSeq controllerInstanceNames)throws InvalidArgumentException, LogicError;

        //creting controllers
        NJointControllerInterface* createNJointController(string className, string instanceName, NJointControllerConfig config) throws InvalidArgumentException, LogicError;
        NJointControllerInterface* createNJointControllerFromVariantConfig(string className, string instanceName, StringVariantBaseMap config) throws InvalidArgumentException, LogicError;
        //deleting controllers
        void deleteNJointController(string controllerInstanceName)throws InvalidArgumentException, LogicError;
        void deleteNJointControllers(Ice::StringSeq controllerInstanceNames)throws InvalidArgumentException, LogicError;
        //loading
        bool loadLibFromPath(string path);
        bool loadLibFromPackage(string package, string libname);

        //units
        ["cpp:const"] idempotent Object* getUnit(string staticIceId);
        ["cpp:const"] idempotent Ice::ObjectProxySeq getUnits();
        ["cpp:const"] idempotent KinematicUnitInterface* getKinematicUnit();
        ["cpp:const"] idempotent ForceTorqueUnitInterface* getForceTorqueUnit();
        ["cpp:const"] idempotent InertialMeasurementUnitInterface* getInertialMeasurementUnit();
        ["cpp:const"] idempotent PlatformUnitInterface* getPlatformUnit();

        //other
        ["cpp:const"] idempotent string getListenerTopicName();
        ["cpp:const"] idempotent bool isRunning();

        //rt-logging
        ["cpp:const"] idempotent bool isRtLoggingEnabled();
        ["cpp:const"] idempotent Ice::StringSeq getLoggingNames();
        RemoteReferenceCounterBase startRtLogging(string filePathFormatString, Ice::StringSeq loggingNames) throws LogicError, InvalidArgumentException;
        RemoteReferenceCounterBase startRtLoggingWithAliasNames(string filePathFormatString, StringStringDictionary aliasNames) throws LogicError, InvalidArgumentException;
        void addMarkerToRtLog(RemoteReferenceCounterBase token, string marker) throws LogicError;
        void stopRtLogging(RemoteReferenceCounterBase token) throws LogicError;

        ["cpp:const"] void writeRecentIterationsToFile(string filePathFormatString) throws LogicError, InvalidArgumentException;
    };

    interface RobotUnitListener
    {
        void nJointControllerStatusChanged(NJointControllerStatus status);
        void controlDeviceStatusChanged(ControlDeviceStatus status);
        void sensorDeviceStatusChanged(SensorDeviceStatus status);
        void nJointControllerClassAdded(string className);
        void nJointControllerCreated(string instanceName);
        void nJointControllerDeleted(string instanceName);
    };
};

