#include "NJointJSDMPController.h"



namespace armarx
{

    NJointControllerRegistration<NJointJSDMPController> registrationControllerNJointJSDMPController("NJointJSDMPController");

    std::string NJointJSDMPController::getClassName(const Ice::Current&) const
    {
        return "NJointJSDMPController";
    }

    NJointJSDMPController::NJointJSDMPController(armarx::NJointControllerDescriptionProviderInterfacePtr prov, const armarx::NJointControllerConfigPtr& config, const VirtualRobot::RobotPtr&)
    {
        NJointJointSpaceDMPControllerConfigPtr cfg = NJointJointSpaceDMPControllerConfigPtr::dynamicCast(config);
        ARMARX_CHECK_EXPRESSION_W_HINT(cfg, "Needed type: NJointJointSpaceDMPControllerConfigPtr");

        for (std::string jointName : cfg->jointNames)
        {
            ControlTargetBase* ct = prov->getControlTarget(jointName, ControlModes::VelocityTorque);
            const SensorValueBase* sv = prov->getSensorValue(jointName);
            targets.insert(std::make_pair(jointName, ct->asA<ControlTarget1DoFActuatorVelocity>()));
            positionSensors.insert(std::make_pair(jointName, sv->asA<SensorValue1DoFActuatorPosition>()));
            torqueSensors.insert(std::make_pair(jointName, sv->asA<SensorValue1DoFActuatorTorque>()));
            gravityTorqueSensors.insert(std::make_pair(jointName, sv->asA<SensorValue1DoFGravityTorque>()));
            velocitySensors.insert(std::make_pair(jointName, sv->asA<SensorValue1DoFActuatorVelocity>()));
        }
        if (cfg->jointNames.size() == 0)
        {
            ARMARX_ERROR << "cfg->jointNames.size() == 0";
        }

        dmpPtr.reset(new DMP::UMIDMP(cfg->kernelSize, cfg->DMPKd, cfg->baseMode, cfg->tau));
        timeDuration = cfg->timeDuration;
        phaseL = cfg->phaseL;
        phaseK = cfg->phaseK;
        phaseDist0 = cfg->phaseDist0;
        phaseDist1 = cfg->phaseDist1;
        phaseKp = cfg->phaseKp;
        dimNames = cfg->jointNames;

        targetVels.resize(cfg->jointNames.size());
        NJointJSDMPControllerControlData initData;
        initData.targetJointVels.resize(cfg->jointNames.size());
        for (size_t i = 0; i < cfg->jointNames.size(); ++i)
        {
            initData.targetJointVels[i] = 0;
        }

        reinitTripleBuffer(initData);


        NJointJSDMPControllerSensorData initSensorData;
        initSensorData.currentTime = 0;
        initSensorData.deltaT = 0;
        initSensorData.currentState.resize(cfg->jointNames.size());
        controllerSensorData.reinitAllBuffers(initSensorData);


        deltaT = 0;
    }

    void NJointJSDMPController::controllerRun()
    {
        if (!controllerSensorData.updateReadBuffer())
        {
            return;
        }

        currentState = controllerSensorData.getUpToDateReadBuffer().currentState;
        double deltaT = controllerSensorData.getUpToDateReadBuffer().deltaT;
        if (canVal > 1e-8)
        {

            double phaseStop = 0;
            double error = 0;
            std::vector<double> currentPosition;
            currentPosition.resize(dimNames.size());

            for (size_t i = 0; i < currentState.size(); i++)
            {
                DMP::DMPState currentPos = currentState[i];
                currentPosition[i] = currentPos.pos;
                error += pow(currentPos.pos - targetState[i], 2);
            }

            double phaseDist;

            if (isDisturbance)
            {
                phaseDist = phaseDist1;
            }
            else
            {
                phaseDist = phaseDist0;
            }

            error = sqrt(error);
            phaseStop = phaseL / (1 + exp(-phaseK * (error - phaseDist)));
            double mpcFactor = 1 - (phaseStop / phaseL);

            if (mpcFactor < 0.1)
            {
                isDisturbance = true;
            }

            if (mpcFactor > 0.9)
            {
                isDisturbance = false;
            }

            canVal -= 1 / tau * deltaT * 1 / (1 + phaseStop);
            double dmpDeltaT = deltaT / timeDuration;
            dmpPtr->setTemporalFactor(tau);

            currentState = dmpPtr->calculateDirectlyVelocity(currentState, canVal / timeDuration, dmpDeltaT, targetState);

            for (size_t i = 0; i < currentState.size(); ++i)
            {
                double vel0 = currentState[i].vel / timeDuration;
                double vel1 = phaseKp * (targetState[i] - currentPosition[i]);
                double vel = mpcFactor * vel0 + (1 - mpcFactor) * vel1;
                targetVels[i] = finished ? 0.0f : vel;

                debugOutputData.getWriteBuffer().latestTargetVelocities[dimNames[i]] = vel;
            }

            debugOutputData.getWriteBuffer().currentCanVal = canVal;
            debugOutputData.getWriteBuffer().mpcFactor = deltaT;
            debugOutputData.commitWrite();

        }
        else
        {
            finished = true;
            for (size_t i = 0; i < dimNames.size(); ++i)
            {
                targetVels[i] = 0;
            }
        }

        LockGuardType guard {controlDataMutex};
        getWriterControlStruct().targetJointVels = targetVels;
        writeControlStruct();
    }

    void NJointJSDMPController::rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        for (size_t i = 0; i < dimNames.size(); i++)
        {
            const auto& jointName = dimNames.at(i);
            DMP::DMPState currentPos;
            currentPos.pos = (positionSensors.count(jointName) == 1) ? positionSensors[jointName]->position : 0.0f;
            currentPos.vel = (velocitySensors.count(jointName) == 1) ? velocitySensors[jointName]->velocity : 0.0f;
            currentPos.vel *= timeDuration;
            controllerSensorData.getWriteBuffer().currentState[i] = currentPos;
        }
        controllerSensorData.getWriteBuffer().deltaT = timeSinceLastIteration.toSecondsDouble();
        controllerSensorData.getWriteBuffer().currentTime += controllerSensorData.getWriteBuffer().deltaT;

        controllerSensorData.commitWrite();

        std::vector<double> targetJointVels = rtGetControlStruct().targetJointVels;
        for (size_t i = 0; i < dimNames.size(); ++i)
        {
            targets[dimNames[i]]->velocity = targetJointVels[i];
        }


    }

    void NJointJSDMPController::learnDMPFromFiles(const Ice::StringSeq& fileNames, const Ice::Current&)
    {
        DMP::Vec<DMP::SampledTrajectoryV2 > trajs;

        DMP::DVec ratios;
        for (size_t i = 0; i < fileNames.size(); ++i)
        {
            DMP::SampledTrajectoryV2 traj;
            traj.readFromCSVFile(fileNames.at(i));
            traj = DMP::SampledTrajectoryV2::normalizeTimestamps(traj, 0, 1);
            trajs.push_back(traj);

            if (i == 0)
            {
                ratios.push_back(1.0);
            }
            else
            {
                ratios.push_back(0.0);
            }
        }
        dmpPtr->learnFromTrajectories(trajs);
        dmpPtr->setOneStepMPC(true);
        dmpPtr->styleParas = dmpPtr->getStyleParasWithRatio(ratios);

        ARMARX_INFO << "Learned DMP ... ";
    }

    void NJointJSDMPController::runDMP(const Ice::DoubleSeq&  goals, double tau, const Ice::Current&)
    {
        targetState.clear();
        targetState.resize(dimNames.size());
        currentState.clear();
        currentState.resize(dimNames.size());
        for (size_t i = 0; i < dimNames.size(); i++)
        {
            const auto& jointName = dimNames.at(i);
            DMP::DMPState currentPos;
            currentPos.pos = (positionSensors.count(jointName) == 1) ? positionSensors[jointName]->position : 0.0f;
            currentPos.vel = (velocitySensors.count(jointName) == 1) ? velocitySensors[jointName]->velocity : 0.0f;

            currentState[i] = currentPos;
            targetState.push_back(currentPos.pos);
        }
        dmpPtr->prepareExecution(goals, currentState, 1,  tau);

        this->goals = goals;
        this->tau = tau;
        canVal = timeDuration;
        finished = false;
        isDisturbance = false;

        ARMARX_INFO << "run DMP...";
        controllerTask->start();
    }

    void NJointJSDMPController::showMessages(const Ice::Current&)
    {
    }

    void NJointJSDMPController::setTemporalFactor(double tau, const Ice::Current&)
    {
        LockGuardType guard(controllerMutex);
        this->tau = tau;
    }

    void NJointJSDMPController::rtPreActivateController()
    {
    }

    void NJointJSDMPController::rtPostDeactivateController()
    {

    }

    void NJointJSDMPController::onPublish(const SensorAndControl&, const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx& debugObs)
    {
        StringVariantBaseMap datafields;
        auto values = debugOutputData.getUpToDateReadBuffer().latestTargetVelocities;
        for (auto& pair : values)
        {
            datafields[pair.first] = new Variant(pair.second);
        }

        datafields["canVal"] = new Variant(debugOutputData.getUpToDateReadBuffer().currentCanVal);
        datafields["mpcFactor"] = new Variant(debugOutputData.getUpToDateReadBuffer().mpcFactor);
        debugObs->setDebugChannel("latestDMPTargetVelocities", datafields);
    }


    void NJointJSDMPController::onInitComponent()
    {
        ARMARX_INFO << "init ...";
        controllerTask = new PeriodicTask<NJointJSDMPController>(this, &NJointJSDMPController::controllerRun, 0.2);
    }

    void NJointJSDMPController::onDisconnectComponent()
    {
        controllerTask->stop();
        ARMARX_INFO << "stopped ...";

    }




}
