/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "PositionControllerHelper.h"

#include <ArmarXCore/core/time/TimeUtil.h>

using namespace armarx;

PositionControllerHelper::PositionControllerHelper(const VirtualRobot::RobotNodePtr& tcp, const VelocityControllerHelperPtr& velocityControllerHelper, const Eigen::Matrix4f& target)
    : posController(tcp), velocityControllerHelper(velocityControllerHelper), currentWaypointIndex(0)
{
    waypoints.push_back(target);
}

PositionControllerHelper::PositionControllerHelper(const VirtualRobot::RobotNodePtr& tcp, const VelocityControllerHelperPtr& velocityControllerHelper, const std::vector<Eigen::Matrix4f>& waypoints)
    : posController(tcp), velocityControllerHelper(velocityControllerHelper), waypoints(waypoints), currentWaypointIndex(0)
{
}

PositionControllerHelper::PositionControllerHelper(const VirtualRobot::RobotNodePtr& tcp, const VelocityControllerHelperPtr& velocityControllerHelper, const std::vector<PosePtr>& waypoints)
    : posController(tcp), velocityControllerHelper(velocityControllerHelper), currentWaypointIndex(0)
{
    for (const PosePtr& pose : waypoints)
    {
        this->waypoints.push_back(pose->toEigen());
    }
}

void PositionControllerHelper::update()
{
    if (!isLastWaypoint() && isCurrentTargetNear())
    {
        currentWaypointIndex++;
    }
    Eigen::VectorXf cv = posController.calculate(getCurrentTarget(), VirtualRobot::IKSolver::All);
    velocityControllerHelper->setTargetVelocity(cv);
}

void PositionControllerHelper::setNewWaypoints(const std::vector<Eigen::Matrix4f>& waypoints)
{
    this->waypoints = waypoints;
    currentWaypointIndex = 0;
}

void PositionControllerHelper::setNewWaypoints(const std::vector<PosePtr>& waypoints)
{
    this->waypoints.clear();
    for (const PosePtr& pose : waypoints)
    {
        this->waypoints.push_back(pose->toEigen());
    }
    currentWaypointIndex = 0;
}

void PositionControllerHelper::addWaypoint(const Eigen::Matrix4f& waypoint)
{
    this->waypoints.push_back(waypoint);
}

void PositionControllerHelper::setTarget(const Eigen::Matrix4f& target)
{
    waypoints.clear();
    waypoints.push_back(target);
    currentWaypointIndex = 0;
}

float PositionControllerHelper::getPositionError() const
{
    return posController.getPositionError(getCurrentTarget());
}

float PositionControllerHelper::getOrientationError() const
{
    return posController.getOrientationError(getCurrentTarget());
}

bool PositionControllerHelper::isCurrentTargetReached() const
{
    return getPositionError() < thresholdPositionReached && getOrientationError() < thresholdOrientationReached;
}

bool PositionControllerHelper::isCurrentTargetNear() const
{
    return getPositionError() < thresholdPositionNear && getOrientationError() < thresholdOrientationNear;
}

bool PositionControllerHelper::isFinalTargetReached() const
{
    return isLastWaypoint() && isCurrentTargetReached();
}

bool PositionControllerHelper::isFinalTargetNear() const
{
    return isLastWaypoint() && isCurrentTargetNear();
}

bool PositionControllerHelper::isLastWaypoint() const
{
    return currentWaypointIndex + 1 >= waypoints.size();
}

const Eigen::Matrix4f& PositionControllerHelper::getCurrentTarget() const
{
    return waypoints.at(currentWaypointIndex);
}

size_t PositionControllerHelper::skipToClosestWaypoint(float rad2mmFactor)
{
    float dist = FLT_MAX;
    size_t minIndex = 0;
    for (size_t i = 0; i < waypoints.size(); i++)
    {
        float posErr = posController.getPositionError(waypoints.at(i));
        float oriErr = posController.getOrientationError(waypoints.at(i));
        float d = posErr + oriErr * rad2mmFactor;
        if (d < dist)
        {
            minIndex = i;
            dist = d;
        }
    }
    currentWaypointIndex = minIndex;
    return currentWaypointIndex;
}

void PositionControllerHelper::setNullSpaceControl(bool enabled)
{
    velocityControllerHelper->setNullSpaceControl(enabled);
}
